/*
 *  WindowMaker window manager
 * 
 *  Copyright (c) 1997 Alfredo K. Kojima
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <stdio.h>
#include <unistd.h>

#include "generic/wwmlib.h"
#include "WindowMaker.h"
#include "core.h"
#include "window.h"
#include "config.h"
#include "client.h"
#include "icon.h"
#include "funcs.h"

#ifdef SOUNDS
#include "sound.h"
#endif

/****** Global Variables ******/
extern WWindow *wWindowList;

extern Time LastTimestamp;

extern Cursor wCursor[WCUR_LAST];


/*
 *---------------------------------------------------------------------- 
 * wSetFocusTo--
 * 	Changes the window focus to the one passed as argument.
 * If the window to focus is not already focused, it will be brought
 * to the head of the list of windows. Previously focused window is
 * unfocused.
 * 
 * Side effects:
 * 	Window list may be reordered and the window focus is changed.
 * 
 *---------------------------------------------------------------------- 
 */
void
wSetFocusTo(WWindow  *wwin)
{
    if (wwin==NULL) {
	/* set focus to the no focus window */
	return;
    }    
    if (!wwin->window_flags.focusable) {
	return;
    }
    if (!wwin->flags.shaded && !wwin->flags.miniaturized) {
	/* set input focus */
	switch (wwin->focus_mode) {
	 case WFM_NO_INPUT:
	    break;
	    
	 case WFM_PASSIVE:
	 case WFM_LOCALLY_ACTIVE:
	    XSetInputFocus(dpy, wwin->client_win, RevertToParent, 
			   LastTimestamp);
	    break;
	    
	 case WFM_GLOBALLY_ACTIVE:
	    break;
	}
    }
    /* if this is not already focused */
    if (wWindowList!=wwin) {
	/* unfocus previous window */	
	wWindowUnfocus(wWindowList);
	
	/* change the window list order */
	if (wwin->prev)
	  wwin->prev->next=wwin->next;
	if (wwin->next)
	  wwin->next->prev=wwin->prev;
	/* bring to head */
	wwin->next = NULL;
	wwin->prev = wWindowList;
	wWindowList->next = wwin;
	wWindowList = wwin;
    }
    wWindowFocus(wwin);	    
}


void 
wRaiseWindow(WWindow  *wwin)
{
    WWindow *ww;
    int wcount=0;

    XRaiseWindow(dpy, wwin->frame->window);
    return;
    for (ww = wWindowList; ww!=NULL; ww=ww->prev) {
	wcount++;
    }
    
}

void 
wLowerWindow(WWindow  *wwin)
{
    if (wwin->window_flags.floating) return;
    XLowerWindow(dpy, wwin->frame->window);
}


static void
selectWindowsInside(int x1, int y1, int x2, int y2)
{
    WWindow *tmpw;
    
    tmpw = wWindowList;
    while (tmpw!=NULL) {
	if ((tmpw->frame_x >= x1) && (tmpw->frame_y >= y1) &&
	    (tmpw->frame->width+tmpw->frame_x <= x2) &&
	    (tmpw->frame->height+tmpw->frame_y <= y2)) {
	    XSetWindowBorder(dpy, tmpw->frame->window, 
			     tmpw->screen_ptr->white_pixel);
	}
	tmpw = tmpw->prev;
    }
}


void
wSelectWindows(WScreen *scr, XEvent *ev)
{
    XEvent event;
    Window root = scr->root_win;
    GC gc = scr->frame_gc;
    int xp = ev->xbutton.x_root;
    int yp = ev->xbutton.y_root;
    int w=0, h=0;
    int x=xp, y=yp;

#ifdef DEBUG
    puts("Selecting windows");
#endif
    if (XGrabPointer(dpy, scr->root_win, False, ButtonMotionMask
		     |ButtonReleaseMask, GrabModeAsync, GrabModeAsync, None, 
		     wCursor[WCUR_DEFAULT], CurrentTime) !=Success) {
	wWarning("pointer grab failed for window selection");
	return;
    }
    XGrabServer(dpy);
    XDrawRectangle(dpy, root, gc, xp, yp, w, h);    
    while(1) {
	XNextEvent(dpy, &event);
	switch (event.type) {
	 case MotionNotify:
	    XDrawRectangle(dpy, root, gc, x, y, w, h);
	    x=event.xmotion.x_root;
	    if (x<xp) {
		w = xp-x;
	    } else {
		w = x-xp;
		x = xp;
	    }
	    y=event.xmotion.y_root;
	    if (y<yp) {
		h = yp-y;
	    } else {
		h = y-yp;
		y = yp;
	    }
	    XDrawRectangle(dpy, root, gc, x, y, w, h);
	    break;

	 case ButtonRelease:
	    XDrawRectangle(dpy, root, gc, x, y, w, h);
	    XUngrabServer(dpy);
	    XUngrabPointer(dpy, CurrentTime);
	    selectWindowsInside(x, y, x+w, y+h);
#ifdef DEBUG
	    puts("End window selection");
#endif
	    return;

	 default:
	    DispatchEvent(&event);
	}
    }
}




void
wShadeWindow(WWindow  *wwin)
{
#ifdef ANIMATIONS    
    int y, s, w, h;

    /* do the shading animation */
    XLowerWindow(dpy, wwin->client_win);    
    h = wwin->frame->height;
    s = h/SHADE_ANIMATION_STEPS;
    w = wwin->frame->width;
    y = wwin->top_width;
    while (h>wwin->top_width) {
	XMoveWindow(dpy, wwin->client_win, 0, y);
	XResizeWindow(dpy, wwin->frame->window, w, h);
	XSync(dpy, 0);
#if (SHADE_ANIMATION_DELAY > 0)	
	usleep(SHADE_ANIMATION_DELAY);
#endif
	h-=s;
	y-=s;
    }
    XMoveWindow(dpy, wwin->client_win, 0, wwin->top_width);
#endif  /* ANIMATIONS */
    wwin->flags.shaded=1;
    wwin->flags.mapped=0;
    /* prevent window withdrawal when getting UnmapNotify */
    XSelectInput(dpy, wwin->client_win, CLIENT_EVENTS & ~StructureNotifyMask);
    XUnmapWindow(dpy, wwin->client_win);
    XSelectInput(dpy, wwin->client_win, CLIENT_EVENTS);
    
    /* for the client it's just like iconification */
    wCoreConfigure(wwin->frame, wwin->frame_x, wwin->frame_y,
		   0,wwin->top_width-1);
    wClientSetState(wwin, IconicState);
    XSync(dpy, 0);
}


void
wUnshadeWindow(WWindow  *wwin)
{
#ifdef ANIMATIONS
    int y, s, w, h;    
#endif /* ANIMATIONS */
    wwin->flags.shaded=0;
    wwin->flags.mapped=1;
    XMapWindow(dpy, wwin->client_win);
    wClientSetState(wwin, NormalState);
#ifdef ANIMATIONS
    /* do the shading animation */    
    XMoveWindow(dpy, wwin->client_win, 0, -wwin->client.height);
    h = wwin->top_width; 
    y = wwin->client.height + wwin->top_width + wwin->bottom_width;
    s = y/SHADE_ANIMATION_STEPS;
    w = wwin->frame->width;
    y = -y;
    while (h<wwin->client.height + wwin->top_width + wwin->bottom_width) {
	XResizeWindow(dpy, wwin->frame->window, w, h);
	XMoveWindow(dpy, wwin->client_win, 0, y);
	XSync(dpy, 0);
#if (SHADE_ANIMATION_DELAY > 0)	
	usleep(SHADE_ANIMATION_DELAY);
#endif
	h+=s;
	y+=s;
    }
    XMoveWindow(dpy, wwin->client_win, 0, wwin->top_width);
#endif /* ANIMATIONS */

    wCoreConfigure(wwin->frame, wwin->frame_x, wwin->frame_y,
		   0,wwin->top_width+wwin->client.height+wwin->bottom_width);
    /* if the window is focused, set the focus again as it was disabled during
     * shading */
    if (wwin->flags.focused)
      wSetFocusTo(wwin);
    XSync(dpy, 0);
}


void
wMaximizeWindow(WWindow *wwin, int full_maximize)
{
    int new_width, new_height, new_x, new_y;
    
    wwin->flags.maximized = 1;
    wwin->old_geometry.width = wwin->frame->width;
    wwin->old_geometry.height = wwin->frame->height;
    wwin->old_geometry.x = wwin->frame_x;
    wwin->old_geometry.y = wwin->frame_y;
    if (full_maximize) {
	new_x = 0;
	new_width = wwin->screen_ptr->scr_width;
    } else {
	new_x = wwin->frame_x;
	new_width = wwin->frame->width;
    }
    new_y = 0;
    new_height = wwin->screen_ptr->scr_height;
    wWindowConstrainSize(wwin, &new_width, &new_height);

    wWindowConfigure(wwin, new_x, new_y, new_width, new_height);
}


void
wUnmaximizeWindow(WWindow *wwin)
{
    wwin->flags.maximized = 0;
    wWindowConfigure(wwin, wwin->old_geometry.x, wwin->old_geometry.y, 
		     wwin->old_geometry.width, wwin->old_geometry.height);
}


static void 
animateResize(WScreen *scr, int x, int y, int w, int h, 
	      int fx, int fy, int fw, int fh)
{
    float cx, cy, cw, ch;
    float xstep, ystep, wstep, hstep;
    int i;
    
    xstep = (float)(fx-x)/MINIATURIZE_ANIMATION_STEPS;
    ystep = (float)(fy-y)/MINIATURIZE_ANIMATION_STEPS;
    wstep = (float)(fw-w)/MINIATURIZE_ANIMATION_STEPS;
    hstep = (float)(fh-h)/MINIATURIZE_ANIMATION_STEPS;

    cx = (float)x;
    cy = (float)y;
    cw = (float)w;
    ch = (float)h;
    XGrabServer(dpy);
    for (i=0; i<MINIATURIZE_ANIMATION_STEPS; i++) {
	XDrawRectangle(dpy, scr->root_win, scr->frame_gc, 
		       (int)cx, (int)cy, (int)cw, (int)ch);
	XSync(dpy, 0);
	usleep(MINIATURIZE_ANIMATION_DELAY);
	XDrawRectangle(dpy, scr->root_win, scr->frame_gc,
		       (int)cx, (int)cy, (int)cw, (int)ch);
	cx+=xstep;
	cy+=ystep;
	cw+=wstep;
	ch+=hstep;
    }
    XUngrabServer(dpy);
}


static void
flushExpose()
{
    XEvent tmpev;
    
    while (XCheckTypedEvent(dpy, Expose, &tmpev))
      DispatchEvent(&tmpev);
}

static void
unmapTransientsFor(WWindow *wwin)
{
    WWindow *tmp;

    tmp = wWindowList;
    while (tmp) {
	if (tmp->transient_for == wwin->client_win) {
	    tmp->flags.miniaturized=1;
	    tmp->flags.mapped=0;
	    XSelectInput(dpy, tmp->client_win, 
			 CLIENT_EVENTS & ~StructureNotifyMask);
	    XUnmapWindow(dpy, tmp->client_win);
	    XSelectInput(dpy, tmp->client_win, CLIENT_EVENTS);

	    XUnmapWindow(dpy, tmp->frame->window);
	    wClientSetState(tmp, IconicState);
	}
	tmp = tmp->prev;
    }
}


static void
mapTransientsFor(WWindow *wwin)
{
    WWindow *tmp;

    tmp = wWindowList;
    while (tmp) {
	if (tmp->transient_for == wwin->client_win) {
	    tmp->flags.miniaturized=0;
	    tmp->flags.mapped=1;
	    XMapSubwindows(dpy, tmp->frame->window);
	    XMapRaised(dpy, tmp->frame->window);
	    wClientSetState(tmp, NormalState);
	}
	tmp = tmp->prev;
    }
}


void
wIconifyWindow(WWindow *wwin)
{
    WWindow  *tmp;
    
    /* icon creation may take a while */
    XGrabPointer(dpy, wwin->screen_ptr->root_win, False, 
		 ButtonMotionMask|ButtonReleaseMask, GrabModeAsync, 
		 GrabModeAsync, None, wCursor[WCUR_WAIT], CurrentTime);

    wwin->flags.miniaturized=1;
    wwin->flags.mapped=0;
    /* make icon position code */
    wwin->icon_x = 0;
    wwin->icon_y = wwin->screen_ptr->scr_height-ICON_HEIGHT;
    wwin->icon = wIconCreate(wwin);

    /* unmap transients */
    unmapTransientsFor(wwin);

    XUngrabPointer(dpy, CurrentTime);
    /* prevent window withdrawal when getting UnmapNotify */
    XSelectInput(dpy, wwin->client_win, CLIENT_EVENTS & ~StructureNotifyMask);
    XUnmapWindow(dpy, wwin->client_win);
    XSelectInput(dpy, wwin->client_win, CLIENT_EVENTS);

    XUnmapWindow(dpy, wwin->frame->window);
    wClientSetState(wwin, IconicState);
    flushExpose();
#ifdef ANIMATIONS
    animateResize(wwin->screen_ptr, wwin->frame_x, wwin->frame_y, 
		  wwin->frame->width, wwin->frame->height,
		  wwin->icon_x, wwin->icon_y,
		  wwin->icon->core->width, wwin->icon->core->height);
#endif
    XMapRaised(dpy, wwin->icon->core->window);
    tmp = wwin->prev;
    while (tmp!=NULL) {
	if (!tmp->flags.miniaturized && !tmp->flags.shaded)
	  break;
	tmp = tmp->prev;
    }
    wSetFocusTo(tmp);
    XSync(dpy, 0);
}


void
wDeiconifyWindow(WWindow  *wwin)
{
    wwin->flags.miniaturized=0;
    wwin->flags.mapped=1;
    XUnmapWindow(dpy, wwin->icon->core->window);
#ifdef ANIMATIONS
    animateResize(wwin->screen_ptr, wwin->icon_x, wwin->icon_y,
		  wwin->icon->core->width, wwin->icon->core->height,
		  wwin->frame_x, wwin->frame_y, 
		  wwin->frame->width, wwin->frame->height);
#endif
    XGrabServer(dpy);
    XMapSubwindows(dpy, wwin->frame->window);
    XMapRaised(dpy, wwin->frame->window);
    wClientSetState(wwin, NormalState);

    mapTransientsFor(wwin);

    wIconDestroy(wwin->icon);
    wwin->icon = NULL;
    XUngrabServer(dpy);
    wSetFocusTo(wwin);
    XSync(dpy,0);
}
