/*
 *  WindowMaker window manager
 * 
 *  Copyright (c) 1997 Alfredo K. Kojima
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <stdlib.h>
#include <stdio.h>

#include "generic/wwmlib.h"
#include "WindowMaker.h"
#include "core.h"
#include "box.h"
#include "button.h"
#include "config.h"

static void buttonMouseDown(WObjDescriptor *desc, XButtonEvent *bev);


WMethodTable wButtonMethods = {
    (WEventHandler)buttonMouseDown,
      (WEventHandler)NULL,
      (WEventHandler)NULL
};

/*
 *---------------------------------------------------------------------- 
 * wButtonCreate--
 * 	Creates a button inside the window in the specified position.
 * 
 * Returns:
 * 	A new initialized (zeroed) button.
 *----------------------------------------------------------------------
 */
WButton *
wButtonCreate(WCoreWindow *core, int x, int y, int width, int height)
{
    WButton *wbut;
    
    wbut = wmalloc(sizeof(WButton));
    wbut->box = wBoxCreate(core, x, y, width, height, WB_NORMAL);
    wbut->box->texture = core->screen_ptr->unfocused_texture;
    wbut->alt_texture = core->screen_ptr->pushed_button_texture;
    wbut->image = NULL;
    wbut->alt_image = NULL;
    wbut->label = NULL;
    wbut->label_gc = core->screen_ptr->menu_entry_gc;
    wbut->font = core->screen_ptr->menu_entry_font;
    wbut->box->core->descriptor.self = wbut;
    wbut->box->core->descriptor.methods = &wButtonMethods;
    wbut->pushed = 0; 
    return wbut;
}


void
wButtonDestroy(WButton *wbut)
{
    if (wbut->image && !wbut->image->shared)
      wPixmapDestroy(wbut->image);

    if (wbut->alt_image && !wbut->alt_image->shared)
      wPixmapDestroy(wbut->alt_image);
#ifdef NEWSTYLE
    if (wbut->image2 && !wbut->image2->shared)
      wPixmapDestroy(wbut->image2);
    
    if (wbut->alt_image2 && !wbut->alt_image2->shared)
      wPixmapDestroy(wbut->alt_image2);
#endif
    /* textures? */

    wBoxDestroy(wbut->box);
    wfree(wbut);
}


void
wButtonPaint(WButton *wbut, int repaint)
{
    WCoreWindow *core = wbut->box->core;
    GC copy_gc = core->screen_ptr->copy_gc;
    WTexture *old_texture;
    WPixmap *image;
    int x=0, y=0, w=0, d=0;

    /* save normal texture */
    old_texture = wbut->box->texture;
    /* setup stuff according to the state */
    if (wbut->pushed) {
	wbut->box->texture = wbut->alt_texture;
	image = (wbut->alt_image ? wbut->alt_image : wbut->image);
#ifdef NEWSTYLE
	wbut->box->border = WB_NONE;
#endif	
	d=1;
    } else {
	image = wbut->image;
#ifdef NEWSTYLE
	wbut->box->border = WB_TITLEBAR;
#endif
	d=0;
    }

    wBoxPaint(wbut->box, repaint);
    if (wbut->label) {
	w = XTextWidth(wbut->font, wbut->label, strlen(wbut->label));
	if (image) {
	    w += image->width + 4;
	}
    } else {
	if (image) {
	    w = image->width;
	}
    }
    /* display text */
    if (wbut->label) {
	y = wbut->font->ascent + d
	  + (core->height - (wbut->font->ascent + wbut->font->descent))/2;
	XDrawString(dpy, core->window, wbut->label_gc, (core->width-w)/2+d, y,
		    wbut->label, strlen(wbut->label));
    }
    /* display image */
    if (image) {
	XSetClipMask(dpy, copy_gc, image->mask);
	if (wbut->label)
	  x = core->width - (core->width-w)/2 - image->width + d;
	else
	  x = (core->width - w)/2 + d;
	y = (core->height - image->height)/2 + d;
	XSetClipOrigin(dpy, copy_gc, x, y);
	if (image->depth==1)
	  XCopyPlane(dpy, image->pixmap, core->window, copy_gc,
		     0, 0, image->width, image->height, x, y, 1);
	else
	  XCopyArea(dpy, image->pixmap, core->window, copy_gc,
		     0, 0, image->width, image->height, x, y);
    }
    /* restore normal texture */
    wbut->box->texture = old_texture;
}


static void
buttonMouseDown(WObjDescriptor *desc, XButtonEvent *bev)
{
    XEvent event;
    WButton *wbut = desc->self;
    int done=0, execute=1;
    
    if (XGrabPointer(dpy, wbut->box->core->window, False, 
		     ButtonMotionMask|ButtonReleaseMask
		     |EnterWindowMask|LeaveWindowMask,
		     GrabModeAsync, GrabModeAsync, None, 
		     None, CurrentTime)!=GrabSuccess) {
	wWarning("pointer grab failed for menu move");
	return;
    }
    
    wbut->pushed = True;
    wButtonPaint(wbut, False);
    while (!done) {
	XNextEvent(dpy, &event);
	switch (event.type) {
	 case LeaveNotify:
	    wbut->pushed = False;
	    execute = 0;
	    wButtonPaint(wbut, False);
	    break;
	    
	 case EnterNotify:
	    wbut->pushed = True;
	    execute = 1;
	    wButtonPaint(wbut, False);    
	    break;

	 case ButtonRelease:
	    done = 1;
	    break;
	}
    }
    if (wbut->pushed) {
	wbut->pushed = False;	
	wButtonPaint(wbut, False);
    }
    XUngrabPointer(dpy, CurrentTime);
    if (execute && desc->methods->click!=NULL) {
	(*desc->methods->click)(desc, &event);
    }
}
