/*
 *  WindowMaker window manager
 * 
 *  Copyright (c) 1997 Alfredo K. Kojima
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <stdlib.h>
#include <stdio.h>

#include "generic/wwmlib.h"
#include "WindowMaker.h"
#include "core.h"
#include "box.h"
#include "window.h"
#include "properties.h"
#include "config.h"
#include "actions.h"
#include "icon.h"

/****** Global Variables ******/

/* contexts */
extern XContext wWinContext;

extern Atom _XA_WM_STATE;
extern Atom _XA_WM_PROTOCOLS;
extern Atom _XA_WM_NAME;
extern Atom _XA_WM_ICON_NAME;

/*
 *--------------------------------------------------------------------
 * wClientRestore--
 * 	Reparent the window back to the root window.
 * 
 *--------------------------------------------------------------------
 */
void
wClientRestore(WWindow *wwin)
{
    XUnmapWindow(dpy, wwin->client_win);
    XReparentWindow(dpy, wwin->client_win, wwin->screen_ptr->root_win,
		    wwin->frame_x, wwin->frame_y);
    XSetWindowBorderWidth(dpy, wwin->client_win, wwin->old_border_width);
    if (wwin->flags.mapped)
      XMapWindow(dpy, wwin->client_win);
}


/*
 *---------------------------------------------------------------------- 
 * wClientSetState--
 * 	Set the state of the client window to one of the window
 * states defined in ICCCM (Iconic, Withdrawn, Normal)
 * 
 * Side effects:
 * 	The WM_STATE property of the window is updated as well as the
 * WWindow.state variable.
 *---------------------------------------------------------------------- 
 */
void
wClientSetState(WWindow *wwin, int state)
{
    unsigned long data[2];
    
    wwin->state = state;

    data[0] = (unsigned long) state;
    data[1] = (unsigned long) None;/* icon window */

    XChangeProperty(dpy, wwin->client_win, _XA_WM_STATE, _XA_WM_STATE, 32, 
		    PropModeReplace, (unsigned char *) data, 2);
}


void
wClientGetGravityOffsets(WWindow *wwin, int *ofs_x, int *ofs_y)
{
    switch (wwin->normal_hints->win_gravity) {	
     case ForgetGravity:
     case CenterGravity:
     case StaticGravity:
	*ofs_x = 0;
	*ofs_y = 0;
	break;
     case NorthWestGravity:
	*ofs_x = -1;
	*ofs_y = -1;
	break;
     case NorthGravity:
	*ofs_x = 0;
	*ofs_y = -1;
	break;
     case NorthEastGravity:
	*ofs_x = 1;
	*ofs_y = -1;
	break;
     case WestGravity:
	*ofs_x = -1;
	*ofs_y = 0;
	break;
     case EastGravity:
	*ofs_x = 1;
	*ofs_y = 0;
	break;
     case SouthWestGravity:
	*ofs_x = -1;
	*ofs_y = 1;
	break;
     case SouthGravity:
	*ofs_x = 0;
	*ofs_y = 1;
	break;
     case SouthEastGravity:
	*ofs_x = 1;
	*ofs_y = 1;
	break;
    }
}



void
wClientConfigure(WWindow *wwin, XConfigureRequestEvent *xcre)
{
    XWindowChanges xwc;
    int nx, ny, nwidth, nheight;
    int ofs_x, ofs_y;

    if (wwin==NULL) {
	/* 
	 * configure a window that was not mapped by us 
	 */
	xwc.x = xcre->x;
	xwc.y = xcre->y;
	xwc.width = xcre->width;
	xwc.height = xcre->height;
	xwc.border_width = xcre->border_width;
	xwc.stack_mode = xcre->detail;
	xwc.sibling = xcre->above;	
	XConfigureWindow(dpy, xcre->window, xcre->value_mask, &xwc);
	return;
    }

    if (xcre->value_mask & CWStackMode) {
	WObjDescriptor *desc;
	WWindow *sibling;

	if ((xcre->value_mask & CWSibling) &&
	    (XFindContext(dpy, xcre->above, wWinContext, 
			  (XPointer *)&desc) == XCSUCCESS)
	    && (desc->parent_type==WCLASS_WINDOW)) {
	    sibling=desc->parent;
	    xwc.sibling = sibling->frame->window;
	} else {
	    xwc.sibling = xcre->above;
	}
	xwc.stack_mode = xcre->detail;
	XConfigureWindow (dpy, wwin->frame->window,
			  xcre->value_mask & (CWSibling | CWStackMode), &xwc);
    }

    wClientGetGravityOffsets(wwin, &ofs_x, &ofs_y);
    
    if (xcre->value_mask & CWBorderWidth) {
	wwin->old_border_width = xcre->border_width;
    }

    
    if (xcre->value_mask & CWX)
      nx = xcre->x - FRAME_BORDER_WIDTH;
    else
      nx = wwin->frame_x;
    
    if (xcre->value_mask & CWY)
      ny = xcre->y - FRAME_BORDER_WIDTH - ((ofs_y < 0) ? 0 : wwin->top_width);
    else
      ny = wwin->frame_y;

    if (xcre->value_mask & CWWidth)
      nwidth = xcre->width;
    else
      nwidth = wwin->frame->width;
    
    if (xcre->value_mask & CWHeight) 
      nheight = xcre->height + wwin->top_width + wwin->bottom_width;
    else
      nheight = wwin->frame->height;

    wWindowConfigure(wwin, nx, ny, nwidth, nheight);
}


void
wClientSendProtocol(WWindow *wwin, Atom protocol, Time time)
{
    XEvent event;

    event.xclient.type = ClientMessage;
    event.xclient.message_type = _XA_WM_PROTOCOLS;
    event.xclient.format = 32;
    event.xclient.display = dpy;
    event.xclient.window = wwin->client_win;
    event.xclient.data.l[0] = protocol;
    event.xclient.data.l[1] = time;

    XSendEvent(dpy, wwin->client_win, False, NoEventMask, &event);
    XSync(dpy, 0);
}


void
wClientKill(WWindow *wwin)
{
    XKillClient(dpy, wwin->client_win);
    XSync(dpy, 0);
}

/*
 *---------------------------------------------------------------------- 
 * wClientCheckProperty--
 * 	Handles PropertyNotify'es, verifying which property was
 * changed and updating internal state according to that, like redrawing
 * the icon title when it is changed.
 * 
 * Side effects:
 * 	Depends on the changed property.
 *---------------------------------------------------------------------- 
 */
void
wClientCheckProperty(WWindow *wwin, XPropertyEvent *event) 
{
    if (event->atom==_XA_WM_NAME) {
	/* window title was changed */
	
	if (wwin->title!=NULL)
	  XFree(wwin->title);
	if (!XFetchName(dpy, wwin->client_win, &(wwin->title))) {
	    wwin->title = NULL;
	    wwin->title_width = XTextWidth(wwin->title_font, DEF_WINDOW_TITLE,
					   strlen(DEF_WINDOW_TITLE));
	} else
	  wwin->title_width = XTextWidth(wwin->title_font, wwin->title,
					 strlen(wwin->title));
	wWindowPaint(wwin, BORDER_TOP, False);
    } else if (event->atom==_XA_WM_ICON_NAME && wwin->icon) {
	/* icon title was changed */
	
	if (wwin->icon->icon_name)
	  XFree(wwin->icon->icon_name);

	if (!XGetIconName(dpy, wwin->client_win, &wwin->icon->icon_name)) {
	    wwin->icon->icon_name = NULL;
	}
	wIconPaint(wwin->icon);
    }
}
