/*
 *  WindowMaker window manager
 * 
 *  Copyright (c) 1997 Alfredo K. Kojima
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>


#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include "generic/wwmlib.h"
#include "WindowMaker.h"
#include "window.h"
#include "actions.h"
#include "client.h"
#include "icon.h"
#include "funcs.h"
#include "dialog.h"


/******** Global Variables **********/
extern XContext wWinContext;

extern Cursor wCursor[WCUR_LAST];

extern WScreen wScreen;

extern Time LastTimestamp;



static void handleMapNotify(), handleUnmapNotify();
static void handleButtonPress(), handleExpose();
static void handleDestroyNotify();
static void handleConfigureRequest();
static void handleMapRequest();
static void handlePropertyNotify();

static void
saveTimestamp(XEvent *event)
{
    LastTimestamp = CurrentTime;

    switch (event->type) {
     case ButtonRelease:
     case ButtonPress:
	LastTimestamp = event->xbutton.time;
	break;
     case KeyPress:
     case KeyRelease:
	LastTimestamp = event->xkey.time;
	break;
     case MotionNotify:
	LastTimestamp = event->xmotion.time;
	break;
     case PropertyNotify:
	LastTimestamp = event->xproperty.time;
	break;
     case EnterNotify:
     case LeaveNotify:
	LastTimestamp = event->xcrossing.time;
	break;
     case SelectionClear:
	LastTimestamp = event->xselectionclear.time;
	break;
     case SelectionRequest:
	LastTimestamp = event->xselectionrequest.time;
	break;
     case SelectionNotify:
	LastTimestamp = event->xselection.time;
	break;
    }
}

void
DispatchEvent(XEvent *event)
{
    saveTimestamp(event);    
    switch (event->type) {
     case MapRequest:
	handleMapRequest(event->xmaprequest.window);
	break;
	
     case ConfigureRequest:
	handleConfigureRequest(event);
	break;
	
     case DestroyNotify:
	handleDestroyNotify(event->xdestroywindow.window);
	break;
	
     case MapNotify:
	handleMapNotify(event->xmap.window);
	break;
	
     case UnmapNotify:
	handleUnmapNotify(event);
	break;
	
     case ButtonPress:
	handleButtonPress(event);
	break;
	
     case Expose:
	handleExpose(event);
	break;
	
     case PropertyNotify:
	handlePropertyNotify(event);
	break;
    }
}

/*
 *----------------------------------------------------------------------
 * EventLoop-
 * 	Processes X and internal events indefinitely.
 * 
 * Returns:
 * 	None
 * 
 * Side effects:
 * 	The LastTimestamp global variable is updated.
 *---------------------------------------------------------------------- 
 */
void
EventLoop()
{
    XEvent event;
    
    for(;;) {
	XNextEvent(dpy, &event);
	DispatchEvent(&event);
    }    
}



static void 
handleMapRequest(Window window)
{
    WWindow *wwin;
    
#ifdef DEBUG
    puts("got map request for");
#endif
    if ((wwin=wFindWWindow(window))) {
	/* deiconify window */
	if (wwin->flags.shaded) {
	    wUnshadeWindow(wwin);
	} else if (wwin->flags.miniaturized) {
	    wDeiconifyWindow(wwin);
	}
	return;
    }
    wwin = wManageWindow(window);
    if (wwin!=NULL)
      wWindowPaint(wwin, BORDER_ALL, False);
}

static void
handleDestroyNotify(Window window)
{
    WWindow *wwin;
#ifdef DEBUG
    puts("got destroy notify");
#endif	    
    
    wwin = wFindWWindow(window);
    if (wwin)
      wUnmanageWindow(wwin, False);
}

static void 
handleExpose(XEvent *event)
{
    WObjDescriptor *desc;
    WWindow *wwin;

#ifdef DEBUG    
    puts("got expose");
#endif
    if (event->xexpose.count!=0) {
      return;
    }
    

    if (XFindContext(dpy, event->xexpose.window, wWinContext, 
		     (XPointer *)&desc)==XCNOENT) {
	return;
    }
    /* Do everything here, instead of in a event handler,
     * because it's simple and faster this way.
     */
    switch (desc->parent_type) {
     case WCLASS_WINDOW:
	wwin = desc->parent;	
	/* 
	 * ignore expose requests to subwindows
	 */
	if (wwin->window_flags.titlebar 
	    && event->xexpose.window==wwin->titlebar->core->window)
	  wWindowPaint(wwin, BORDER_TOP, True);
	else if (wwin->window_flags.resizable
		 && event->xexpose.window==wwin->handle->core->window)
	  wWindowPaint(wwin, BORDER_BOTTOM, True);
	break;

     case WCLASS_MENU:
	wMenuPaint(desc->parent);
	break;
	
     case WCLASS_ICON:
	wIconPaint(desc->parent);
	break;
	
     case WCLASS_DIALOG:
	wDialogPaint(desc->parent);
	break;
    }
}


/* bindable */
static void
handleButtonPress(XEvent *event)
{
    WObjDescriptor *desc;

#ifdef DEBUG    
    puts("got button press");
#endif

    if (event->xbutton.window==wScreen.root_win) {
	switch (event->xbutton.button) {
	 case Button3:
	    OpenRootMenu(&wScreen, event->xbutton.x_root, 
			 event->xbutton.y_root);
	    break;
	 case Button1:
	    wSelectWindows(&wScreen, event);
	    break;
	 default:
#ifdef DEBUG    
	    if (event->xbutton.window==wScreen.root_win)
	      pusage();
#endif
	}
	return;
    }
    if (XFindContext(dpy, event->xbutton.window, wWinContext, 
		 (XPointer *)&desc)==XCNOENT) {
	return;
    }
    
    /* detect double click */
    if ((desc->click_timestamp>0) 
	&& (LastTimestamp - desc->click_timestamp <= DEF_DBLCLICK_TIME)) {
	
	desc->click_timestamp = LastTimestamp;
	
	if (desc->methods && desc->methods->dbl_click!=NULL) {
	    (*desc->methods->dbl_click)(desc, event);
	    return;
	}
    }
    desc->click_timestamp = LastTimestamp;

    /* mouse down. single click is called by the mouse_down handler */
    if (desc->methods && desc->methods->mouse_down!=NULL) {
	(*desc->methods->mouse_down)(desc, event);
	return;
    }
}


static void
handleMapNotify(Window window)
{
    WWindow *wwin;
    
#ifdef DEBUG
    puts("got map");
#endif
    wwin= wFindWWindow(window);
    if (wwin) {
	XGrabServer(dpy);
	XMapSubwindows(dpy, wwin->frame->window);
	
	XMapWindow(dpy, wwin->frame->window);

	wwin->flags.mapped=1;
	wClientSetState(wwin, NormalState);

	XUngrabServer(dpy);
	XSync(dpy,0);
    }
}


static void
handleUnmapNotify(XEvent *event)
{
    WWindow *wwin;
    XEvent junk;
    
#ifdef DEBUG
    puts("got unmap");
#endif

    wwin = wFindWWindow(event->xunmap.window);
    if (!wwin)
      return;

    if (event->xany.window==wScreen.root_win) {
	/* client wants to go to Withdrawn state */
#ifdef DEBUG
	puts("withdraw");
#endif	
	if (wwin->flags.mapped)
	 wUnmanageWindow(wwin, True);
	 
	wwin->flags.mapped=0;

	wClientSetState(wwin, WithdrawnState);
	return;
    }
    XGrabServer(dpy);
    /* check if the window was destroyed */
    if (XCheckTypedWindowEvent(dpy, event->xunmap.window, DestroyNotify,
			       &junk)) {
	wUnmanageWindow(wwin, False);
	XUngrabServer(dpy);
	return;
    }
    wwin->flags.mapped=0;
    wClientSetState(wwin, WithdrawnState);
    wUnmanageWindow(wwin, True);
    XUngrabServer(dpy);
}


static void
handleConfigureRequest(XEvent *event)
{
    WWindow *wwin;
    
#ifdef DEBUG
    puts("got configure request");
#endif
    if (!(wwin=wFindWWindow(event->xconfigurerequest.window))) {
	/*
	 * Configure request for unmapped window
	 */
	wClientConfigure(NULL, &(event->xconfigurerequest));
    } else {
	wClientConfigure(wwin, &(event->xconfigurerequest));
    }
}


static void
handlePropertyNotify(XEvent *event)
{
    WWindow *wwin;
    
#ifdef DEBUG
    puts("got property notify");
#endif
    if ((wwin=wFindWWindow(event->xproperty.window))) {
	wClientCheckProperty(wwin, &event->xproperty);
    }    
}
