/*
 *  WindowMaker window manager
 * 
 *  Copyright (c) 1997 Alfredo K. Kojima
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <X11/Xlib.h>
#include <stdlib.h>
#include <stdio.h>

#include "generic/wwmlib.h"
#include "config.h"
#include "WindowMaker.h"
#include "core.h"
#include "pixmap.h"
#include "texture.h"
#include "window.h"
#include "icon.h"
#include "actions.h"
#include "funcs.h"

/**** Global variables ****/
extern char *PixmapPath;

static void iconMouseDown(WObjDescriptor *desc, XEvent *event);
static void iconDblClick(WObjDescriptor *desc, XEvent *event);

WMethodTable wIconMethods = {
    (WEventHandler)iconMouseDown,
      NULL,
      (WEventHandler)iconDblClick
};



extern Cursor wCursor[WCUR_LAST];



static void
getSize(Drawable d, unsigned int *w, unsigned int *h, unsigned int *dep)
{
    Window rjunk;
    int xjunk, yjunk;
    unsigned int bjunk;
    
    XGetGeometry(dpy, d, &rjunk, &xjunk, &yjunk, w, h, &bjunk, dep);
}


WIcon*
wIconCreate(WWindow *wwin)
{
    WIcon *icon;
    unsigned int width, height, depth;

    icon = wmalloc(sizeof(WIcon));
    icon->core = wCoreCreateTopLevel(wwin->screen_ptr, wwin->icon_x, 
				     wwin->icon_y, ICON_WIDTH, ICON_HEIGHT);
    icon->core->descriptor.methods = &wIconMethods;
    icon->core->descriptor.parent_type = WCLASS_ICON;
    icon->core->descriptor.parent = icon;
    icon->owner = wwin;
    icon->group_icon = 0;

    icon->icon_win = None;
    icon->client_window = 0;
    icon->pixmap = NULL;
    icon->show_title = 1;
    if (!XGetIconName(dpy, wwin->client_win, &icon->icon_name)) {
	icon->icon_name = NULL;
    }

    /*
     * The icon that will be used is selected in this order:
     * 
     * 1. user specified icon for instance name
     * 2. user specified icon for class name
     * 3. application supplied icon window
     * 4. application supplied icon pixmap
     * 5. user specified icon for application name
     * 6. built-in icon
     * 
     * the icon names will also be displayed if it's available
     * 
     */


    /* check if the client supplies an icon window */
    if (wwin->wm_hints && (wwin->wm_hints->flags & IconWindowHint)) {
	int resize=0;

	icon->icon_win = wwin->wm_hints->icon_window;
	XReparentWindow(dpy, icon->icon_win, icon->core->window,
			ICON_BORDER_WIDTH, ICON_BORDER_WIDTH);
	
	getSize(icon->icon_win, &width, &height, &depth);

	if (width > (ICON_WIDTH-2*ICON_BORDER_WIDTH)) {
	    resize = 1;
	    width = ICON_WIDTH-2*ICON_BORDER_WIDTH;
	}	
	if (height > (ICON_HEIGHT-2*ICON_BORDER_WIDTH)) {
	    resize = 1;
	    height = ICON_HEIGHT-2*ICON_BORDER_WIDTH;
	}
	XSetWindowBorderWidth(dpy, icon->icon_win, 0);
	if (resize)
	  XResizeWindow(dpy, icon->icon_win, width, height);

	XMapWindow(dpy, icon->icon_win);
    } else if (wwin->wm_hints && (wwin->wm_hints->flags & IconPixmapHint)) {
	/* check for a client supplied pixmap */	
	icon->client_pixmap = 1;
	icon->pixmap = wmalloc(sizeof(WPixmap));
	icon->pixmap->pixmap = wwin->wm_hints->icon_pixmap;
	if (wwin->wm_hints->flags & IconMaskHint)
	  icon->pixmap->mask = wwin->wm_hints->icon_mask;
	else
	  icon->pixmap->mask = None;
	/* we accept icon pixmaps of depth > 1, althought it's not 
	 * ICCCM compliant */
	getSize(icon->pixmap->pixmap, (unsigned*)&icon->pixmap->width, 
		(unsigned*)&icon->pixmap->height, 
		(unsigned*)&icon->pixmap->depth);
	
	if (icon->pixmap->width > MAX_ICON_WIDTH)
	  icon->pixmap->width = MAX_ICON_WIDTH;

	if (icon->pixmap->height > MAX_ICON_HEIGHT)
	  icon->pixmap->height = MAX_ICON_HEIGHT;
    } else {
	if (wwin->icon_file) {
	    /* look for user supplied icon file */
	    char *path;
	    path = find_file(PixmapPath, wwin->icon_file);
	    if (!path)
	      goto use_default_icon;
	    icon->pixmap = wPixmapLoadXPMFile(wwin->screen_ptr, path);
	    free(path);
	} else {
	    use_default_icon:
	    icon->pixmap = wwin->screen_ptr->default_icon;
	}
    }
    
    return icon;
}


void
wIconDestroy(WIcon *icon)
{
    if (icon->client_pixmap) {
	/* do not destroy pixmap data */
	wfree(icon->pixmap);
    } else if (icon->pixmap && !icon->pixmap->shared) {
	wPixmapDestroy(icon->pixmap);
    } else if (icon->client_window) {
	XReparentWindow(dpy, icon->icon_win, icon->core->screen_ptr->root_win,
			0, 0);
    }

    if (icon->icon_name) 
      XFree(icon->icon_name);
    wCoreDestroy(icon->core);
    wfree(icon);
}


void
wIconPaint(WIcon *icon)
{
    WScreen *scr=icon->core->screen_ptr;
    GC gc = scr->icon_title_gc;
    GC copy_gc = scr->copy_gc;
    int x, y;
    int h = scr->icon_title_font->ascent+scr->icon_title_font->descent;

    if (icon->client_window) {
	/* redraw only borders and exit */
	wTexturePaint(icon->core, icon->core->screen_ptr->icon_texture, 
		      WB_ICON, True);
	return;
    }
    wTexturePaint(icon->core, icon->core->screen_ptr->icon_texture, 
		  WB_ICON, False);

    if (!icon->show_title || !icon->icon_name) {
	h = 0;
    }
    /* draw the icon pixmap */
    if (icon->pixmap) {
	WPixmap *image = icon->pixmap;
	
	XSetClipMask(dpy, copy_gc, image->mask);
	x = (icon->core->width - image->width - 4)/2;
	y = h + (icon->core->height - h - image->height - 4)/2;
	XSetClipOrigin(dpy, copy_gc, x, y);
	if (image->depth==1)
	  XCopyPlane(dpy, image->pixmap, icon->core->window, copy_gc,
		     0, 0, image->width, image->height, x, y, 1);
	else
	  XCopyArea(dpy, image->pixmap, icon->core->window, copy_gc,
		     0, 0, image->width, image->height, x, y);
    }

    /* draw the icon title */
    if (icon->show_title && icon->icon_name!=NULL) {
	int l = strlen(icon->icon_name);

	XSetForeground(dpy, gc, scr->black_pixel);
	XFillRectangle(dpy, icon->core->window, gc, 1, 1, 
		       icon->core->width-2, h);
	x = (icon->core->width 
	     - XTextWidth(scr->icon_title_font, icon->icon_name, l))/2;
	XSetForeground(dpy, gc, scr->white_pixel);
	XDrawString(dpy, icon->core->window, gc, x, 
		    1 + scr->icon_title_font->ascent, icon->icon_name, l);
    }
}


static void 
iconDblClick(WObjDescriptor *desc, XEvent *event)
{
    WIcon *icon = desc->parent;
    
    wDeiconifyWindow(icon->owner);
}


static void 
iconMouseDown(WObjDescriptor *desc, XEvent *event)
{
    WIcon *icon = desc->parent;
    WWindow *wwin = icon->owner;
    XEvent ev;
    int x=wwin->icon_x, y=wwin->icon_y;
    int dx=event->xbutton.x_root, dy=event->xbutton.y_root;
    int grabbed=0;

#ifdef DEBUG
    puts("Moving icon");
#endif
    XRaiseWindow(dpy,icon->core->window);
    if (XGrabPointer(dpy, icon->core->window, False, 
		     ButtonMotionMask|ButtonReleaseMask,
		     GrabModeAsync, GrabModeAsync, None, 
		     wCursor[WCUR_MOVE], CurrentTime) !=GrabSuccess) {
	wWarning("pointer grab failed for icon move");
    }
    while(1) {
	XNextEvent(dpy, &ev);
	switch (ev.type) {
	 case Expose:
	    wIconPaint(icon);
	    break;

	 case MotionNotify:
	    if (!grabbed) {
		XChangeActivePointerGrab(dpy, ButtonMotionMask
					 |ButtonReleaseMask, 
					 wCursor[WCUR_MOVE], CurrentTime);
		grabbed=1;
	    }
	    x += ev.xmotion.x_root - dx;
	    y += ev.xmotion.y_root - dy;
	    dx = ev.xmotion.x_root;
	    dy = ev.xmotion.y_root;
	    XMoveWindow(dpy, icon->core->window, x, y);
	    break;

	 case ButtonRelease:
	    wwin->icon_x = x;
	    wwin->icon_y = y;
#ifdef DEBUG
	    puts("End icon move");
#endif
	    XUngrabPointer(dpy, CurrentTime);
	    return;
	 default:
	    DispatchEvent(&ev);
	}
    }
}
    
