/*
 *  WindowMaker window manager
 * 
 *  Copyright (c) 1997 Alfredo K. Kojima
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "generic/wwmlib.h"
#include "WindowMaker.h"
#include "core.h"
#include "box.h"
#include "button.h"
#include "menu.h"
#include "config.h"
#include "actions.h"
#include "funcs.h"

/****** Global Variables ******/

extern Cursor wCursor[WCUR_LAST];

extern XContext wWinContext;



static void menuMouseDown(WObjDescriptor *desc, XButtonEvent *bev);
static void menuTitleMouseDown(WObjDescriptor *desc, XButtonEvent *bev);
static void menuCloseClick(WObjDescriptor *desc, XEvent *bev);

static WMethodTable wMenuMethods = {
    (WEventHandler)menuMouseDown,
      NULL,
      NULL
};

static WMethodTable wMenuTitlebarMethods = {
    (WEventHandler)menuTitleMouseDown,
      NULL,
      NULL
};


static WMethodTable wMenuButtonMethods;

void
wMenuInitialize()
{
    memcpy(&wMenuButtonMethods, &wButtonMethods, sizeof(WMethodTable));
    wMenuButtonMethods.click = menuCloseClick;
}


WMenu *
wMenuCreate(WScreen *screen, char *title)
{
    WMenu *menu;
    WButton *wbut;
    int title_h;
    int width, height;
    int bsize;
    
    menu = wmalloc(sizeof(WMenu));

    menu->parent = NULL;
    memset(&(menu->flags), 0, sizeof(menu->flags));
    
    menu->title = strdup(title);
    menu->title_width = XTextWidth(screen->menu_title_font, title, 
				   strlen(title));
    menu->entry_no = 0;
    menu->alloced_entries = 8;
    menu->selected_entry = -1;
    menu->entries = wmalloc(sizeof(WMenuEntry)*menu->alloced_entries);
    
    title_h = screen->menu_title_font->ascent 
	+ screen->menu_title_font->descent + 8;
    
    menu->screen_ptr = screen;
    width = XTextWidth(screen->title_font, title, strlen(title)) 
	+ title_h + 6;
    height = title_h*2;
    menu->frame = wCoreCreateTopLevel(screen, 8, 2, width, height);
    menu->frame_x = screen->app_menu_x;
    menu->frame_x = screen->app_menu_y;
    menu->frame->descriptor.parent = menu;
    menu->frame->descriptor.parent_type = WCLASS_MENU;
    menu->frame->descriptor.methods = &wMenuMethods;
    
    /* create borders */ 
    menu->titlebar = wBoxCreate(menu->frame, 0, 0, menu->frame->width,
				title_h, WB_NORMAL);
    menu->titlebar->texture = screen->focused_texture;
    menu->titlebar->core->descriptor.parent = menu;
    menu->titlebar->core->descriptor.parent_type = WCLASS_MENU;
    menu->titlebar->core->descriptor.methods = &wMenuTitlebarMethods;

    menu->menu = wCoreCreate(menu->frame, 0, title_h, width, height-title_h);

    menu->menu->descriptor.parent = menu;
    menu->menu->descriptor.parent_type = WCLASS_MENU;
    menu->menu->descriptor.methods = &wMenuMethods;
    
    /* create close button */
    bsize = (title_h - 7);
    wbut = wButtonCreate(menu->titlebar->core, width - 3 - bsize, 
			 (title_h-bsize)/2, bsize, bsize);
    /* setup object descriptor */
    wbut->box->core->descriptor.parent = menu;
    wbut->box->core->descriptor.parent_type = WCLASS_MENU;
    wbut->box->core->descriptor.methods = &wMenuButtonMethods;
    wbut->image = screen->b_pixmaps[WBUT_CLOSE];
    menu->close_button = wbut;
    
    XMapSubwindows(dpy, menu->frame->window);
    
    XSync(dpy, 0);
    
    return menu;
}


void
wMenuAddItem(WMenu *menu, char *text, int code)
{
    WMenuEntry *entry;
  
#ifdef DEBUG
    if (!menu) {
	printf("Passed NULL as menu parameter to wMenuAddItem() \n");
	return;
    }    
#endif
    menu->flags.realized = 0;
    /* wreallocate array if it's too small */
    if (menu->entry_no == menu->alloced_entries) {
	void *tmp;
#ifdef DEBUG
	puts("doing wrealloc()");
#endif
	tmp = wrealloc(menu->entries,
		      sizeof(WMenuEntry)*menu->alloced_entries+4);
	if (tmp==NULL) {
	    wWarning("wrealloc() failed while trying to add menu item");
	    return;
	}
	menu->entries = tmp;
	menu->alloced_entries += 4;
    }
    entry = wmalloc(sizeof(WMenuEntry));
    entry->flags.selected = 0;
    entry->flags.enabled = 1;
    entry->text = strdup(text);
    entry->rtext = NULL;
    entry->cascade = NULL;
    entry->code = code;
    entry->callback = NULL;
    entry->order = menu->entry_no;
    menu->entries[menu->entry_no] = entry;
    menu->entry_no++;
}

void wMenuAddCallback(WMenu *menu, char *text, 
		      void (*callback)(WMenu *menu, WMenuEntry *entry), 
		      int code)
{
    WMenuEntry *entry;

#ifdef DEBUG
    if (!menu) {
	printf("Passed NULL as menu parameter to wMenuAddCallback() \n");
	return;
    }    
#endif
    
    menu->flags.realized = 0;
    /* wreallocate array if it's too small */
    if (menu->entry_no == menu->alloced_entries) {
	void *tmp;
#ifdef DEBUG
	puts("doing wrealloc()");
#endif	
	tmp = wrealloc(menu->entries,
		      sizeof(WMenuEntry)*menu->alloced_entries+4);
	if (tmp==NULL) {
	    wWarning("wrealloc() failed while trying to add menu item");
	    return;
	}
	menu->entries = tmp;
	menu->alloced_entries += 4;
    }
    entry = wmalloc(sizeof(WMenuEntry));
    entry->flags.selected = 0;
    entry->flags.enabled = 1;    
    entry->text = strdup(text);
    entry->rtext = NULL;
    entry->cascade = NULL;
    entry->code = code;
    entry->callback = callback;
    entry->order = menu->entry_no;    
    menu->entries[menu->entry_no] = entry;
    menu->entry_no++;
}


void 
wMenuAddCascade(WMenu *menu, char *text, WMenu *cascade)
{
    int n;
    WMenuEntry *entry;

#ifdef DEBUG
    if (!menu) {
	printf("Passed NULL as menu parameter to wMenuAddCascade() \n");
	return;
    }    
#endif
    
    cascade->parent = menu;
    menu->flags.realized = 0;
    /* wreallocate array if it's too small */
    if (menu->entry_no == menu->alloced_entries) {
	void *tmp;
#ifdef DEBUG
	puts("doing wrealloc()");
#endif	
	tmp = wrealloc(menu->entries, 
		      sizeof(WMenuEntry)*menu->alloced_entries+4);
	if (tmp==NULL) {
	    wWarning("wrealloc() failed while trying to add menu item");
	    return;
	}
	menu->entries = tmp;
	menu->alloced_entries += 4;
    }
    n = menu->entry_no++;
    
    entry = wmalloc(sizeof(WMenuEntry));
    entry->flags.selected = 0;
    entry->flags.enabled = 1;
    entry->text = strdup(text);
    entry->rtext = NULL;
    entry->cascade = cascade;
    entry->code = 0;
    entry->callback = NULL;
    entry->order = n;
    menu->entries[n] = entry;
}


void 
wMenuRealize(WMenu *menu)
{
    int i;
    int width, mwidth, extra_width, bsize;
    int theight, eheight;
    WScreen *scr = menu->frame->screen_ptr;

#ifdef DEBUG
    if (!menu) {
	printf("Trying to realize NULL menu \n");
	return;
    }    
#endif
    menu->flags.realized = 1;
    mwidth = menu->title_width;
    theight = scr->menu_title_font->ascent + scr->menu_title_font->descent + 8;
    eheight = scr->menu_entry_font->ascent + scr->menu_entry_font->descent + 6;
    menu->entry_height = eheight;
    
    extra_width = XTextWidth(scr->menu_entry_font, "F8", 2) + 15;
    
    for (i=0; i<menu->entry_no; i++) {
	char *text;

	text = menu->entries[i]->text;
	width = XTextWidth(menu->frame->screen_ptr->menu_entry_font,
			   text, strlen(text));
	menu->entries[i]->text_width = width;
	if (width > mwidth)
	  mwidth = width;
    }
    wCoreConfigure(menu->frame, menu->frame_x, menu->frame_y, 
		   mwidth + extra_width, theight + menu->entry_no * eheight);
    wCoreConfigure(menu->titlebar->core, 0, 0, mwidth + extra_width, 0);
    wCoreConfigure(menu->menu, 0, theight, mwidth + extra_width, 
		   menu->entry_no * eheight);
    
    bsize = menu->close_button->box->core->width;
    XMoveWindow(dpy,menu->close_button->box->core->window, mwidth 
		+ extra_width - 3 - bsize, (theight-bsize)/2);
    XSetWindowBackground(dpy, menu->menu->window, 
			 scr->unfocused_texture->solid.color.pixel);
}


void
wMenuDestroy(WMenu *menu)
{
    int i;

#ifdef DEBUG
    if (!menu) {
	printf("Trying to destroy NULL menu \n");
	return;
    }
#endif
    /* destroy items */
    for (i=0; i<menu->entry_no; i++) {
	wfree(menu->entries[i]->text);
	if (menu->entries[i]->rtext)
	    wfree(menu->entries[i]->rtext);
	if (menu->entries[i]->cascade!=NULL)
	  wMenuDestroy(menu->entries[i]->cascade);
	wfree(menu->entries[i]);
    }
    if (menu->entries)
	wfree(menu->entries);
    if (menu->title)
	wfree(menu->title);
    wButtonDestroy(menu->close_button);
    wBoxDestroy(menu->titlebar);
    wCoreDestroy(menu->menu);
    wCoreDestroy(menu->frame);
    wfree(menu);
}


static void
paintEntry(WMenu *menu, int index, int selected)
{
    int y, w, h;
    GC light, dim, dark;
    XSegment segs[2];
    WScreen *scr=menu->frame->screen_ptr;
    Window win = menu->menu->window;
    WMenuEntry *entry=menu->entries[index];

    /* FIX textures */
    h = menu->entry_height;
    w = menu->menu->width;
    y = index * h;
    /* paint background */
    if (selected) {
	XFillRectangle(dpy, win, scr->select_menu_gc, 1, y+1, w-3, h-3);
    } else {
	switch (scr->unfocused_texture->type) {
	 case WTEX_SOLID:
	    XFillRectangle(dpy, win, scr->unfocused_texture->solid.gc, 
			   1, y+1, w-3, h-3);
	    break;
	}
    }
    /* draw text */
    if (entry->flags.enabled) {
	XDrawString(dpy, win, scr->menu_entry_gc, 5, 
		    2+y+scr->menu_entry_font->ascent, entry->text, 
		    strlen(entry->text));
    } else {
	XDrawString(dpy, win, scr->disabled_menu_entry_gc, 5, 
		    2+y+scr->menu_entry_font->ascent, entry->text, 
		    strlen(entry->text));
    }
    light = scr->unfocused_texture->any.relief->light_gc;
    dim = scr->unfocused_texture->any.relief->dim_gc;
    dark = scr->unfocused_texture->any.relief->dark_gc;

    if (entry->cascade) {
	/* draw the cascade indicator */
	XDrawLine(dpy,win,dim, w-11, y+6, w-6, y+h/2-1);
	XDrawLine(dpy,win,light, w-11, y+h-8, w-6, y+h/2-1);
	XDrawLine(dpy,win,dark, w-12, y+6, w-12, y+h-8);
    }

    /* draw the frame */
    segs[0].x1 = 0;
    segs[0].y1 = y;
    segs[0].x2 = w-2;
    segs[0].y2 = y;
    segs[1].x1 = 0;
    segs[1].y1 = y;
    segs[1].x2 = 0;
    segs[1].y2 = y + h - 2;
    XDrawSegments(dpy, win, light, segs, 2);

    segs[0].x1 = 0;
    segs[0].x2 = w-1;
    segs[0].y1 = segs[0].y2 = y + h-1;
    segs[1].x1 = segs[1].x2 = w-1;
    segs[1].y1 = y;
    segs[1].y2 = y + h-1;
    XDrawSegments(dpy, win, dark, segs, 2);

    segs[0].x1 = 1;
    segs[0].y1 = segs[0].y2 = y + h-2;
    segs[0].x2 = w-2;
    segs[1].x1 = segs[1].x2 = w-2;
    segs[1].y1 = y + 1;
    segs[1].y2 = y + h-2;
    XDrawSegments(dpy, win, dim, segs, 2);

}

void
wMenuMapAt(WMenu *menu, int x, int y)
{
    if (!menu->flags.realized) {
	menu->flags.realized=1;
	wMenuRealize(menu);
    }
    if (x<0) x = 0;
    if (y<0) y = 0;
    XMoveWindow(dpy, menu->frame->window, x, y);
    menu->frame_x = x;
    menu->frame_y = y;
    XMapRaised(dpy, menu->frame->window);
    menu->flags.mapped = 1;
}


void
wMenuMap(WMenu *menu)
{
    if (!menu->flags.realized) {
	menu->flags.realized=1;
	wMenuRealize(menu);
    }
    if (menu->flags.app_menu) {
	menu->frame_x = menu->frame->screen_ptr->app_menu_x;
	menu->frame_y = menu->frame->screen_ptr->app_menu_y;
	XMoveWindow(dpy, menu->frame->window, menu->frame_x, menu->frame_y);
    }
    XMapRaised(dpy, menu->frame->window);
    menu->flags.mapped = 1;
}


void
wMenuUnmap(WMenu *menu)
{
    int i;
    
    XUnmapWindow(dpy, menu->frame->window);
    XUnmapWindow(dpy, menu->close_button->box->core->window);
    menu->flags.buttoned = 0;
    menu->flags.mapped = 0;
    
    for (i=0; i<menu->entry_no; i++) {
	if (menu->entries[i]->cascade!=NULL 
	    && menu->entries[i]->cascade->flags.mapped 
	    && !menu->entries[i]->cascade->flags.buttoned)
	  wMenuUnmap(menu->entries[i]->cascade);
	menu->entries[i]->flags.selected=0;
    }
    menu->selected_entry = -1;
} 


void 
wMenuPaint(WMenu *menu)
{
    int i;
    WScreen *scr = menu->frame->screen_ptr;
    
    /* paint titlebar */
    wBoxPaint(menu->titlebar, False);
    XDrawString(dpy, menu->titlebar->core->window, scr->menu_title_gc, 5,
		scr->menu_title_font->ascent + 3, menu->title, 
		strlen(menu->title));
    /* paint entries */
    for(i=0; i<menu->entry_no; i++) {
	paintEntry(menu, i, i==menu->selected_entry);
    }
    if (menu->flags.buttoned)
      wButtonPaint(menu->close_button, False);
}

void
wMenuEnableEntry(WMenu *menu, int i, int enable)
{
    menu->entries[i]->flags.enabled=enable;
    paintEntry(menu, i, i==menu->selected_entry);
}


/* ====================================================================== */


static void
selectEntry(WMenu *menu, int entry_no)
{
    WMenuEntry *entry;
    
    if (menu->selected_entry!=entry_no) {
	/* unselect previous entry */
	if (menu->selected_entry>=0) {
	    entry = menu->entries[menu->selected_entry];
	    if (!entry->cascade || entry->cascade->flags.buttoned ||
		entry_no>=0) {
		paintEntry(menu, menu->selected_entry, False);
		entry = menu->entries[menu->selected_entry];
		entry->flags.selected=0;
		/* unmap cascade */
		if (entry->cascade && !entry->cascade->flags.buttoned) {
		    wMenuUnmap(entry->cascade);
		}
	    } else return;
	}
	if (entry_no<0) {
	    menu->selected_entry = -1;
	    return;
	}
	entry = menu->entries[entry_no];
	if (entry->flags.enabled) {
	    if (entry->cascade!=NULL)
	      /* map cascade */
	      if (!entry->cascade->flags.mapped) {
		  wMenuMapAt(entry->cascade, menu->frame_x 
			     + menu->frame->width, menu->frame_y);
	      } else return;
	    menu->selected_entry = entry_no;
	    entry->flags.selected=1;
	    paintEntry(menu, entry_no, True);
	} else {
	    menu->selected_entry = -1;
	}
    }    
}

static WMenu*
findMenu(WScreen *scr)
{
    WObjDescriptor *desc;
    Window root_ret, win;
    int x, y, wx, wy;
    unsigned int mask;
    
    XQueryPointer(dpy, scr->root_win, &root_ret, &win, &x, &y, &wx, &wy, 
		  &mask);
    if (win==None) return NULL;
    if (XFindContext(dpy, win, wWinContext, (XPointer *)&desc)==XCNOENT)
      return NULL;
    if (desc->parent_type==WCLASS_MENU)
      return desc->parent;
    return NULL;
}


#define getEntryAt(menu, x, y)   ((y)<0 ? -1 : (y)/(menu->entry_height))

static void 
menuMouseDown(WObjDescriptor *desc, XButtonEvent *bev)
{
    WMenu *menu=desc->parent;
    WMenu *smenu;
    WScreen *scr=menu->frame->screen_ptr;    
    WMenuEntry *entry=NULL;
    XEvent event;
    int done=0;
    int entry_no;

    XRaiseWindow(dpy,menu->frame->window);
    
    if (bev->button==Button3) {
	wMenuUnmap(menu);
	return;
    }
    entry_no = getEntryAt(menu, bev->x, bev->y);
    entry = menu->entries[entry_no];
    if (entry->flags.enabled && entry->cascade!=NULL) {
	/* map cascade */
	if (!entry->cascade->flags.mapped) {
	    selectEntry(menu, entry_no);
	} else if (!entry->cascade->flags.buttoned) {
	    selectEntry(menu, -1);
	}
    } else
      selectEntry(menu, entry_no);
    
    while (!done) {
	XNextEvent(dpy, &event);
	switch (event.type) {
	 case MotionNotify:
	    smenu = findMenu(scr);
	    if (smenu == NULL) {
		if (menu==NULL) break;
		selectEntry(menu, -1);
		menu = smenu;
		break;
	    } else if (menu!=smenu && menu && smenu==menu->parent) {
		selectEntry(menu, -1);
	    }
	    menu = smenu;
	    entry_no = getEntryAt(menu, event.xmotion.x, event.xmotion.y);
	    selectEntry(menu, entry_no);
	    break;

	 case ButtonRelease:
	    done=1;
	    break;

	 default:
	    DispatchEvent(&event);
	}
    }
    if (menu && menu->selected_entry>=0) {
	entry = menu->entries[menu->selected_entry];
	/* erase menu selection */
	if (entry->callback!=NULL) {
	    selectEntry(menu, -1);
	    /* unmap the menu, it's parents and call the callback */
	    if (!menu->flags.buttoned) {
		WMenu *parent=menu;
		while (parent!=NULL && !parent->flags.buttoned) {
		    selectEntry(parent, -1);
		    wMenuUnmap(parent);
		    parent = parent->parent;
		}
		if (parent)
		  selectEntry(parent, -1);
	    }
	    (*entry->callback)(menu, entry);
	}
    }
}


/*
 *---------------------------------------------------------------------- 
 * menuTitleMouseDown--
 * 	Handle ButtonPress events on the menu titlebar.
 * 
 * Side effects:
 * 	The pointer is grabbed until the button is released, but 
 * events are dispatched properly.
 * 
 * Notes:
 * 	Moves the menu according to user action and enables the close
 * button (stick) if the menu is moved.
 *---------------------------------------------------------------------- 
 */
static void 
menuTitleMouseDown(WObjDescriptor *desc, XButtonEvent *bev)
{
    WMenu *menu=desc->parent;    
    XEvent event;
    int x=menu->frame_x, y=menu->frame_y;
    int dx=bev->x_root, dy=bev->y_root;

#ifdef DEBUG
    puts("Moving menu");
#endif
    if (XGrabPointer(dpy, menu->titlebar->core->window, False, 
		     ButtonMotionMask|ButtonReleaseMask,
		     GrabModeAsync, GrabModeAsync, None, 
		     wCursor[WCUR_MOVE], CurrentTime)!=GrabSuccess) {
	wWarning("pointer grab failed for menu move");
	return;
    }
    XRaiseWindow(dpy,menu->frame->window);
    
    /* tear off the menu if it's a root menu or a cascade application menu */
    if (!menu->flags.buttoned && ((menu->flags.app_menu && menu->parent!=NULL)
	|| !menu->flags.app_menu)) {
	
	XMapWindow(dpy, menu->close_button->box->core->window);
	menu->flags.buttoned=1;
	if (menu->parent) {
	    /* turn off selected menu entry in parent menu */
	    selectEntry(menu->parent, -1);
	}
    }
    while(1) {
	XNextEvent(dpy, &event);
	switch (event.type) {
	 case Expose:
	    wMenuPaint(menu);
	    break;

	 case MotionNotify:
	    x += event.xmotion.x_root - dx;
	    y += event.xmotion.y_root - dy;
	    dx = event.xmotion.x_root;
	    dy = event.xmotion.y_root;
	    XMoveWindow(dpy, menu->frame->window, x, y);
	    break;

	 case ButtonRelease:
	    menu->frame_x = x;
	    menu->frame_y = y;
	    if (menu->flags.app_menu) {
		menu->frame->screen_ptr->app_menu_x = x;
		menu->frame->screen_ptr->app_menu_y = y;
	    }
#ifdef DEBUG
	    puts("End menu move");
#endif
	    XUngrabPointer(dpy, CurrentTime);
	    return;
	 default:
	    DispatchEvent(&event);
	}
    }
}

/*
 *----------------------------------------------------------------------
 * menuCloseClick--
 * 	Handles mouse click on the close butoon of menus. The menu is
 * closed when the button is clicked. 
 * 
 * Side effects:
 * 	None
 *---------------------------------------------------------------------- 
 */
static void 
menuCloseClick(WObjDescriptor *desc, XEvent *bev)
{
    WMenu *menu=desc->parent;
    
    wMenuUnmap(menu);
}


