/*
 *  WindowMaker window manager
 * 
 *  Copyright (c) 1997 Alfredo K. Kojima
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/xpm.h>
#include <stdlib.h>

#include "generic/wwmlib.h"
#include "config.h"
#include "WindowMaker.h"
#include "core.h"


/*
 *---------------------------------------------------------------------- 
 * wPixmapCreateFromXPMData--
 * 	Creates a QPixmap structure and initializes it with the supplied
 * XPM structure data.
 * 
 * Returns:
 * 	A WPixmap structure or NULL on failure.
 * 
 * Notes:
 * 	DEF_XPM_CLOSENESS specifies the XpmCloseness
 *---------------------------------------------------------------------- 
 */
WPixmap *
wPixmapCreateFromXPMData(WScreen *scr, char **data)
{
    WPixmap *pix;
    XpmAttributes attribs;
    
    pix = wmalloc(sizeof(WPixmap));
    pix->shared = 0;
    attribs.valuemask = XpmSize|XpmReturnPixels|XpmColormap|XpmCloseness;
    attribs.closeness = DEF_XPM_CLOSENESS;
    attribs.colormap = scr->colormap;
    if (XpmCreatePixmapFromData(dpy, scr->w_win, data, &(pix->pixmap),
			    &(pix->mask), &attribs)!=XpmSuccess) {
	wfree(pix);
	return NULL;
    }
    pix->width = attribs.width;
    pix->height = attribs.height;
    pix->depth = attribs.depth;
    return pix;
}



/*
 *---------------------------------------------------------------------- 
 * wPixmapLoadXPMFile--
 * 	Creates a WPixmap structure and loads a XPM file into it.
 * 
 * Returns:
 * 	A WPixmap structure or NULL on failure.
 * 
 * Notes:
 * 	DEF_XPM_CLOSENESS specifies the XpmCloseness
 *---------------------------------------------------------------------- 
 */
WPixmap *
wPixmapLoadXPMFile(WScreen *scr, char *path)
{
    WPixmap *pix;
    XpmAttributes attribs;
    
    if (!path) return NULL;
    pix = wmalloc(sizeof(WPixmap));
    pix->shared = 0;
    attribs.valuemask = XpmSize|XpmReturnPixels|XpmColormap|XpmCloseness;
    attribs.closeness = DEF_XPM_CLOSENESS;
    attribs.colormap = scr->colormap;
    if (XpmReadFileToPixmap(dpy, scr->w_win, path, &(pix->pixmap),
			    &(pix->mask), &attribs)!=XpmSuccess) {
	wfree(pix);
	return NULL;
    }
    pix->width = attribs.width;
    pix->height = attribs.height;
    pix->depth = attribs.depth;
    return pix;
}


/*
 *---------------------------------------------------------------------- 
 * wPixmapLoadXBMFile--
 * 	Creates a WPixmap structure and loads a XBM file into it with
 * an optional mask file. If a mask is not wanted, mask_path should be
 * NULL.
 * 
 * Returns:
 * 	A WPixmap structure or NULL on failure.
 * 
 * Notes:
 * 	If the mask bitmap is not successfully loaded the operation
 * continues as no mask was supplied.
 *---------------------------------------------------------------------- 
 */
WPixmap *
wPixmapLoadXBMFile(WScreen *scr, char *path, char *mask_path)
{
    WPixmap *pix;
    int junk;

    if (!path) return NULL;
    pix = wmalloc(sizeof(WPixmap));
    
    pix->shared = 0;
    if (XReadBitmapFile(dpy, scr->w_win, path, (unsigned *)&(pix->width), 
			(unsigned *)&(pix->height),
			&(pix->pixmap), &junk, &junk)!=BitmapSuccess) {
	wfree(pix);
	return NULL;
    }
    if (mask_path!=NULL) {
	if (XReadBitmapFile(dpy, scr->w_win, path, (unsigned *)&junk, 
			    (unsigned *)&junk, &(pix->mask), 
			    &junk, &junk) !=BitmapSuccess) {
	    wWarning("could not load mask bitmap file %s. Ignoring..", 
		     mask_path);
	    pix->mask = None;
	}
    }
    pix->depth = 1;
    return pix;
}



/*
 *---------------------------------------------------------------------- 
 * wPixmapDestroy--
 * 	Destroys a WPixmap structure and the pixmap/mask it holds.
 * 
 * Returns:
 * 	None
 *----------------------------------------------------------------------
 */
void 
wPixmapDestroy(WPixmap *pix)
{
    if (pix->mask)
      XFreePixmap(dpy, pix->mask);
    if (pix->pixmap)
      XFreePixmap(dpy, pix->pixmap);
    wfree(pix);
}

