/*
 *  WindowMaker window manager
 * 
 *  Copyright (c) 1997 Alfredo K. Kojima
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>


#include "generic/wwmlib.h"
#include "config.h"

#include "WindowMaker.h"
#include "window.h"
#include "resource.h"


/*
 * Decoration setting is done in this order:
 * - see if there's a default in the resource database
 * - if the application is GNUStep set the resources asked by the app
 * - if not, try to guess the settings
 * 
 * For icons: (ascending precedence)
 * - use global default
 * - use application specified icon bitmap
 * - use application specified icon window
 * - use database with window class
 * - use database with window instance
 * - use database with window name
 */

static HashTable *wResDatabase=NULL;

WResource wResourceDefault;

extern char *PixmapPath;


static char*
cropline(char *line)
{
    char *start, *end;
    
    start = line;
    end = &(line[strlen(line)]);
    while (isspace(*line) && *line!=0) line++;
    while (isspace(*end) && end!=line) end--;
    end--;    
    *end=0;
    return line;
}

static void
parsePreferences(FILE *file)
{
    char buffer[MAXLINE], *line;

    while (!feof(file)) {
	int size;
	
	if (!fgets(buffer, MAXLINE, file)) break;
	line = cropline((char*)buffer);
	if (iscomment(line)) continue;
	if (*line=='[') return;
	size = strcspn(line, "=");
	while (isspace(line[size])) size--;
	if (strncasecmp(line, "PixmapPath", size)==0) {
	    free(PixmapPath);
	    PixmapPath = strdup(&(line[size+1]));
	}
    }
}



static HashTable*
resourceOpen(char *path)
{
    FILE *file;
    char buffer[MAXLINE], *line;
    HashTable *table;
    WResource *res;

    file = fopen(path,"r");
    if (!file) {
	wSysError("could not open configuration file: \"%s\"", path);
	return NULL;
    }

    table = wmalloc(sizeof(HashTable));
    table_init(table);

    while (!feof(file)) {
	char *pos, *end;

	if (!fgets(buffer, MAXLINE, file)) break;
	line = cropline((char*)buffer);
	if (iscomment(line)) continue;

	if (*line=='[') {
	    if (strcasecmp(line, "[Preferences]")==0)
	      parsePreferences(file);
	    continue;
	}

	/* this is a global default */
	if (*line=='*') {
	    res = &wResourceDefault;
	} else {
	    /* application default */
	    pos=strchr(line, '.');
	    if (!pos)
		goto syntax_error;

	    *pos = 0;
	    res = table_sget(table, line);
	    if (!res) {
		/* create new entry */
		res = wmalloc(sizeof(WResource));
		memset(res, 0, sizeof(WResource));
		table_sput(table, line, res);
	    }
	    line = pos;
	}
	line++; /* skip . or * */
	/* global default */
	pos=strchr(line, ':');
	if (!pos) 
	    goto syntax_error;
	*pos=0;
	pos++;

#define ISTRUE(c) ((c)=='1' || (c)=='t' || (c)=='T')
	/* trim parameter */
	while (isspace(*pos)) pos++;
	if (*pos==0)
	    goto syntax_error;
	end=pos+strlen(pos)-1;
	while (isspace(*end)) end--;
	end++;
	*end=0;
	if (strcasecmp(line, "ICON")==0) {
	    res->icon_file = strdup(pos);
	} else if (strcasecmp(line, "TITLEBAR")==0) {
	    res->flags.titlebar=1;
	    res->values.titlebar=ISTRUE(pos[0]);
	} else if (strcasecmp(line, "SHADEABLE")==0) {
	    res->flags.shadeable=1;
	    res->values.shadeable=ISTRUE(pos[0]);
	} else if (strcasecmp(line, "MINIATURIZABLE")==0) {
	    res->flags.miniaturizable=1;
	    res->values.miniaturizable=ISTRUE(pos[0]);
	} else if (strcasecmp(line, "CLOSABLE")==0) {
	    res->flags.closable=1;
	    res->values.closable=ISTRUE(pos[0]);
	} else if (strcasecmp(line, "RESIZABLE")==0) {
	    res->flags.resizable=1;
	    res->values.resizable=ISTRUE(pos[0]);
	} else if (strcasecmp(line, "FOCUSABLE")==0) {
	    res->flags.focusable=1;
	    res->values.focusable=ISTRUE(pos[0]);
	} else if (strcasecmp(line, "BROKENCLOSE")==0) {
	    res->flags.broken_close=1;
	    res->values.broken_close=ISTRUE(pos[0]);
	} else if (strcasecmp(line, "FLOATING")==0) {
	    res->flags.floating=1;
	    res->values.floating=ISTRUE(pos[0]);
	} else {
	    wWarning("unknown option in file %s: \"%s\"", path, line);
	    continue;
	}
	continue;
	syntax_error:
	wWarning("syntax error in file %s: \"%s\"", path, buffer);
    }
    return table;
}



void
wResourceInitialize()
{
    char *path;
    
    path=find_file(DEF_CONFIG_PATHS, DEF_PREFERENCES);

    /* default global defaults */
    wResourceDefault.icon_file = NULL;
    wResourceDefault.flags.titlebar=1;
    wResourceDefault.flags.resizable=1;
    wResourceDefault.flags.closable=1;
    wResourceDefault.flags.broken_close=1;
    wResourceDefault.flags.miniaturizable=1;
    wResourceDefault.flags.shadeable=1;
    wResourceDefault.flags.focusable=1;
    wResourceDefault.flags.sticky=1;
    wResourceDefault.flags.skip_window_list=1;
    wResourceDefault.flags.floating=1;
    wResourceDefault.flags.ignore_keys=1;
    
    wResourceDefault.values.titlebar=1;
    wResourceDefault.values.resizable=1;
    wResourceDefault.values.closable=1;
    wResourceDefault.values.broken_close=1;
    wResourceDefault.values.miniaturizable=1;
    wResourceDefault.values.shadeable=1;
    wResourceDefault.values.focusable=1;
    wResourceDefault.values.sticky=0;
    wResourceDefault.values.skip_window_list=0;
    wResourceDefault.values.floating=0;
    wResourceDefault.values.ignore_keys=0;
    if (path) {
	wResDatabase = resourceOpen(path);
	if (wResourceDefault.icon_file) {
	    char *file;
	    file = wResourceDefault.icon_file;
	    wResourceDefault.icon_file=find_file(PixmapPath, file);
	    free(file);
	}
	free(path);
    } else
      wWarning("could not find a preference file (%s) in \"%s\"", 
	       DEF_PREFERENCES, DEF_CONFIG_PATHS);
}



/*
 *---------------------------------------------------------------------- 
 * wResourceGet--
 * 	Get the attributes set by the user in the resource database 
 * (our own DB, not X's)
 * 
 * Returns:
 * 	The resource structure or
 * 	NULL if it's not defined for the window
 * 
 * Notes:
 * 	Order of precedence is (lower to higher):
 * 		window class
 * 		window instance name
 * 		window name (title)
 *---------------------------------------------------------------------- 
 */
WResource *
wResourceGet(char *name, char *instance, char *wclass)
{
    WResource *res;

    if (!wResDatabase) 
      return NULL;

    if (name) {
	res = table_sget(wResDatabase, name);
	if (res)
	  return res;
    }

    if (instance) {
	res = table_sget(wResDatabase, instance);
	if (res)
	  return res;
    }

    if (wclass) {
	res = table_sget(wResDatabase, wclass);
	if (res)
	  return res;
    }

    return NULL;
}
