/*
 *  WindowMaker window manager
 * 
 *  Copyright (c) 1997 Alfredo K. Kojima
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>


#include "generic/wwmlib.h"
#include "WindowMaker.h"
#include "texture.h"
#include "config.h"
#include "def_pixmaps.h"
#include "screen.h"
#include "pixmap.h"
#include "button.h"
#include "menu.h"
#include "resource.h"

#define EVENT_MASK (LeaveWindowMask|EnterWindowMask|PropertyChangeMask\
		|SubstructureNotifyMask \
		|SubstructureRedirectMask|ButtonPressMask|ButtonReleaseMask\
		|KeyPressMask|KeyReleaseMask)


extern Cursor wCursor[WCUR_LAST];
extern WResource wResourceDefault;



static int 
alreadyRunningError(Display *dpy, XErrorEvent *error)
{
    wFatal("it seems that there already is a window manager running");
    exit(1);
    return -1;
}


static void
allocButtonPixmaps(WScreen *scr)
{
    WPixmap *pix;    
    
    pix = wPixmapCreateFromXPMData(scr, PRED_CLOSE_XPM);
    pix->shared = True;
    scr->b_pixmaps[WBUT_CLOSE] = pix;

    pix = wPixmapCreateFromXPMData(scr, PRED_CLOSE2_XPM);
    pix->shared = True;
    scr->b_pixmaps[WBUT_CLOSE2] = pix;

    pix = wPixmapCreateFromXPMData(scr, PRED_BROKENCLOSE_XPM);
    pix->shared = True;
    scr->b_pixmaps[WBUT_BROKENCLOSE] = pix;

    pix = wPixmapCreateFromXPMData(scr, PRED_BROKENCLOSE2_XPM);
    pix->shared = True;
    scr->b_pixmaps[WBUT_BROKENCLOSE2] = pix;

    pix = wPixmapCreateFromXPMData(scr, PRED_ICONIFY_XPM);
    pix->shared = True;
    scr->b_pixmaps[WBUT_ICONIFY] = pix;
    
    pix = wPixmapCreateFromXPMData(scr, PRED_ICONIFY2_XPM);
    pix->shared = True;
    scr->b_pixmaps[WBUT_ICONIFY2] = pix;
            
#ifdef NEWSTYLE
    pix = wPixmapCreateFromXPMData(scr, PRED_CLOSE_XPM2);
    pix->shared = True;
    scr->b_pixmaps[WBUT_LAST+WBUT_CLOSE] = pix;

    pix = wPixmapCreateFromXPMData(scr, PRED_CLOSE2_XPM2);
    pix->shared = True;
    scr->b_pixmaps[WBUT_LAST+WBUT_CLOSE2] = pix;

    pix = wPixmapCreateFromXPMData(scr, PRED_BROKENCLOSE_XPM2);
    pix->shared = True;
    scr->b_pixmaps[WBUT_LAST+WBUT_BROKENCLOSE] = pix;

    pix = wPixmapCreateFromXPMData(scr, PRED_BROKENCLOSE2_XPM2);
    pix->shared = True;
    scr->b_pixmaps[WBUT_LAST+WBUT_BROKENCLOSE2] = pix;

    pix = wPixmapCreateFromXPMData(scr, PRED_ICONIFY_XPM2);
    pix->shared = True;
    scr->b_pixmaps[WBUT_LAST+WBUT_ICONIFY] = pix;
    
    pix = wPixmapCreateFromXPMData(scr, PRED_ICONIFY2_XPM2);
    pix->shared = True;
    scr->b_pixmaps[WBUT_LAST+WBUT_ICONIFY2] = pix;
#endif    
}



static void
allocResources(WScreen *scr)
{
    XGCValues gcv;
    XColor color;
    WPixmap *pix;
    
    /* fonts */
    scr->title_font = XLoadQueryFont(dpy, DEF_TITLE_FONT);
    scr->menu_title_font = XLoadQueryFont(dpy, DEF_MENU_TITLE_FONT);
    scr->menu_entry_font = XLoadQueryFont(dpy, DEF_MENU_ENTRY_FONT);
    scr->icon_title_font = XLoadQueryFont(dpy, DEF_ICON_TITLE_FONT);
    scr->panel_title_font = XLoadQueryFont(dpy, DEF_PANEL_TITLE_FONT);
    scr->panel_text_font = XLoadQueryFont(dpy, DEF_PANEL_TEXT_FONT);
    scr->info_text_font =  XLoadQueryFont(dpy, DEF_INFO_TEXT_FONT);
      
    /* menu GCs */
    gcv.foreground = scr->white_pixel;
    gcv.font = scr->menu_title_font->fid;
    scr->menu_title_gc = XCreateGC(dpy, scr->w_win, GCForeground|GCFont,
				   &gcv);

    gcv.foreground = scr->black_pixel;
    gcv.font = scr->menu_entry_font->fid;
    scr->menu_entry_gc = XCreateGC(dpy, scr->w_win, GCForeground|GCFont,
				   &gcv);
    color.red = 0x7777;
    color.green = 0x7777;
    color.blue = 0x7777;
    wAllocColor(scr, &color);
    gcv.foreground = color.pixel;
    gcv.font = scr->menu_entry_font->fid;
    scr->disabled_menu_entry_gc = XCreateGC(dpy, scr->w_win, 
					    GCForeground|GCFont, &gcv);
    /* frame GC */
    if (!XParseColor(dpy, scr->colormap, DEF_FRAME_COLOR, &color)) {
	wWarning("could not parse color %s", DEF_FRAME_COLOR);
	color.red=0xffff;
	color.green=0xffff;
	color.blue=0xffff;
    }
    gcv.foreground = ~color.pixel;
    gcv.function = GXxor;
    gcv.line_width = DEF_FRAME_THICKNESS;
    gcv.subwindow_mode = IncludeInferiors;
    scr->frame_gc = XCreateGC(dpy, scr->root_win, GCForeground
			      |GCFunction|GCSubwindowMode|GCLineWidth, 
			      &gcv);
    
    /* line GC */
    gcv.foreground = ~color.pixel;
    gcv.function = GXxor;
    gcv.subwindow_mode = IncludeInferiors;
    gcv.font = scr->info_text_font->fid;
    scr->line_gc = XCreateGC(dpy, scr->root_win, GCForeground|GCFont
			     |GCFunction|GCSubwindowMode, &gcv);
    
    /* copy GC */
    gcv.foreground = scr->white_pixel;
    scr->copy_gc = XCreateGC(dpy, scr->w_win, GCForeground, &gcv);
    

    /* selected menu entry GC */
    gcv.foreground = scr->white_pixel;
    gcv.background = scr->white_pixel;    
    scr->select_menu_gc = XCreateGC(dpy, scr->w_win, 
				    GCForeground|GCBackground, &gcv);
    
    /* default text GCs */
    gcv.foreground = scr->white_pixel;
    gcv.font = scr->title_font->fid;
    scr->focused_text_gc = XCreateGC(dpy, scr->w_win, GCForeground|GCFont,
				     &gcv);

    gcv.foreground = scr->black_pixel;
    scr->unfocused_text_gc = XCreateGC(dpy, scr->w_win, GCForeground|GCFont,
				       &gcv);
    
    /* icon title GC */
    gcv.foreground = scr->white_pixel;
    gcv.font = scr->icon_title_font->fid;
    scr->icon_title_gc = XCreateGC(dpy, scr->w_win, GCForeground|GCFont,
				     &gcv);
    
    /* default shared textures */
    /* focused titlebar */
    if (!XParseColor(dpy, scr->colormap, DEF_FOCUSED_TEXTURE_COLOR, &color)) {
	wWarning("could not parse color %s", DEF_FOCUSED_TEXTURE_COLOR);
	color.red=0;
	color.green=0;
	color.blue=0;
    }
    scr->focused_texture = wTextureCreate(scr);
    wAllocColor(scr, &color);
    wTextureSetSolidColor(scr, scr->focused_texture, &color);

    /* unfocused titlebar */
    if (!XParseColor(dpy, scr->colormap, DEF_UNFOCUSED_TEXTURE_COLOR, &color)){
	wWarning("could not parse color %s", DEF_UNFOCUSED_TEXTURE_COLOR);
	color.red=0xbebe;
	color.green=0xbebe;
	color.blue=0xbebe;
    }
    scr->unfocused_texture = wTextureCreate(scr);
    wAllocColor(scr, &color);
    wTextureSetSolidColor(scr, scr->unfocused_texture, &color);

    /* pushed buttons */
#ifdef NEWSTYLE    
    color.red = 0x7777;
    color.green = 0x7777;
    color.blue = 0x7777;
#else    
    color.red = 0xffff;
    color.green = 0xffff;
    color.blue = 0xffff;
#endif
    scr->pushed_button_texture = wTextureCreate(scr);
    wAllocColor(scr, &color);
    wTextureSetSolidColor(scr, scr->pushed_button_texture, &color);
    scr->pushed_button_texture->any.relief->type = WREL_SUNKEN;

    /* icon texture */
    if (!XParseColor(dpy, scr->colormap, DEF_ICON_TEXTURE_COLOR, &color)){
	wWarning("could not parse color %s", DEF_ICON_TEXTURE_COLOR);
	color.red=0xbebe;
	color.green=0xbebe;
	color.blue=0xbebe;
    }
    scr->icon_texture = wTextureCreate(scr);
    wAllocColor(scr, &color);
    wTextureSetSolidColor(scr, scr->icon_texture, &color); 
    
    /* load pixmaps */
    pix = wPixmapCreateFromXPMData(scr, BUTTON_ARROW_XPM);
    pix->shared = 1;
    scr->button_arrow = pix;

    pix = wPixmapLoadXPMFile(scr, wResourceDefault.icon_file);
    if (pix)
      pix->shared = 1;
    scr->default_icon = pix;

    /* titlebar button pixmaps */
    allocButtonPixmaps(scr);

}


static void
createInternalWindows(WScreen *scr)
{
    int vmask;
    XSetWindowAttributes attribs;

    /* window for displaying geometry information during resizes and moves */
    vmask = CWBackPixmap|CWBackPixel|CWCursor|CWSaveUnder;
    attribs.save_under = True;
    attribs.cursor = wCursor[WCUR_DEFAULT];
    attribs.background_pixmap = None;
    attribs.background_pixel = scr->unfocused_texture->solid.color.pixel;
    scr->geometry_display =
      XCreateWindow(dpy, scr->root_win, (scr->scr_width-80)/2, 
		    (scr->scr_height-20)/2, 80, 20, 1, scr->depth, 
		    CopyFromParent, scr->visual, vmask, &attribs);
}


WScreen *
wScreenInit(int screen_number)
{
    WScreen *scr;
    XIconSize icon_size;
    
    scr = wmalloc(sizeof(WScreen));

    /* initialize globals */
    scr->screen = screen_number;
    scr->root_win = RootWindow(dpy, screen_number);
    scr->visual = DefaultVisual(dpy, screen_number);    
    scr->depth = DefaultDepth(dpy, screen_number);
    /* change the following to user settable */
    scr->w_win = scr->root_win;
    scr->w_visual = scr->visual;
    scr->w_depth = scr->depth;

    scr->black_pixel = BlackPixel(dpy, screen_number);
    scr->white_pixel = WhitePixel(dpy, screen_number);
    
    scr->colormap = DefaultColormap(dpy, screen_number);

    scr->scr_width = WidthOfScreen(ScreenOfDisplay(dpy, screen_number));
    scr->scr_height = HeightOfScreen(ScreenOfDisplay(dpy, screen_number));

    XSetErrorHandler((XErrorHandler)alreadyRunningError);
    XSelectInput(dpy, scr->root_win, EVENT_MASK);
    XSync(dpy, 0);

    /* allocate shared resources */
    allocResources(scr);
    createInternalWindows(scr);
    /* set icon sizes we can accept */
    icon_size.min_width = 1;
    icon_size.min_height = 1;
    icon_size.max_width = MAX_ICON_WIDTH;
    icon_size.max_height = MAX_ICON_HEIGHT;
    icon_size.width_inc = 1;
    icon_size.height_inc = 1;
    XSetIconSizes(dpy, scr->root_win, &icon_size, 1);
    return scr;    
}

