/*
 *  WindowMaker window manager
 * 
 *  Copyright (c) 1997 Alfredo K. Kojima
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include <stdlib.h>
#include <stdio.h>

#include "generic/wwmlib.h"
#include "WindowMaker.h"
#include "core.h"
#include "texture.h"





static void
drawFrame(WCoreWindow *core, WTexture *texture, short border_width);




WTexture *
wTextureCreate(WScreen *scr)
{
    WTexture *texture;
    
    texture = wmalloc(sizeof(WTexture));
    texture->any.relief = NULL;
    return texture;
}


void
wTextureDestroy(WScreen *scr, WTexture *texture)
{
#ifdef DEBUG    
    if (texture==NULL) {
	printf("BUG: trying to free NULL texture\n");
	return;
    }
#endif
    if (texture->any.relief) {
	unsigned long colors[3];

	XFreeGC(dpy, texture->any.relief->light_gc);
	if (texture->any.relief->light_gc)
	  XFreeGC(dpy, texture->any.relief->dark_gc);
	XFreeGC(dpy, texture->any.relief->dim_gc);
	colors[0] = texture->any.relief->light.pixel;
	colors[1] = texture->any.relief->dim.pixel;
	colors[2] = texture->any.relief->dark.pixel;
	XFreeColors(dpy, scr->colormap, colors, 2, 1);
	wfree(texture->any.relief);
    }
    switch (texture->type) {
     case WTEX_SOLID:
	XFreeGC(dpy, texture->solid.gc);
	break;
    }
    wfree(texture);
}



void
wTextureSetSolidColor(WScreen *scr, WTexture *texture, XColor *color)
{
    XColor light, dim, dark;
    int gcm;
    XGCValues gcv;
    int r, g, b;

    texture->type = WTEX_SOLID;
    texture->solid.color = *color;
    if (color->red==0 && color->blue==0 && color->green == 0) {	
	light.red = 0xb6da;
	light.green = 0xb6da;
	light.blue = 0xb6da;
	dim.red = 0x6185;
	dim.green = 0x6185;
	dim.blue = 0x6185;
	dark.red = 0;
	dark.green = 0;
	dark.blue = 0;	
    } else {
 	r = (color->red * 14)/10;
	g = (color->green * 14)/10;
	b = (color->blue * 14)/10;
	light.red = (r > 0xffff ? 0xffff : r);
	light.green = (g > 0xffff ? 0xffff : g);
	light.blue = (b > 0xffff ? 0xffff : b);
 	r = (color->red * 5)/10;
	g = (color->green * 5)/10;
	b = (color->blue * 5)/10;	
	dim.red = r;
	dim.green = g;
	dim.blue = b;
	/*
 	r = (color->red)/10;
	g = (color->green)/10;
	b = (color->blue)/10;	
	dark.red = r;
	dark.green = g;
	dark.blue = b;
	 */
	dark.red = 0;
	dark.green = 0;
	dark.blue = 0;
    }
    if (texture->solid.relief!=NULL) {
	XSetForeground(dpy, texture->solid.relief->light_gc, light.pixel);
	XSetForeground(dpy, texture->solid.relief->dim_gc, dim.pixel);
	XSetForeground(dpy, texture->solid.relief->dark_gc, dark.pixel);

	XSetForeground(dpy, texture->solid.gc, color->pixel);
    } else {
	texture->solid.relief = wmalloc(sizeof(W3DRelief));
	texture->solid.relief->type = WREL_RAISED;
	wAllocColor(scr,&light);
	wAllocColor(scr,&dim);
	wAllocColor(scr,&dark);
	texture->solid.relief->light = light;
	texture->solid.relief->dim = dim;
	texture->solid.relief->dark = dark;

	gcm = GCForeground;
	gcv.foreground = light.pixel;
	texture->solid.relief->light_gc = XCreateGC(dpy, scr->w_win,gcm, &gcv);
	gcv.foreground = dim.pixel;	
	texture->solid.relief->dim_gc = XCreateGC(dpy, scr->w_win, gcm, &gcv);
	gcv.foreground = dark.pixel;
	texture->solid.relief->dark_gc = XCreateGC(dpy, scr->w_win, gcm, &gcv);

	gcv.foreground = color->pixel;
	texture->solid.gc = XCreateGC(dpy, scr->w_win, gcm, &gcv);
    }
}



void
wTexturePaint(WCoreWindow *core, WTexture *texture, int border, 
	      int repaint) 
{
    int d;

    switch (border) {
     case WB_TITLEBAR:
	d = 1;
	break;
     case WB_ICON:
     case WB_NORMAL:
	d = 1;
	break;
     default:
	d = 0;
    }
    switch (texture->type) {
     case WTEX_SOLID:
	if (!repaint) {
	    XSetWindowBackground(dpy, core->window, 
				 texture->solid.color.pixel);
	    XClearArea(dpy, core->window, d, d, core->width-2*d,
		       core->height-2*d, False);
	}
	drawFrame(core, texture, border);
	break;
    }
}


static void
drawFrame(WCoreWindow *core, WTexture *texture, short border)
{
    GC light, dim, dark;
    XSegment segs[2];

    if (border==WB_NONE) return;
    light = texture->any.relief->light_gc;
    dim = texture->any.relief->dim_gc;
    dark = texture->any.relief->dark_gc;
    switch (texture->any.relief->type) {
     case WREL_FLAT: 
	return;
     case WREL_RAISED:
	switch (border) {
	 case WB_ICON:
	 case WB_NORMAL:
	    segs[0].x1 = 1;
	    segs[0].x2 = core->width - 2;
	    segs[0].y2 = segs[0].y1 = core->height - 2;
	    segs[1].x1 = core->width - 2;
	    segs[1].y1 = 1;
	    segs[1].x2 = core->width - 2;
	    segs[1].y2 = core->height - 2;
	    XDrawSegments(dpy, core->window, dim, segs, 2);
	    segs[0].x1 = 0;
	    segs[0].x2 = core->width - 1;
	    segs[0].y2 = segs[0].y1 = core->height - 1;
	    segs[1].x1 = segs[1].x2 = core->width - 1;
	    segs[1].y1 = 0;
	    segs[1].y2 = core->height - 1;
	    XDrawSegments(dpy, core->window, dark, segs, 2);
	    break;

	 case WB_TITLEBAR:
	    segs[0].x1 = 1;
	    segs[0].x2 = core->width - 1;
	    segs[0].y2 = segs[0].y1 = core->height - 2;
	    segs[1].x2 = segs[1].x1 = core->width - 1;
	    segs[1].y1 = 0;
	    segs[1].y2 = core->height - 1;
	    XDrawSegments(dpy, core->window, dim, segs, 2);
	    segs[0].x1 = 0;
	    segs[0].x2 = core->width - 1;
	    segs[0].y2 = segs[0].y1 = core->height - 1;
	    XDrawSegments(dpy, core->window, dark, segs, 1);
	    break;
	}

	segs[0].x1 = segs[0].y1 = segs[0].y2 = 0;
	segs[0].x2 = core->width - 2;
	segs[1].x1 = segs[1].y1 = 0;
	segs[1].x2 = 0;
	segs[1].y2 = core->height - 2;
	XDrawSegments(dpy, core->window, light, segs, 2);
	
	if (border==WB_ICON) {
	    segs[0].x1 = segs[0].y1 = segs[0].y2 = 1;
	    segs[0].x2 = core->width - 2;
	    segs[1].x1 = segs[1].y1 = 1;
	    segs[1].x2 = 1;
	    segs[1].y2 = core->height - 2;
	    XDrawSegments(dpy, core->window, light, segs, 2);
	}
	break;
	
     case WREL_SUNKEN:
	switch (border) {
	 case WB_ICON:
	 case WB_NORMAL:
	    /*
	    segs[0].x1 = 1;
	    segs[0].x2 = core->width - 2;
	    segs[0].y2 = segs[0].y1 = 1;
	    segs[1].x2 = segs[1].x1 = 1;
	    segs[1].y1 = 1;
	    segs[1].y2 = core->height - 2;
	    XDrawSegments(dpy, core->window, dark, segs, 2);
	    */
	    segs[0].x1 = 0;
	    segs[0].x2 = core->width - 1;
	    segs[0].y2 = segs[0].y1 = 0;
	    segs[1].x1 = segs[1].x2 = 0;
	    segs[1].y1 = 0;
	    segs[1].y2 = core->height - 1;
	    XDrawSegments(dpy, core->window, dark, segs, 2);
	    
	    segs[0].x1 = 0;
	    segs[0].y1 = segs[0].y2 = core->height - 1;
	    segs[0].x2 = core->width - 1;
	    segs[1].x2 = segs[1].x1 = core->width - 1;
	    segs[1].y1 = 1;
	    segs[1].y2 = core->height - 1;
	    XDrawSegments(dpy, core->window, light, segs, 2);
	    break;
	}
	break;	
    }
}

