/*
 *  WindowMaker window manager
 * 
 *  Copyright (c) 1997 Alfredo K. Kojima
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "generic/wwmlib.h"
#include "WindowMaker.h"
#include "GNUStep.h"
#include "core.h"
#include "box.h"
#include "window.h"
#include "icon.h"
#include "properties.h"
#include "actions.h"
#include "client.h"
#include "appmenu.h"
#include "resource.h"

/****** Global Variables ******/

extern WWindow *wWindowList;

extern WScreen wScreen;
/* modules */
extern int ModuleNo;

/* contexts */
extern XContext wWinContext;

/* cursors */
extern Cursor wCursor[WCUR_LAST];

/* protocol atoms */
extern Atom _XA_WM_DELETE_WINDOW;


/* local functions */
static FocusMode getFocusMode(WWindow *wwin);
static void getNormalHints(WWindow *wwin, XWindowAttributes *wattribs);
static void putWindowBorders(WWindow *wwin, int width, int height, 
			     int *child_x, int *child_y);
static void setupGrabs(WWindow *wwin);
static void removeGrabs(WWindow *wwin);
static void setupResources(WWindow *wwin, WResource *resource);

/* event handlers */

/* titlebar */
static void titlebarMouseDown(WObjDescriptor *desc, XEvent *event);
static void titlebarDblClick(WObjDescriptor *desc, XEvent *event);

static WMethodTable wWindowTitlebarMethods = {
    (WEventHandler)titlebarMouseDown,
      NULL,
      (WEventHandler)titlebarDblClick
};

/* handle */
static void handleMouseDown(WObjDescriptor *desc, XEvent *event);

static WMethodTable wWindowHandleMethods = {
    (WEventHandler)handleMouseDown,
      NULL,
      NULL
};



/* frame window (during window grabs) */
static void frameMouseDown(WObjDescriptor *desc, XEvent *event);

static WMethodTable wWindowFrameMethods = {
    (WEventHandler)frameMouseDown,
      NULL,
      NULL
};


/* close button */
static void windowCloseClick(WObjDescriptor *desc, XEvent *event);
static void windowCloseDblClick(WObjDescriptor *desc, XEvent *event);

static WMethodTable wWindowCloseMethods;

/* iconify button */
static void windowIconifyClick(WObjDescriptor *desc, XEvent *event);

static WMethodTable wWindowIconifyMethods;


/*
 *----------------------------------------------------------------------
 * wWindowInitialize--
 * 	Initialize window "class".
 * 
 * Notes:
 * 	Method table initialization to user defined bindings is done 
 * here.
 *---------------------------------------------------------------------- 
 */
void
wWindowInitialize()
{
    memcpy(&wWindowCloseMethods, &wButtonMethods, sizeof(WMethodTable));
    wWindowCloseMethods.click = windowCloseClick;
    wWindowCloseMethods.dbl_click = windowCloseDblClick;    
    
    memcpy(&wWindowIconifyMethods, &wButtonMethods, sizeof(WMethodTable));
    wWindowIconifyMethods.click = windowIconifyClick;
}



WWindow *
wFindWWindow(Window window)
{
    WObjDescriptor *desc;

    if (XFindContext(dpy, window, wWinContext, (XPointer*)&desc)==XCNOENT)
      return NULL;
    if (desc->parent_type==WCLASS_WINDOW)
      return desc->parent;
    else
      return NULL;
}




WWindow *
wWindowCreate()
{
    WWindow *wwin;
    
    wwin = wmalloc(sizeof(WWindow));

    memset(wwin, 0, sizeof(WWindow));

    /* setup global decoration defaults */
    wwin->window_flags = wResourceDefault.values;

    wwin->client_descriptor.methods = NULL;
    wwin->client_descriptor.parent = wwin;
    wwin->client_descriptor.self = wwin;
    wwin->client_descriptor.parent_type = WCLASS_WINDOW;
    
    wwin->belongs_to = NULL;
    return wwin;
}


void
wWindowDestroy(WWindow *wwin)
{
    int i;

    if (wwin->title!=NULL)
      XFree(wwin->title);
        
    if (wwin->normal_hints)
      wfree(wwin->normal_hints);
    
    if (wwin->wm_hints)
      XFree(wwin->wm_hints);
    
    if (wwin->wm_instance)
      XFree(wwin->wm_instance);
    
    if (wwin->wm_gnustep_attr)
      XFree(wwin->wm_gnustep_attr);

    XDeleteContext(dpy, wwin->client_win, wWinContext);

    for (i=0; i<wwin->l_button_no; i++) {
	if (wwin->l_buttons[i]!=NULL) {	    
	    wButtonDestroy(wwin->l_buttons[i]);
	}	
    }    
    for (i=0; i<wwin->r_button_no; i++) {
	if (wwin->r_buttons[i]!=NULL) {
	    wButtonDestroy(wwin->r_buttons[i]);
	}
    }

    if (wwin->titlebar)
	wBoxDestroy(wwin->titlebar);

    if (wwin->handle)
   	wBoxDestroy(wwin->handle);    

    wCoreDestroy(wwin->frame);

    if (wwin->icon) 
      wIconDestroy(wwin->icon);

    wfree(wwin);
}


/*
 *----------------------------------------------------------------
 * wManageWindow--
 * 	reparents the window and allocates a descriptor for it.
 * 
 * Returns:
 * 	the new window descriptor
 * 
 * Side effects:
 * 	The window is reparented and appropriate notification
 * is done to the client. Input mask for the window is setup.
 * 	The window descriptor is also associated with various window 
 * contexts and inserted in the head of the window list.
 * Event handler contexts are associated for some objects.
 *---------------------------------------------------------------- 
 */
WWindow *
wManageWindow(Window window)
{
    WWindow *wwin;
    int px, py, width, height;
    XWindowAttributes wattribs;    
    XSetWindowAttributes attribs;
    WResource *resource;

    wwin = wWindowCreate();
#ifdef DEBUG
    printf("managing window %x\n", (unsigned)window);
#endif
    
    /* setup descriptor */
    wwin->client_win = window;
    wwin->screen_ptr = &wScreen;
    wwin->focused_texture = wwin->screen_ptr->focused_texture;
    wwin->unfocused_texture = wwin->screen_ptr->unfocused_texture;

    wwin->title_font = wwin->screen_ptr->title_font;

    wwin->focused_text_gc = wwin->screen_ptr->focused_text_gc;
    wwin->unfocused_text_gc = wwin->screen_ptr->unfocused_text_gc;

    XSelectInput(dpy, window, PropertyChangeMask);

    /* prepare for reparenting */
#ifndef NOGRAB
    XGrabServer(dpy);
#endif
    XGetWindowAttributes(dpy, window, &wattribs);
    wwin->old_border_width = wattribs.border_width;

    attribs.event_mask = CLIENT_EVENTS;
    attribs.do_not_propagate_mask = ButtonPressMask | ButtonReleaseMask;
    XChangeWindowAttributes(dpy, window, CWEventMask|CWDontPropagate,
			    &attribs);

    XSetWindowBorderWidth(dpy, window, 0);
			  
    /* fill in property/hint data */
    if (!XFetchName(dpy, window, &(wwin->title))) {
	wwin->title = NULL;
	wwin->title_width = XTextWidth(wwin->title_font, DEF_WINDOW_TITLE,
				       strlen(DEF_WINDOW_TITLE));
    } else
      wwin->title_width = XTextWidth(wwin->title_font, wwin->title,
				     strlen(wwin->title));

    
    PropGetWMClass(window, &wwin->wm_class, &wwin->wm_instance);

    /* get hints from GNUStep app */
    if (PropGetGNUStepAttr(window, &wwin->wm_gnustep_attr)) {
	wwin->gnustep_app = 1;
    }

    wwin->wm_hints = XGetWMHints(dpy, window);
    
    if (wwin->wm_hints)  {
	if (wwin->wm_hints->flags & WindowGroupHint) {
	    if (wwin->wm_hints->window_group==window)
	      wwin->group_leader=1;
	    wwin->group_id = wwin->wm_hints->window_group;
	} else {
	    wwin->group_id = None;
	}

	if (wwin->wm_hints->flags & UrgencyHint)
	  wwin->urgent = 1;
    } else {
	wwin->group_id = None;
    }
    

    PropGetProtocols(window, &wwin->protocols);
    
    if (!PropGetTransientForHint(window, &wwin->transient_for)) {
	wwin->transient_for = None;
    } else {
	if (wwin->transient_for==0 || wwin->transient_for==window) {
	    wwin->transient_for = wwin->screen_ptr->root_win;
	}
    }    

    /* guess the focus mode */
    wwin->focus_mode = getFocusMode(wwin);

    /* get geometry stuff */
    getNormalHints(wwin, &wattribs);

    
    /* Setup the decoration/window attributes */
    
    /*
     * Decoration setting is done in this precedence (lower to higher)
     * - use global defaults (done in wWindowCreate())
     * - try to guess some settings
     * - use default in the resource database
     * - use GNUStep window attributes
     */

    if (wwin->protocols.DELETE_WINDOW || wwin->protocols.SAVE_YOURSELF)
      wwin->window_flags.broken_close = 0;
    else
      wwin->window_flags.broken_close = 1;

    /* transients can't be iconified or maximized */
    if (wwin->transient_for) {
	wwin->window_flags.miniaturizable=0;
#ifdef DEBUG
	printf("%x is transient for %x\n", (unsigned)window, 
	       (unsigned)wwin->transient_for);
#endif	
    }

    /* if the window can't be resized, remove the resize handle */
    if (wwin->normal_hints->flags & (PMinSize|PMaxSize)
	&& (wwin->normal_hints->min_width==wwin->normal_hints->max_width)
	&& (wwin->normal_hints->min_height==wwin->normal_hints->max_height)) {
	wwin->window_flags.resizable=0;
    }
    
    /* get default window attribute stuff from resource database */
    if ((resource=wResourceGet(wwin->title, wwin->wm_instance, 
			       wwin->wm_class))) {
	setupResources(wwin, resource);	
    }

    /* set GNUStep window attributes */
    if (wwin->gnustep_app) {
	wwin->window_flags.titlebar = ((wwin->wm_gnustep_attr->window_flags 
					& NSTitledWindowMask)?1:0);
	wwin->window_flags.closable = ((wwin->wm_gnustep_attr->window_flags 
				       & NSClosableWindowMask)?1:0);
	
	wwin->window_flags.miniaturizable = 
	  ((wwin->wm_gnustep_attr->window_flags 
	    & NSMiniaturizableWindowMask)?1:0);
	
	wwin->window_flags.resizable = ((wwin->wm_gnustep_attr->window_flags 
					& NSResizableWindowMask)?1:0);
	
	
	wwin->window_flags.floating = ((wwin->wm_gnustep_attr->window_level
				       == NSFloatingWindowLevel)?1:0);
    } 

    /* setup window geometry */
    width = wwin->client.width;
    height = wwin->client.height;

    wWindowConstrainSize(wwin, &width, &height);
    
    wwin->client.width = width;
    wwin->client.height = height;
#ifdef DEBUG
    printf("Geometry %i,%i %ix%i\n",wwin->client.x, wwin->client.y,
	   width,height);
#endif
    wwin->frame_x = wwin->client.x;
    wwin->frame_y = wwin->client.y;
    
    /* 
     * Start Reparenting 
     */
    wwin->frame = wCoreCreateParent(&wScreen, wwin->frame_x, wwin->frame_y,
				    width, height);
    /* create borders */
    putWindowBorders(wwin, width, height, &px, &py);

    wwin->client.x += px;
    wwin->client.y += py;
    XReparentWindow(dpy, wwin->client_win, wwin->frame->window, px, py);
    XResizeWindow(dpy, wwin->client_win, wwin->client.width, 
		  wwin->client.height);
#ifndef NOGRAB
    XUngrabServer(dpy);
#endif

    /* setup the frame descriptor */
    wwin->frame->descriptor.methods = &wWindowFrameMethods;
    wwin->frame->descriptor.parent = wwin;
    wwin->frame->descriptor.parent_type = WCLASS_WINDOW;

    /* do not save context for our windows, as they may have their own
     * contexts saved */
    if (!wwin->wm_class || strcmp(wwin->wm_class, "WindowMaker")!=0) {
	/* save a dummy descriptor to associate the client to it's wwin */
	XSaveContext(dpy, window, wWinContext, 
		     (XPointer)&wwin->client_descriptor);
    } else {
	/* don't let our windows be shaded */
	wwin->window_flags.shadeable = 0;
    }

    /* update internal descriptors */
    /* if window wants focus, include at the head of the 
     * window list, as it will receive it now */
    if ((wWindowList==NULL) || wwin->window_flags.focusable) {
	wwin->next = NULL;
	wwin->prev = wWindowList;
	if (wWindowList!=NULL) {
	    wWindowList->next = wwin;
	    wWindowUnfocus(wWindowList);
	}
	wWindowList = wwin;
    } else {
	wwin->next = wWindowList;
	wwin->prev = wWindowList->prev;
	if (wWindowList->prev) {
	    wWindowList->prev->next = wwin;
	}
	wWindowList->prev = wwin;
    }
    XLowerWindow(dpy, window);    
    XMapSubwindows(dpy, wwin->frame->window);
    XMapRaised(dpy, wwin->frame->window);
    
    /*
     * Prevent window withdrawal when getting the 
     * unmap notifies generated during reparenting
     */
    wwin->flags.mapped=0;
    wClientSetState(wwin, NormalState);
    
    wSetFocusTo(wwin);    
    XSync(dpy, 0);
    return wwin;
}


/*
 *---------------------------------------------------------------------- 
 * wUnmanageWindow--
 * 	Removes the frame window from a window and destroys all data
 * related to it. The window will be reparented back to the root window
 * if restore is True.
 * 
 * Side effects:
 * 	Everything related to the window is destroyed and the window
 * is removed from the window list. Focus is set to the next on the
 * window list.
 *---------------------------------------------------------------------- 
 */
void
wUnmanageWindow(WWindow *wwin, int restore)
{    
    /* reparent the window back to the root */
    if (restore)
      wClientRestore(wwin);

    /* remove from the window list */
    if (wwin->prev)
      wwin->prev->next = wwin->next;
    
    if (wwin->next)
      wwin->next->prev = wwin->prev;
    else {	
	/* this window was the focused window, so pass the focus to
	 * the previous one in the window list */
	wWindowList = wwin->prev;
  	if (wwin->prev)
	  wSetFocusTo(wwin->prev);
    }
#ifdef DEBUG
    printf("destroying window %x frame %x\n", (unsigned)wwin->client_win,
	   (unsigned)wwin->frame->window);
#endif 
    wWindowDestroy(wwin);
    XSync(dpy, 0);
}

/*
 *---------------------------------------------------------------------- 
 * wWindowPaint--
 * 	Paints the requested parts of the window.
 * If repaint is true, the background will be repainted from scratch.
 * 
 * Returns:
 * 	None
 * 
 * Side effects:
 * 	None
 *---------------------------------------------------------------------- 
 */
void
wWindowPaint(WWindow *wwin, int sides, int repaint)
{
    int x, i;
    GC gc, light_gc, dim_gc;

    if (wwin->window_flags.titlebar && (sides & BORDER_TOP)) {
	wBoxPaint(wwin->titlebar, repaint);
	if (wwin->flags.focused) {
	    gc = wwin->focused_text_gc;
	} else {
	    gc = wwin->unfocused_text_gc;
	}
	x = (wwin->titlebar->core->width - wwin->title_width)/2;
	if (wwin->title)
	  XDrawString(dpy, wwin->titlebar->core->window, gc, x,
		      wwin->title_font->ascent + 3, wwin->title,
		      strlen(wwin->title));
	else 
	  XDrawString(dpy, wwin->titlebar->core->window, gc, x,
		      wwin->title_font->ascent + 3, DEF_WINDOW_TITLE,
		      strlen(DEF_WINDOW_TITLE));

	for (i=0; i<wwin->l_button_no; i++) {
	    wButtonPaint(wwin->l_buttons[i], repaint);
	}
	for (i=0; i<wwin->r_button_no; i++) {
	    wButtonPaint(wwin->r_buttons[i], repaint);
	}
    }
    if (wwin->window_flags.resizable && (sides & BORDER_BOTTOM)) {
	Window win;
	int w, h;
	int cw;

	w = wwin->handle->core->width;
	h = wwin->handle->core->height;
	cw = wwin->handle_corner_width;
	light_gc = wwin->handle->texture->any.relief->light_gc;
	dim_gc = wwin->handle->texture->any.relief->dim_gc;
	win = wwin->handle->core->window;

	wBoxPaint(wwin->handle, repaint);

	XDrawLine(dpy, win, dim_gc, 0, 0, w, 0);
	XDrawLine(dpy, win, light_gc, 0, 1, w, 1);

	XDrawLine(dpy, win, dim_gc, cw, 2, cw, h);
	XDrawLine(dpy, win, light_gc, cw+1, 2, cw+1, h);

	XDrawLine(dpy, win, dim_gc, w-cw-2, 2, w-cw-2, h);
	XDrawLine(dpy, win, light_gc, w-cw-1, 2, w-cw-1, h);
    }
}


     
void
wWindowFocus(WWindow *wwin)
{
#ifdef NEWSTYLE    
    int i;
    WPixmap *pix;
#endif
    
    if (wwin->titlebar) {	
	wwin->titlebar->texture = wwin->focused_texture;
#ifdef NEWSTYLE	
	if (wwin->flags.focused) {
	    for (i=0; i<wwin->l_button_no; i++) {
		wwin->l_buttons[i]->box->texture = wwin->focused_texture;
		pix = wwin->l_buttons[i]->image;
		wwin->l_buttons[i]->image = wwin->l_buttons[i]->image2;
		wwin->l_buttons[i]->image2 = pix;
		pix = wwin->l_buttons[i]->alt_image;
		wwin->l_buttons[i]->alt_image = wwin->l_buttons[i]->alt_image2;
		wwin->l_buttons[i]->alt_image2 = pix;
	    }
	    for (i=0; i<wwin->r_button_no; i++) {
		wwin->r_buttons[i]->box->texture = wwin->focused_texture;
		pix = wwin->r_buttons[i]->image;
		wwin->r_buttons[i]->image = wwin->r_buttons[i]->image2;
		wwin->r_buttons[i]->image2 = pix;
		pix = wwin->r_buttons[i]->alt_image;
		wwin->r_buttons[i]->alt_image = wwin->r_buttons[i]->alt_image2;
		wwin->r_buttons[i]->alt_image2 = pix;
	    }
	}
#endif
    }    
    if (wwin->flags.focused)
      wWindowPaint(wwin, BORDER_ALL, True);
    else {
	wwin->flags.focused=1;
	wWindowPaint(wwin, BORDER_ALL, False);
    }
    removeGrabs(wwin);    

}

void
wWindowUnfocus(WWindow *wwin)
{
#ifdef NEWSTYLE    
    int i;
    WPixmap *pix;
#endif

    if (wwin->titlebar) {
	wwin->titlebar->texture = wwin->unfocused_texture;
#ifdef NEWSTYLE	
	if (wwin->flags.focused) {
	    for (i=0; i<wwin->l_button_no; i++) {
		wwin->l_buttons[i]->box->texture = wwin->unfocused_texture;
		pix = wwin->l_buttons[i]->image;
		wwin->l_buttons[i]->image = wwin->l_buttons[i]->image2;
		wwin->l_buttons[i]->image2 = pix;
		pix = wwin->l_buttons[i]->alt_image;
		wwin->l_buttons[i]->alt_image = wwin->l_buttons[i]->alt_image2;
		wwin->l_buttons[i]->alt_image2 = pix;
	    }
	    for (i=0; i<wwin->r_button_no; i++) {
		wwin->r_buttons[i]->box->texture = wwin->unfocused_texture;
		pix = wwin->r_buttons[i]->image;
		wwin->r_buttons[i]->image = wwin->r_buttons[i]->image2;
		wwin->r_buttons[i]->image2 = pix;
		pix = wwin->r_buttons[i]->alt_image;
		wwin->r_buttons[i]->alt_image = wwin->r_buttons[i]->alt_image2;
		wwin->r_buttons[i]->alt_image2 = pix;
	    }
	}
#endif	
    }
    wwin->flags.focused=0;
    wWindowPaint(wwin, BORDER_ALL, False);

    setupGrabs(wwin);

}


void
wWindowConstrainSize(WWindow *wwin, int *nwidth, int *nheight)
{
    XSizeHints *sizeh=wwin->normal_hints;
    int width = *nwidth;
    int height = *nheight;
    int winc = sizeh->width_inc;
    int hinc = sizeh->height_inc;
    int aspw = (sizeh->min_aspect.x + sizeh->max_aspect.x)/2;
    int asph = (sizeh->min_aspect.y + sizeh->max_aspect.y)/2;

    if (width < sizeh->min_width)
      width = sizeh->min_width;
    if (height < sizeh->min_height)
      height = sizeh->min_height;

    if (width > sizeh->max_width) 
      width = sizeh->max_width;
    if (height > sizeh->max_height) 
      height = sizeh->max_height;
    
    if (sizeh->flags & PAspect) {
	/* adjust min aspect ratio */
	if (width * sizeh->min_aspect.x > height * sizeh->min_aspect.y) {
	    if (aspw > asph) {
		height = (aspw/asph) * width;
		width = (asph/aspw) * height;
	    } else {
		width = (asph/aspw) * height;
		height = (aspw/asph) * width;
	    }
	}
	/* adjust max aspect ratio */
	if (width * sizeh->max_aspect.x < height * sizeh->max_aspect.y) {
	    if (aspw > asph) {
		height = (aspw/asph) * width;
		width = (asph/aspw) * height;
	    } else {
		width = (asph/aspw) * height;
		height = (aspw/asph) * width;
	    }
	}	
    }
    width = (((width - sizeh->base_width) / winc) * winc) + sizeh->base_width; 
    height = (((height - sizeh->base_height) / hinc) * hinc)
      + sizeh->base_height;
    
    *nwidth = width;
    *nheight = height;
}

#ifdef NEWSTYLE
static void
adjustButtons(WWindow *wwin)
{
    int i;
    int x, w, bsize;

    w = wwin->frame->width;
    bsize = wwin->titlebar->core->height;

    if ((bsize * (wwin->l_button_no + wwin->r_button_no) > w/2) 
	&& !(wwin->flags.no_buttons)) {
	wwin->flags.no_buttons=1;
	for (i=0; i<wwin->r_button_no; i++) {	
	    XUnmapWindow(dpy, wwin->r_buttons[i]->box->core->window);
	}
    }
    if (wwin->flags.no_buttons) {
	for (i=0; i<wwin->r_button_no; i++) {	
	    XMapWindow(dpy, wwin->r_buttons[i]->box->core->window);
	}
    }
    wwin->flags.no_buttons=0;
    x = w - bsize + 1;
    for (i=0; i<wwin->r_button_no; i++) {
	XMoveWindow(dpy, wwin->r_buttons[i]->box->core->window, x, 0);
	x -= bsize;
    }
}
#else /* !NEWSTYLE */
static void
adjustButtons(WWindow *wwin)
{
    int i;
    int x, y, w, bsize;

    w = wwin->frame->width;
    bsize = wwin->titlebar->core->height - 7;

    if ((bsize * (wwin->l_button_no + wwin->r_button_no) > w/2) 
	&& !wwin->flags.no_buttons) {
	wwin->flags.no_buttons=1;
	XUnmapSubwindows(dpy, wwin->titlebar->core->window);
	return;
    }
    if (wwin->flags.no_buttons) {
	XMapSubwindows(dpy, wwin->titlebar->core->window);
    }
    wwin->flags.no_buttons=0;
    x = w - 3 - bsize;
    y = (wwin->titlebar->core->height - bsize)/2;
    for (i=0; i<wwin->r_button_no; i++) {
	XMoveWindow(dpy, wwin->r_buttons[i]->box->core->window, x, y);
	x -= (bsize + 3);
    }
}
#endif /* NEWSTYLE */

void 
wWindowConfigure(wwin, req_x, req_y, req_width, req_height)
WWindow *wwin;
int req_x, req_y;		       /* new position of the frame */
int req_width, req_height;	       /* new size of the frame */
{
    int synth_notify = False;
    int resize;

    resize = (req_width!=wwin->frame->width 
	      || req_height!=wwin->frame->height);
    /* 
     * if the window is being moved but not resized then
     * send a synthetic ConfigureNotify
     */
    if ((req_x!=wwin->frame_x || req_y!=wwin->frame_y) && !resize) {
	synth_notify = True;
    }
    /* do not let the window titlebar get unreachable */
    if (req_y<0) 
      req_y = 0;

    if (resize) {
	int cx, cy, cw, ch;

	if (req_width <= 0)
	  req_width = 1;
	if (req_height <= 0)
	  req_height = 1;

	cx = req_x;
	cy = req_y;
	cw = req_width;
	ch = req_height;
	/* resize the titlebar horizontally */
	if (wwin->window_flags.titlebar) {
#ifdef NEWSTYLE
	    int ls = (wwin->titlebar->core->height-1)*wwin->l_button_no;
	    int rs = (wwin->titlebar->core->height-1)*wwin->r_button_no;
	    wCoreConfigure(wwin->titlebar->core, ls, 0, req_width-ls-rs, 0);
#else	    
	    wCoreConfigure(wwin->titlebar->core, 0, 0, req_width, 0);
#endif
	    ch -= wwin->top_width;
	    cy += wwin->top_width;
	}
	/* resize the handle horizontally and move it */
	if (wwin->window_flags.resizable) {
	    ch -= wwin->handle->core->height;
	    wCoreConfigure(wwin->handle->core, 0, 
			   wwin->top_width + ch,
			   req_width, 0);
	}
	wwin->client.x = cx;
	wwin->client.y = cy;
	wwin->client.width = cw;
	wwin->client.height = ch;
	XMoveResizeWindow(dpy, wwin->client_win, 0, wwin->top_width, cw, ch);
	wCoreConfigure(wwin->frame, req_x, req_y, req_width, req_height);
	if (wwin->window_flags.titlebar)
	  adjustButtons(wwin);
    } else {
	int cx, cy;
	
	cx = req_x;
	cy = req_y;
	if (wwin->window_flags.titlebar) {
	    cy += wwin->top_width;
	}
	wCoreConfigure(wwin->frame, req_x, req_y, 0, 0);
	wwin->client.x = cx;
	wwin->client.y = cy;
    }
    wwin->frame_x = req_x;
    wwin->frame_y = req_y;
    if (synth_notify) {
	XEvent sevent;
	
	sevent.type = ConfigureNotify;
	sevent.xconfigure.display = dpy;
	sevent.xconfigure.event = wwin->client_win;
	sevent.xconfigure.window = wwin->client_win;

	sevent.xconfigure.x = wwin->client.x;
	sevent.xconfigure.y = wwin->client.y;
	sevent.xconfigure.width = wwin->client.width;
	sevent.xconfigure.height = wwin->client.height;
 
	sevent.xconfigure.border_width = wwin->old_border_width;
	if (wwin->titlebar)
	  sevent.xconfigure.above = wwin->titlebar->core->window;
	else
	  sevent.xconfigure.above = None;
	sevent.xconfigure.override_redirect = False;
	XSendEvent(dpy, wwin->client_win, False, StructureNotifyMask, &sevent);
    }
    if (req_width < (HANDLE_CORNER_WIDTH*2 + HANDLE_MIN_WIDTH)) {
	wwin->handle_corner_width = (req_width - HANDLE_MIN_WIDTH)/2;
	if (wwin->handle_corner_width < 0)
	  wwin->handle_corner_width = 0;
    } else {
	wwin->handle_corner_width = HANDLE_CORNER_WIDTH;
    }
}


void
wWindowDumpInfo(WWindow *wwin)
{
    printf("Window: %s [%x]\n", wwin->title, (unsigned)wwin->client_win); 
    printf("Geometry: %i,%i %ix%i\n", wwin->client.x, wwin->client.y,
	   wwin->client.width, wwin->client.height);
    printf("Frame [%x]: %ix%i\n", (unsigned)wwin->frame->window,
	   wwin->frame->width, wwin->frame->height);    
    if (wwin->window_flags.titlebar) {
	printf("Titlebar [%x]: %ix%i\n", (unsigned)wwin->titlebar->core->window,
	       wwin->titlebar->core->width, wwin->titlebar->core->height);
    }
    if (wwin->window_flags.resizable) {
	printf("Handle [%x]: %ix%i\n", (unsigned)wwin->handle->core->window,
	       wwin->handle->core->width, wwin->handle->core->height);
    }
    printf("Gravity: %i\n", wwin->normal_hints->win_gravity);
}


/* ====================================================================== */

static FocusMode
getFocusMode(WWindow *wwin)
{    
    FocusMode mode;
    
    if ((wwin->wm_hints) && (wwin->wm_hints->flags & InputHint)) {
	if (wwin->wm_hints->input == True) {
	    if (wwin->protocols.TAKE_FOCUS)
	      mode = WFM_LOCALLY_ACTIVE;
	    else
	      mode = WFM_PASSIVE;
	} else {
	    if (wwin->protocols.TAKE_FOCUS)
	      mode = WFM_GLOBALLY_ACTIVE;
	    else
	      mode = WFM_NO_INPUT;
	}
    } else {
	mode = WFM_PASSIVE;
    }
    return mode;
}


/*
 *---------------------------------------------------------------------- 
 * getNormalHints--
 * 	Get size (normal) hints and a default geometry for the client 
 * window. The hints are also checked for inconsistency.
 * 
 * Side effects:
 * 	normal_hints is filled with valid data and geometry information 
 * about the client is also filled.
 *---------------------------------------------------------------------- 
 */
static void
getNormalHints(WWindow *wwin, XWindowAttributes *wattribs)
{
    int pre_icccm=0;
    
    /* find a position for the window */
    wwin->normal_hints = XAllocSizeHints();
    if (!PropGetNormalHints(wwin->client_win, wwin->normal_hints, &pre_icccm)) {
	wwin->normal_hints->flags = 0;
    }
    wwin->client.x = wattribs->x;
    wwin->client.y = wattribs->y;
    
    wwin->client.width = wattribs->width;
    wwin->client.height = wattribs->height;

    if (!CKFLAG(wwin->normal_hints->flags, PWinGravity)) {
	wwin->normal_hints->win_gravity = NorthWestGravity;
    }
    if (!CKFLAG(wwin->normal_hints->flags, PMinSize)) {
	wwin->normal_hints->min_width = MIN_WSIZE;
	wwin->normal_hints->min_height = MIN_WSIZE;
    }
    if (!CKFLAG(wwin->normal_hints->flags, PBaseSize)) {
	wwin->normal_hints->base_width = wwin->normal_hints->min_width;
	wwin->normal_hints->base_height = wwin->normal_hints->min_height;      
    }    
    if (!CKFLAG(wwin->normal_hints->flags, PMaxSize)) {
	wwin->normal_hints->max_width = wwin->screen_ptr->scr_width*2;
	wwin->normal_hints->max_height = wwin->screen_ptr->scr_height*2;
    }
    
    if (!CKFLAG(wwin->normal_hints->flags, PResizeInc)) {
	wwin->normal_hints->width_inc = 1;
	wwin->normal_hints->height_inc = 1;
    } else {	
	if (wwin->normal_hints->width_inc <= 0)
	  wwin->normal_hints->width_inc = 1;
	if (wwin->normal_hints->height_inc <= 0)
	  wwin->normal_hints->height_inc = 1;
    }

    if (wwin->normal_hints->min_width <= 0) 
      wwin->normal_hints->min_width = MIN_WSIZE;
    if (wwin->normal_hints->min_height <= 0)
      wwin->normal_hints->min_height = MIN_WSIZE;

    if (wwin->normal_hints->min_height > wwin->normal_hints->max_height) {
	wwin->normal_hints->min_height = wwin->normal_hints->max_height;
    }
    if (wwin->normal_hints->min_width > wwin->normal_hints->max_width) {
	wwin->normal_hints->min_width = wwin->normal_hints->max_width;
    }

    /* pre ICCCM (old) client */
    if (pre_icccm) {
#ifdef DEBUG
	printf("PRE ICCCM\n");	
#endif	
	if (CKFLAG(wwin->normal_hints->flags, (USPosition|USPosition))) {
	    wwin->client.x = wwin->normal_hints->x;
	    wwin->client.y = wwin->normal_hints->y;
	}
	if (CKFLAG(wwin->normal_hints->flags, (USSize|PSize))) {
	    wwin->client.width = wwin->normal_hints->width;
	    wwin->client.height = wwin->normal_hints->height;
	}
    }
}

#ifdef DEBUG
void
printWindowList()
{
    WWindow *wwin;
    puts("-------------");
    for(wwin=wWindowList; wwin!=NULL && wwin!=wWindowList; wwin=wwin->prev) {
	printf("%s\n",wwin->title);
    }
    if (wwin==wWindowList) {
	puts("CIRCULAR LIST");	     
    }    
    puts("-------------");    
}

#endif


static void
setupGrabs(WWindow *wwin)
{
    return;
    /* setup passive grabs on the client window */
    XGrabButton(dpy, Button1, AnyModifier, wwin->client_win, True,
		ButtonPressMask | ButtonReleaseMask,
		GrabModeSync, GrabModeSync, None, None);
    XGrabButton(dpy, Button2, AnyModifier, wwin->client_win, True,
		ButtonPressMask | ButtonReleaseMask,
		GrabModeSync, GrabModeSync, None, None);
}

static void
removeGrabs(WWindow *wwin)
{
    return;
    /* remove passive grabs on the client window */
    XUngrabButton(dpy, Button1, AnyModifier, wwin->client_win);
    XUngrabButton(dpy, Button2, AnyModifier, wwin->client_win);
}



static void
setupResources(WWindow *wwin, WResource *resource)
{
    wwin->icon_file = resource->icon_file;

    if (resource->flags.titlebar)
      wwin->window_flags.titlebar=resource->values.titlebar;
    
    if (resource->flags.resizable)
      wwin->window_flags.resizable=resource->values.resizable;
    
    if (resource->flags.closable)
      wwin->window_flags.closable=resource->values.closable;

    if (resource->flags.broken_close)
      wwin->window_flags.broken_close=resource->values.broken_close;
    
    if (resource->flags.miniaturizable)
      wwin->window_flags.miniaturizable=resource->values.miniaturizable;
    
    if (resource->flags.shadeable)
      wwin->window_flags.shadeable=resource->values.shadeable;
    
    if (resource->flags.focusable)
      wwin->window_flags.focusable=resource->values.focusable;
    
    if (resource->flags.sticky)
      wwin->window_flags.sticky=resource->values.sticky;
    
    if (resource->flags.skip_window_list)
      wwin->window_flags.skip_window_list=resource->values.skip_window_list;
    
    if (resource->flags.floating)
      wwin->window_flags.floating=resource->values.floating;
    
    if (resource->flags.ignore_keys)
      wwin->window_flags.ignore_keys=resource->values.ignore_keys;
}
 

/*
 *---------------------------------------------------------------------- 
 * putWindowBorders--
 * 	Puts the titlebar and handle if specified in wwin->window_flags.
 *
 * Returns:
 * 	The position to place the child window inside the frame (child_x,
 *								 child_y)
 * 
 * Side effects:
 * 	The size of the frame is adjusted to accomodate the titlebar and
 * handle. The position of the frame is also adjusted according to the
 * gravity. wwin is associated to the created windows.
 * 
 *---------------------------------------------------------------------- 
 */
static void
putWindowBorders(wwin, width, height, child_x, child_y)
WWindow *wwin;
int width, height;		       /* size of the client window */
int *child_x, *child_y;		       /* returns the position of the client
					* window in the frame */
{
    int resized=0;
    int oy, ox;
    WScreen *scr = wwin->screen_ptr;
    int bsize;
    int new_height = wwin->frame->height;
    
    *child_x = 0;
    *child_y = 0;
    
    wClientGetGravityOffsets(wwin, &ox, &oy);

    if (wwin->window_flags.titlebar) {
	WButton *wbut=NULL;
	int x, w, h;
	
	h = wwin->title_font->ascent + wwin->title_font->descent + 8;
	w = width;
	wwin->titlebar = wBoxCreate(wwin->frame, 0, 0, w, h, WB_TITLEBAR);
	wwin->titlebar->texture = wwin->unfocused_texture;
	/* setup object descriptor */
	wwin->titlebar->core->descriptor.parent = wwin;
	wwin->titlebar->core->descriptor.parent_type = WCLASS_WINDOW;
	wwin->titlebar->core->descriptor.methods = &wWindowTitlebarMethods;

	resized = 1;
	*child_y += h;
	new_height += h;
	wwin->top_width += h;

	bsize = h - 7;
	/* put default titlebar buttons on left */
	if (wwin->window_flags.miniaturizable) {
#ifdef NEWSTYLE
	    x = 0;
	    wbut = wButtonCreate(wwin->frame, x, 0, h-1, h);
#else
	    x = 3;
	    wbut = wButtonCreate(wwin->titlebar->core, x, (h-bsize)/2, 
				 bsize, bsize);
#endif
	    /* setup object descriptor */
	    wbut->box->core->descriptor.parent = wwin;
	    wbut->box->core->descriptor.parent_type = WCLASS_WINDOW;
	    wbut->box->core->descriptor.methods = &wWindowIconifyMethods;
	    wbut->image = scr->b_pixmaps[WBUT_ICONIFY];
	    wbut->alt_image = scr->b_pixmaps[WBUT_ICONIFY2];
#ifdef NEWSTYLE
	    wbut->image2 = scr->b_pixmaps[WBUT_LAST+WBUT_ICONIFY];
	    wbut->alt_image2 = scr->b_pixmaps[WBUT_LAST+WBUT_ICONIFY2];
#endif
	    wwin->l_buttons[wwin->l_button_no++] = wbut;
	}

	/* put default titlebar buttons on right */
	if (wwin->window_flags.closable) {
#ifdef NEWSTYLE
	    x = width - h + 1;
	    wbut = wButtonCreate(wwin->frame, x, 0, h-1, h);
#else
	    x = width - 3 - bsize;
	    wbut = wButtonCreate(wwin->titlebar->core, x, (h-bsize)/2, 
				 bsize, bsize);
#endif
	    /* setup object descriptor */
	    wbut->box->core->descriptor.parent = wwin;
	    wbut->box->core->descriptor.parent_type = WCLASS_WINDOW;
	    wbut->box->core->descriptor.methods = &wWindowCloseMethods;
	    if (wwin->window_flags.broken_close) {
		wbut->image = scr->b_pixmaps[WBUT_BROKENCLOSE];
		wbut->alt_image = scr->b_pixmaps[WBUT_BROKENCLOSE2];
#ifdef NEWSTYLE
		wbut->image2 = scr->b_pixmaps[WBUT_LAST+WBUT_BROKENCLOSE];
		wbut->alt_image2=scr->b_pixmaps[WBUT_LAST+WBUT_BROKENCLOSE2];
#endif
	    } else {
		wbut->image = scr->b_pixmaps[WBUT_CLOSE];
		wbut->alt_image = scr->b_pixmaps[WBUT_CLOSE2];
#ifdef NEWSTYLE
		wbut->image2 = scr->b_pixmaps[WBUT_LAST+WBUT_CLOSE];
		wbut->alt_image2 = scr->b_pixmaps[WBUT_LAST+WBUT_CLOSE2];
#endif
	    }
	    wwin->r_buttons[wwin->r_button_no++] = wbut;
	}
	XMapSubwindows(dpy, wwin->titlebar->core->window);
    }

    if (wwin->window_flags.resizable) {
	wwin->handle = wBoxCreate(wwin->frame, 0, *child_y+height, width,
				  HANDLE_HEIGHT, WB_NONE);
	/* setup object descriptor */
	wwin->handle->core->descriptor.parent = wwin;
	wwin->handle->core->descriptor.parent_type = WCLASS_WINDOW;
	wwin->handle->core->descriptor.methods = &wWindowHandleMethods;
	wwin->handle->texture = wwin->unfocused_texture;
	
	resized = 1;
	wwin->bottom_width += HANDLE_HEIGHT;
	new_height += HANDLE_HEIGHT;
	
	wwin->handle_corner_width = HANDLE_CORNER_WIDTH;
	if (width < HANDLE_CORNER_WIDTH*2 + HANDLE_MIN_WIDTH) {
	    wwin->handle_corner_width = (width - HANDLE_MIN_WIDTH)/2;
	    if (wwin->handle_corner_width < 0)
	      wwin->handle_corner_width = 0;
	}
    }
    wwin->frame_x += ox * (wwin->old_border_width - FRAME_BORDER_WIDTH);
    wwin->frame_y += oy * (wwin->old_border_width - FRAME_BORDER_WIDTH);
    if (resized) {
	wWindowConfigure(wwin, wwin->frame_x, wwin->frame_y,
			 wwin->frame->width, new_height);
    }
}


/* ====================================================================== */

static void 
handleMouseDown(WObjDescriptor *desc, XEvent *event)
{
    WWindow *wwin = desc->parent;
    
    if (event->xbutton.state & Mod1Mask) {
	/* move the window */
	if (XGrabPointer(dpy, wwin->titlebar->core->window, False, 
			 ButtonMotionMask|ButtonReleaseMask,
			 GrabModeAsync, GrabModeAsync, None, 
			 None, CurrentTime)!=GrabSuccess) {
	    wWarning("pointer grab failed for window move");
	    return;
	}
	wMoveWindow(wwin);
	XUngrabPointer(dpy, CurrentTime);		
    } else {
	if (event->xbutton.button == Button1
	    || event->xbutton.button == Button2)
	  wSetFocusTo(wwin);
	if (event->xbutton.button == Button1) {
	    wRaiseWindow(wwin);
	    /* resize the window */
	    if (XGrabPointer(dpy, wwin->titlebar->core->window, False, 
			     ButtonMotionMask|ButtonReleaseMask,
			     GrabModeAsync, GrabModeAsync, None, 
			     None, CurrentTime)!=GrabSuccess) {
		wWarning("pointer grab failed for window resize");
		return;
	    }
	    wResizeWindow(wwin);
	    XUngrabPointer(dpy, CurrentTime);
	}
    }
}


static void 
titlebarDblClick(WObjDescriptor *desc, XEvent *event)
{
    WWindow *wwin = desc->parent;

    if (event->xbutton.state & ControlMask) {
	/* maximize vertically */
	if (wwin->window_flags.resizable) {
	    if (wwin->flags.maximized)
	      wUnmaximizeWindow(wwin);
	    else
	      wMaximizeWindow(wwin, False);
	}
    } else if (wwin->window_flags.shadeable) {
	/* shade window */
	if (wwin->flags.shaded)
	  wUnshadeWindow(wwin);
	else
	  wShadeWindow(wwin);
    }
}


static void 
titlebarMouseDown(WObjDescriptor *desc, XEvent *event)
{
    WWindow *wwin = desc->parent;

    wSetFocusTo(wwin);
    if (event->xbutton.button == Button1
	|| event->xbutton.button == Button2) {
	if (event->xbutton.button == Button1) {
	    if (event->xbutton.state & Mod1Mask)
	      wLowerWindow(wwin);
	    else
	      wRaiseWindow(wwin);
	}	
	/* move the window */
	if (XGrabPointer(dpy, wwin->titlebar->core->window, False, 
			 ButtonMotionMask|ButtonReleaseMask,
			 GrabModeAsync, GrabModeAsync, None, 
			 None, CurrentTime)!=GrabSuccess) {
	    wWarning("pointer grab failed for window move");
	    return;
	}
	wMoveWindow(wwin);

	XUngrabPointer(dpy, CurrentTime);
    }
}


static void
frameMouseDown(WObjDescriptor *desc, XEvent *event)
{
    WWindow *wwin = desc->parent;

    if (event->xbutton.button == Button1
	|| event->xbutton.button == Button2) {
	if (event->xbutton.button == Button1)
	  wRaiseWindow(wwin);

	#if 0
	if (event->xbutton.state & Mod1Mask) {
	    /* move the window */
	    if (XGrabPointer(dpy, wwin->frame->window, False, 
			     ButtonMotionMask|ButtonReleaseMask,
			     GrabModeAsync, GrabModeAsync, None, 
			     None, CurrentTime)!=GrabSuccess) {
		wWarning("pointer grab failed for window move");
		return;
	    }
	    wMoveWindow(wwin);
	    XUngrabPointer(dpy, CurrentTime);
	}
	#endif
    }
    wSetFocusTo(wwin);
}


static void
windowCloseClick(WObjDescriptor *desc, XEvent *event)
{
    WWindow *wwin=desc->parent;

    /* if control click, kill the client */
    if (event->xkey.state & ControlMask)
      wClientKill(wwin);
    /* send delete message */
    if (wwin->protocols.DELETE_WINDOW) {
	wClientSendProtocol(wwin, _XA_WM_DELETE_WINDOW, CurrentTime);
    }
}


static void
windowCloseDblClick(WObjDescriptor *desc, XEvent *event)
{
    WWindow *wwin=desc->parent;

    /* send delete message */
    if (wwin->protocols.DELETE_WINDOW) {
	wClientSendProtocol(wwin, _XA_WM_DELETE_WINDOW, CurrentTime);
    } else {
	wClientKill(wwin);
    }
}


static void
windowIconifyClick(WObjDescriptor *desc, XEvent *event)
{
    wIconifyWindow(desc->parent);
}

