/*
 *  WindowMaker window manager
 * 
 *  Copyright (c) 1997 Alfredo K. Kojima
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <sys/wait.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Intrinsic.h>
#include <X11/cursorfont.h>


#include "generic/wwmlib.h"
#include "WindowMaker.h"
#include "GNUstep.h"
#include "texture.h"
#include "screen.h"
#include "window.h"
#include "wconfig.h"

/****** Global Variables ******/

extern WPreferences wPreferences;

/* currently only single screen support */
extern WScreen wScreen;
/* modules */
extern int ModuleNo;

/* contexts */
extern XContext wWinContext;

/* atoms */
extern Atom _XA_WM_STATE;
extern Atom _XA_WM_CHANGE_STATE;
extern Atom _XA_WM_PROTOCOLS;
extern Atom _XA_WM_TAKE_FOCUS;
extern Atom _XA_WM_DELETE_WINDOW;
extern Atom _XA_WM_SAVE_YOURSELF;
extern Atom _XA_WM_CLIENT_LEADER;
extern Atom _XA_WM_COLORMAP_WINDOWS;

extern Atom _XA_GNUSTEP_WM_ATTR;

/* cursors */
extern Cursor wCursor[WCUR_LAST];



static void manageAllWindows();


static int 
catchXError(Display *dpy, XErrorEvent *error)
{
    char buffer[MAXLINE];
    
    wFormatXError(dpy, error, buffer, MAXLINE); 
    wWarning("internal X error: %s\n", buffer);
    return -1;
}


/*
 *----------------------------------------------------------------------
 * handleSig--
 * 	general signal handler. Exits the program gently.
 *---------------------------------------------------------------------- 
 */
static void 
handleSig(int signal)
{
    wFatal("got signal %i\n", signal);

    wAbort();
    exit(1);
}


static void
handleChild(int signal)
{
    waitpid(-1, NULL, WNOHANG);
}

	       
/*
 *---------------------------------------------------------- 
 * StartUp--
 * 	starts the window manager and setup global data
 * 
 * display - display to run on
 * 
 * Side effects:
 * global data declared in main.c is initialized
 *----------------------------------------------------------
 */
void 
StartUp(char *on_display)
{
    char *display_name;
    
    ModuleNo = 1; /* 0 is ourself */

    display_name = XDisplayName(on_display);
    dpy = XOpenDisplay(on_display);
    if (dpy == NULL) {
	wFatal("could not open display %s", display_name);
	exit(1);
    }
#ifdef DEBUG
    XSynchronize(dpy, True);
#endif

    /* setup default preferences */
    wPreferences.size_display=0;
    wPreferences.move_display=0;
    wPreferences.colormap_follows_mouse=0;
    wPreferences.focus_follows_mouse=0;
    wPreferences.use_real_colors=1;

    wWinContext = XUniqueContext();
    _XA_WM_STATE = XInternAtom(dpy, "WM_STATE", False);
    _XA_WM_CHANGE_STATE = XInternAtom(dpy, "WM_CHANGE_STATE", False);
    _XA_WM_PROTOCOLS = XInternAtom(dpy, "WM_PROTOCOLS", False);
    _XA_WM_TAKE_FOCUS = XInternAtom(dpy, "WM_TAKE_FOCUS", False);
    _XA_WM_DELETE_WINDOW = XInternAtom(dpy, "WM_DELETE_WINDOW", False);
    _XA_WM_SAVE_YOURSELF = XInternAtom(dpy, "WM_SAVE_YOURSELF", False);    
    _XA_WM_CLIENT_LEADER = XInternAtom(dpy, "WM_CLIENT_LEADER", False);
    _XA_WM_COLORMAP_WINDOWS = XInternAtom(dpy, "WM_COLORMAP_WINDOWS", False);
    
    _XA_GNUSTEP_WM_ATTR = XInternAtom(dpy, GNUSTEP_WM_ATTR_NAME, False);
    /* cursors */
    wCursor[WCUR_NORMAL]=XCreateFontCursor(dpy, XC_top_left_arrow);
    wCursor[WCUR_MOVE]=XCreateFontCursor(dpy, XC_fleur);
    wCursor[WCUR_RESIZE]=XCreateFontCursor(dpy, XC_sizing);
    wCursor[WCUR_WAIT]=XCreateFontCursor(dpy, XC_watch);
    wCursor[WCUR_ARROW]=XCreateFontCursor(dpy, XC_left_ptr);
    
    /* right now, only works for the deault screen */
    wScreen = *wScreenInit(DefaultScreen(dpy));
    XSetErrorHandler((XErrorHandler)catchXError);
    
    /* emergency exit... */
    signal(SIGINT, handleSig);
    signal(SIGHUP, handleSig);
    signal(SIGTERM, handleSig);
    signal(SIGQUIT, handleSig);  
    
    signal(SIGCHLD, handleChild);
    /* manage all windows that were already 
     * here before us */
    manageAllWindows(&wScreen);
}

/*
 *-----------------------------------------------------------------------
 * manageAllWindows--
 * 	Manages all windows in the screen.
 * 
 * Notes:
 * 	Called when the wm is being started.
 *----------------------------------------------------------------------- 
 */
static void
manageAllWindows(WScreen *scr)
{
    Window root, parent;
    Window *children;
    unsigned int nchildren;
    XWindowAttributes wattribs;
    unsigned int i;
    
    XGrabServer(dpy);
    XQueryTree(dpy, scr->root_win, &root, &parent, &children, &nchildren);

    /* map all windows without OverrideRedirect */
    for (i=0; i<nchildren; i++) {
	XGetWindowAttributes(dpy, children[i], &wattribs);
	if (!wattribs.override_redirect && (wattribs.map_state!=IsUnmapped)) {
	    XUnmapWindow(dpy, children[i]);
	    wManageWindow(children[i]);
	}
    }
    XUngrabServer(dpy);
    XFree(children);
}



