/*
 *  WindowMaker window manager
 * 
 *  Copyright (c) 1997 Alfredo K. Kojima
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include <stdlib.h>

#include "generic/wwmlib.h"
#include "WindowMaker.h"
#include "wcore.h"
#include "wconfig.h"


/****** Global Variables ******/
/* cursors */
extern Cursor wCursor[WCUR_LAST];

extern XContext wWinContext;


static WMethodTable wNullMethods = {
    NULL,
    NULL,
    NULL
};


/*
 *---------------------------------------------------------------------- 
 * wCoreCreateParent--
 * 	Creates a window whose parent is the root win. Used for
 * window frames (parent of client windows).
 * 	The border width is DEF_BORDER_WIDTH and the color is black
 * Returns:
 * 	A initialized core window structure.
 * 
 * Side effects:
 * 	A window context is saved for the created window.
 * 
 * Notes:
 * 	The event mask is initialized to a default value.
 *---------------------------------------------------------------------- 
 */
WCoreWindow *
wCoreCreateParent(WScreen *screen, int x, int y, int width, int height)
{
    WCoreWindow *core;
    int vmask;
    XSetWindowAttributes attribs;
    
    core=wmalloc(sizeof(WCoreWindow));
    
    vmask = CWBackPixmap|CWBackPixel|CWBorderPixel|CWCursor|CWEventMask;
    attribs.cursor = wCursor[WCUR_DEFAULT];
    attribs.background_pixmap = None;
    attribs.background_pixel = screen->black_pixel;
    attribs.border_pixel = screen->black_pixel;
    attribs.event_mask = SubstructureRedirectMask | ButtonPressMask 
      | ButtonReleaseMask | EnterWindowMask | LeaveWindowMask;
    
    core->window =
      XCreateWindow(dpy, screen->root_win, x, y, width, height, 
		    FRAME_BORDER_WIDTH, screen->w_depth, CopyFromParent, 
		    screen->w_visual, vmask, &attribs);
    core->width=width;
    core->height=height;
    core->screen_ptr = screen;

    core->descriptor.self = core;
    core->descriptor.parent = NULL;
    core->descriptor.parent_type = 0;
    core->descriptor.methods = &wNullMethods;
    core->descriptor.click_timestamp = 0;
    XSaveContext(dpy, core->window, wWinContext, (XPointer)&core->descriptor);
    return core;
}


/*
 *---------------------------------------------------------------------- 
 * wCoreCreateTopLevel--
 * 	Creates a toplevel window used for icons, menus and dialogs.
 * 	Border width is 0 and some attributes are set specially.
 *
 * Returns:
 * 	The created window.
 *---------------------------------------------------------------------- 
 */
WCoreWindow *
wCoreCreateTopLevel(WScreen *screen, int x, int y, int width, int height)
{
    WCoreWindow *core;
    int vmask;
    XSetWindowAttributes attribs;
    
    core=wmalloc(sizeof(WCoreWindow));
    
    vmask = CWBackPixmap|CWBackPixel|CWCursor|CWEventMask|CWSaveUnder;
    attribs.save_under = True;
    attribs.cursor = wCursor[WCUR_ARROW];
    attribs.background_pixmap = None;
    attribs.background_pixel = screen->unfocused_texture->color.pixel;
    attribs.event_mask = SubstructureRedirectMask | ButtonPressMask 
      | ButtonReleaseMask | ExposureMask;
    
    core->window =
      XCreateWindow(dpy, screen->root_win, x, y, width, height, 
		    0, screen->w_depth, CopyFromParent, 
		    screen->w_visual, vmask, &attribs);
    core->width=width;
    core->height=height;
    core->screen_ptr = screen;

    core->descriptor.self = core;
    core->descriptor.parent = NULL;
    core->descriptor.parent_type = 0;
    core->descriptor.methods = NULL;
    core->descriptor.click_timestamp = 0;
    XSaveContext(dpy, core->window, wWinContext, (XPointer)&core->descriptor);
    return core;
}


/*
 *---------------------------------------------------------------------- 
 * wCoreCreate--
 * 	Creates a brand new child window.
 * 	The window will have a border width of 0 and color is black.
 * 
 * Returns:
 * 	A initialized core window structure.
 * 
 * Side effects:
 * 	A window context for the created window is saved.
 * 
 * Notes:
 * 	The event mask is initialized to a default value.
 * 
 *---------------------------------------------------------------------- 
 */
WCoreWindow *
wCoreCreate(WCoreWindow *parent, int x, int y, int width, int height)
{
    WCoreWindow *core;
    int vmask;
    XSetWindowAttributes attribs;
    
    core=wmalloc(sizeof(WCoreWindow));
    
    vmask = CWBackPixmap|CWBackPixel|CWBorderPixel|CWCursor|CWEventMask;
    attribs.cursor = wCursor[WCUR_ARROW];
    attribs.background_pixmap = None;
    attribs.background_pixel = parent->screen_ptr->black_pixel;
    attribs.event_mask = KeyPressMask | ButtonPressMask | ButtonReleaseMask 
      | ButtonMotionMask | ExposureMask;
    
    core->window =
      XCreateWindow(dpy, parent->window, x, y, width, height, 0,
		    parent->screen_ptr->w_depth, CopyFromParent, 
		    parent->screen_ptr->w_visual, vmask, &attribs);
    core->width=width;
    core->height=height;
    core->screen_ptr = parent->screen_ptr;

    core->descriptor.self = core;
    core->descriptor.parent = NULL;
    core->descriptor.parent_type = 0;
    core->descriptor.methods = NULL;
    core->descriptor.click_timestamp = 0;    
    XSaveContext(dpy, core->window, wWinContext, (XPointer)&core->descriptor);
    return core;
}



void
wCoreDestroy(WCoreWindow *core)
{
    XDeleteContext(dpy, core->window, wWinContext);
    XDestroyWindow(dpy, core->window);
    wfree(core);
}


void
wCoreConfigure(WCoreWindow *core, int req_x, int req_y, int req_w, int req_h)
{
    XWindowChanges xwc;
    unsigned int mask;
    
    mask = CWX|CWY;
    xwc.x = req_x;
    xwc.y = req_y;
    
    if (req_w <= 0) 
      req_w = core->width;
    if (req_h <= 0)
      req_h = core->height;
    
    if (req_w != core->width || req_h != core->height) {
	mask |= CWWidth | CWHeight;
	xwc.width = req_w;
	xwc.height = req_h;
	core->width = req_w;
	core->height = req_h;
    }
    XConfigureWindow(dpy, core->window, mask, &xwc);
}
