/*
 *  WindowMaker window manager
 * 
 *  Copyright (c) 1997 Alfredo K. Kojima
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef WINDOWMAKER_H_
#define WINDOWMAKER_H_

#include "wconfig.h"


/* max. number of distinct window levels */
#define MAX_WINDOW_LEVELS	5


/* class codes */
#define WCLASS_UNKNOWN 	0
#define WCLASS_WINDOW	1	       /* managed client windows */
#define WCLASS_MENU	2	       /* root menus */
#define WCLASS_ICON	3
#define WCLASS_DIALOG	4


/*
 * WObjDescriptor will be used by the event dispatcher to
 * send events to a particular object through the methods in the 
 * method table. If all objects of the same class share the
 * same methods, the class method table should be used, otherwise
 * a new method table must be created for each object.
 * It is also assigned to find the parent structure of a given
 * window (like the WWindow or WMenu for a button)
 */
typedef struct WObjDescriptor {
    void *self;			       /* the object that will be called */
    struct WMethodTable *methods;
    
    int parent_type;		       /* type code of the parent */
    void *parent;		       /* parent object (WWindow or WMenu) */
    
    Time click_timestamp;	       /* timestamp of the last click */
} WObjDescriptor;


typedef void (*WEventHandler)(WObjDescriptor *sender, XEvent *event);

/*
 * WMethodTable is a table for callbacks to be called by the event
 * dispatcher. The sender parameter will be set by the dispatcher
 * to the self field of the WEventHandler structure.
 * There are 3 types of method tables:
 * # classwide method tables, which is a table that will be shared
 * among all the objects of a class
 * 	WCore
 * 	WBox
 * # object method tables, where each object has it's own different
 * table.
 * 	-
 * # mixed method tables, each object of the same type points to the
 * same method table. (ex. all close button objects point to the same
 *		       table, as well as iconify etc.)
 * 	WButton
 */

/* DO NOT CHANGE THE ORDER OF THIS. NEW EVENTS SHOULD BE ADDED AT THE END 
 * OF THE TABLE
 */
typedef struct WMethodTable {
    WEventHandler mouse_down;
    WEventHandler click;    
    WEventHandler dbl_click;
} WMethodTable;





/* cursors */
#define WCUR_DEFAULT	0
#define WCUR_NORMAL 	0
#define WCUR_MOVE	1
#define WCUR_RESIZE	2
#define WCUR_WAIT	3
#define WCUR_ARROW	4
#define WCUR_LAST	5


/* geometry displays */
#define WDIS_NEW	0	       /* new style */
#define WDIS_CENTER	1	       /* center of screen */
#define WDIS_TOPLEFT	2	       /* top left corner of screen */
#define WDIS_FRAME_CENTER 3	       /* center of the frame */


/* keyboard focus mode */
#define WKF_CLICK	0
#define WKF_POINTER	1
#define WKF_SLOPPY	2

/* window placement mode */
#define WPM_MANUAL	0
#define WPM_CASCADE	1
#define WPM_AUTO	2


typedef struct WPreferences {
    char size_display;		       /* display type for resize geometry */
    char move_display;		       /* display type for move geometry */
    char window_placement;	       /* window placement mode */
    char colormap_follows_mouse;       /* Colormaps should be installed when
					* the mouse enters a window */
    char focus_mode;		       /* window focusing mode */
    char save_colors;		       /* use dithering colors for bevels
					* etc. */
    char use_dithering;		       /* if dithering should be used for
					* icons and stuff */
    char align_menus;		       /* align menu with their parents */
    
    char floating_transients;	       /* transient windows are floating window
					* level */
    struct {
	short red, green,	       /* size of colormap for dithering */
	  blue;
    } cmap;
    char *icon_back_file;	       /* background image for icons */
    
    struct {
	unsigned int nocpp:1;	       /* don't use cpp */
    } flags;			       /* internal flags */
} WPreferences;

/****** Global Variables  ******/
extern Display	*dpy;
extern char *ProgName;

/****** Global Functions ******/
void wAbort();

#endif
