/*
 *  WindowMaker window manager
 * 
 *  Copyright (c) 1997 Alfredo K. Kojima
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "wconfig.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <string.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include "generic/wwmlib.h"
#include "WindowMaker.h"
#include "actions.h"
#include "menu.h"
#include "funcs.h" 
#include "dialog.h"


extern char *MenuConfigPath;
extern wCursor[WCUR_LAST];



static char **CommandTable=NULL;
static int LastCommand=0;
static int TableSize=0;


/*
 * Syntax:
 * # main menu
 * "Menu Name" MENU
 * 	"Title" EXEC command_to_exec -params
 * 	"Submenu" MENU
 * 		"Title" EXEC command_to_exec -params
 * 	"Submenu" END
 * 	"Workspaces" WORKSPACE_MENU
 * 	"Title" built_in_command
 * "Menu Name" END
 * 
 * Built-in commands:
 * 
 * SHUTDOWN - closes the X server
 * REFRESH - forces the desktop to be repainted
 * EXIT - exit the window manager
 * EXEC <program> - execute an external program
 * WORKSPACE_MENU - places the workspace submenu
 */


/* menu commands */

/* change this to a workspace manager call */
static void
execCommand(WMenu *menu, WMenuEntry *entry)
{
    XGrabPointer(dpy, menu->frame->screen_ptr->root_win, True, 0,
		 GrabModeAsync, GrabModeAsync, None, wCursor[WCUR_WAIT],
		 CurrentTime);
    XSync(dpy, 0);
    if (fork()==0) {
	execl("/bin/sh", "/bin/sh", "-c", CommandTable[entry->code], NULL);
	exit(0);
    }
    XUngrabPointer(dpy, CurrentTime);
    XSync(dpy, 0);
}

static void
exitCommand(WMenu *menu, WMenuEntry *entry)
{
    /* prevent reentrant calls */
    wMenuSetEnabled(menu, entry->order, False);
    if (wMessageDialog(menu->frame->screen_ptr, "Exit", "Exit window manager?",
		       WD_EXIT_CONFIRM)==WDB_EXIT) {
	printf("Exiting WindowMaker.\n");	
	RestoreDesktop(menu->frame->screen_ptr);
	exit(0);
    }
    wMenuSetEnabled(menu, entry->order, True);
}


static void
shutdownCommand(WMenu *menu, WMenuEntry *entry)
{
    /* prevent reentrant calls */
    wMenuSetEnabled(menu, entry->order, False);
    if (wMessageDialog(menu->frame->screen_ptr, "Close X session", 
	       "Close window session?\n(all applications will be closed)",
	       WD_EXIT_CONFIRM)==WDB_EXIT) {
	printf("Closing...\n");	
	WipeDesktop(menu->frame->screen_ptr);
	exit(0);
    }
    wMenuSetEnabled(menu, entry->order, True);
}


static void
refreshCommand(WMenu *menu, WMenuEntry *entry)
{
    wRefreshDesktop(menu->frame->screen_ptr);
}


static void
switchWSCommand(WMenu *menu, WMenuEntry *entry)
{
    wMenuSetEnabled(menu, menu->screen_ptr->workspace+1, True);
    wChangeWorkspace(menu->screen_ptr, entry->code);
    wMenuSetEnabled(menu, menu->screen_ptr->workspace+1, False);
}


static void
newWSCommand(WMenu *menu, WMenuEntry *entry)
{
    char title[MAX_MENU_TITLE_LENGTH+6];
    int ws;

    wMenuSetEnabled(menu, menu->screen_ptr->workspace+1, True);
    ws = wNewWorkspace(menu->screen_ptr);
    if (ws<0) {
	wMenuSetEnabled(menu, menu->screen_ptr->workspace+1, False);
	wMenuSetEnabled(menu, 0, False);
	return;
    }
    if (ws>=MAX_WORKSPACES) {
	wMenuSetEnabled(menu, 0, False);
    }
    sprintf(title, "Switch to %i", ws);
    wMenuAddCallback(menu, title, switchWSCommand, ws);
    wMenuRealize(menu);
    wMenuSetEnabled(menu, menu->screen_ptr->workspace+1, False);    
    wMenuPaint(menu);
}


/*******************************/

/* menu config. parser */

static void
destroyCommands(char **table, int size)
{
    int i;
    
    if (size<=0) return;
    for (i=0; i<size; i++) {
	free(table[i]);
    }
    wfree(table);
}


static int 
registerCommand(char *command)
{
    /* enlarge table */
    if (TableSize>=LastCommand) {
	char **oldtable=CommandTable;
	
	TableSize += 5;
	CommandTable = wrealloc(CommandTable, sizeof(char *)*TableSize);
	if (CommandTable==NULL) {
	    wWarning("out of memory to store menu commands. Nuking it.");
	    destroyCommands(oldtable, LastCommand);
	    return -1;
	}
    }
    
    CommandTable[LastCommand] = strdup(command);
    return LastCommand++;
}



static void
addWorkspaceMenu(WScreen *scr, WMenu *menu, char *title)
{
    char wstitle[MAX_MENU_TITLE_LENGTH+6];
    WMenu *wsmenu;
    int i;
    
    wsmenu = wMenuCreate(scr, "Workspace");
    if (!wsmenu) {
	wWarning("could not create Workspace menu");
	return;
    }
    wMenuAddCallback(wsmenu, "New", newWSCommand, -1);    
    if (scr->last_workspace>=MAX_WORKSPACES) {
	wMenuSetEnabled(wsmenu, 0, False);
    }
    for (i=0; i<=scr->last_workspace; i++) {
	sprintf(wstitle, "Switch to %i", i);
	wMenuAddCallback(wsmenu, wstitle, switchWSCommand, i);
    }
    wMenuSetEnabled(wsmenu, scr->workspace+1, False);
    wMenuAddCascade(menu, title, wsmenu);
}


static WMenu*
parseCascade(WScreen *scr, WMenu *menu, FILE *file)
{
    char linebuf[MAXLINE];
    char title[MAXLINE];
    char command[MAXLINE];
    char params[MAXLINE];
    char *line;
    
    while (!feof(file)) {
	fgets(linebuf, MAXLINE, file);
	line = cropline(linebuf);
	if (line[0]==0 || line[0]=='#' || (line[0]=='/' && line[1]=='/')) 
	  continue;
	
	if (separateline(line, title, command, params) < 2) {
	    wWarning("syntax error in menu config. %s", line);
	    goto error;
	}

	if (strcmp(command, "MENU")==0) {
	    WMenu *cascade;

	    cascade = wMenuCreate(scr, title);
	    if (parseCascade(scr, cascade, file)==NULL) {
		wMenuDestroy(cascade);
	    } else {
		wMenuAddCascade(menu, title, cascade);
	    }
	} else {
	    /* see if the menu has ended */
	    if (strcmp(command, "END")==0) {
		return menu;
	    }
	    /* add new item */
	    if (strcmp(command, "EXEC")==0) {
		int code;

		code = registerCommand(params);
		if (code<0)
		  goto error;
		wMenuAddCallback(menu, title, execCommand, code);

	    } else if (strcmp(command, "EXIT")==0) {
		wMenuAddCallback(menu, title, exitCommand, -1);
		
	    } else if (strcmp(command, "SHUTDOWN")==0) {
		wMenuAddCallback(menu, title, shutdownCommand, -1);
		
	    } else if (strcmp(command, "REFRESH")==0) {
		wMenuAddCallback(menu, title, refreshCommand, -1);
		
	    } else if (strcmp(command, "WORKSPACE_MENU")==0) {
		addWorkspaceMenu(scr, menu, title);
	    } else {
		wWarning("unknown command \"%s\" in menu config.", command);
	    }
	}
    }
    wWarning("syntax error in menu config. END declaration missing");
    return menu;

    error:
    if (menu)
      wMenuDestroy(menu);
    return NULL;
}


static WMenu*
parseMenu(WScreen *scr, FILE *file)
{
    WMenu *menu=NULL;
    char linebuf[MAXLINE];
    char title[MAXLINE];
    char command[MAXLINE];
    char params[MAXLINE];
    char *line;
    
    while (!feof(file)) {
	fgets(linebuf, MAXLINE, file);
	line = cropline(linebuf);
	if (line[0]==0 || line[0]=='#' || (line[0]=='/' && line[1]=='/')) 
	  continue;

	if (separateline(line, title, command, params) < 2) {
	    wWarning("syntax error in menu config. %s", line);
	    return NULL;
	}

	if (strcasecmp(command, "MENU")==0) {
	    menu = wMenuCreate(scr, title);
	    return parseCascade(scr, menu, file);
	} else {
	    wWarning("no title given for the root menu");
	    return NULL;
	}
    }
    return NULL;
}

static WMenu *
makeDefaultMenu(WScreen *scr)
{
    WMenu *menu=NULL;
    
    menu = wMenuCreate(scr, "Commands");
    wMenuAddCallback(menu, "Exit...", exitCommand, -1);
    return menu;
}


/*
 *---------------------------------------------------------------------- 
 * OpenRootMenu--
 * 	Opens the root menu. 
 * 
 * Side effects:
 * 	The command table and the menu may be remade.
 * 
 * Notes:
 * 	If the menu was never mapped or the configuration file was 
 * changed since the last mapping the menu will be reconstructed. 
 * If the menu reconstruction fails, the previous menu will be used.
 *---------------------------------------------------------------------- 
 */
void
OpenRootMenu(WScreen *scr, int x, int y)
{
    WMenu *menu;
    FILE *file=NULL;
    struct stat stats;
    static time_t last_changed=0;
#ifdef USECPP
    char command[MAXLINE], *args;
    int cpp=0;
#endif
    if (stat(MenuConfigPath, &stats)<0) {
	wSysError("could not stat menu configuration file \"%s\"", 
		   MenuConfigPath);
	return;
    }
    /* user changed menu config. Reread menu definition */
    if (scr->root_menu==NULL || stats.st_mtime!=last_changed) {
	char **cmd_table;
	int last_cmd;
	int table_size;

#ifdef DEBUG
	puts("Reparsing menu configs");
#endif	
	last_changed = stats.st_mtime;

#ifdef USECPP
	args = MakeCPPArgs();
	if (!args) {
	    wWarning("could not make arguments for menu file preprocessor");
	} else {
	    sprintf(command, "%s %s %s", CPP_PATH, args, MenuConfigPath);
	    wfree(args);
	    file = popen(command, "r");
	    if (!file) {
		wSysError("could not open menu configuration preprocessor");
	    } else {
		cpp = 1;
	    }
	}
#endif /* USECPP */
	if (!file) {
	    file = fopen(MenuConfigPath, "r");
	    if (file == NULL) {
		wSysError("could not open menu configuration file \"%s\"", 
			  MenuConfigPath);
		return;
	    }
	}
	/* save the current command table */
	cmd_table = CommandTable;
	last_cmd = LastCommand;
	table_size = TableSize;
	/* reset commands */
	CommandTable = NULL;
	LastCommand = 0;
	TableSize = 0;

	menu = parseMenu(scr, file);
#ifdef CPP
	if (cpp) {
	    pclose(file):
	} else {
	    fclose(file);
	}
#else
	fclose(file);
#endif
	if (menu!=NULL) {
	    if (scr->root_menu) {
		wMenuDestroy(scr->root_menu);
		/* destroy the old command table */
		destroyCommands(cmd_table, last_cmd);
	    }
	    scr->root_menu = menu;
	} else {
	    wMessageDialog(scr, "Error", "Could not read menu configuration\n"\
			"(see console log for details)", WD_INFORMATION);
	    /* restore the old command table */
	    destroyCommands(CommandTable, LastCommand);
	    CommandTable = cmd_table;
	    LastCommand = last_cmd;
	    TableSize = table_size;
	}
    }
    menu = scr->root_menu;
    if (!menu) {
	menu = makeDefaultMenu(scr);
	scr->root_menu = menu;
    }

    if (menu) {
	if (menu->flags.mapped) {
	    if (!menu->flags.buttoned)
	      wMenuUnmap(menu);
	 } else {
	     if (!menu->flags.realized)
	       wMenuRealize(menu);
	     wMenuMapAt(menu, x-menu->frame->width/2, 
			y-menu->titlebar->height-menu->entry_height/2);
	 }
    }
}


