/*
 *  WindowMaker window manager
 * 
 *  Copyright (c) 1997 Alfredo K. Kojima
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "wconfig.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <sys/wait.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Intrinsic.h>
#include <X11/cursorfont.h>
#include <X11/Xproto.h>

#include "generic/wwmlib.h"
#include "WindowMaker.h"
#include "GNUstep.h"
#include "texture.h"
#include "screen.h"
#include "window.h"
#include "actions.h"
#include "client.h"

/****** Global Variables ******/

extern WPreferences wPreferences;

/* currently only single screen support */
extern WScreen wScreen;
/* modules */
extern int ModuleNo;

/* contexts */
extern XContext wWinContext;

/* atoms */
extern Atom _XA_WM_STATE;
extern Atom _XA_WM_CHANGE_STATE;
extern Atom _XA_WM_PROTOCOLS;
extern Atom _XA_WM_TAKE_FOCUS;
extern Atom _XA_WM_DELETE_WINDOW;
extern Atom _XA_WM_SAVE_YOURSELF;
extern Atom _XA_WM_CLIENT_LEADER;
extern Atom _XA_WM_COLORMAP_WINDOWS;

extern Atom _XA_GNUSTEP_WM_ATTR;

extern Atom _XA_WINDOWMAKER_MENU;
extern Atom _XA_WINDOWMAKER_WM_PROTOCOLS;

/* cursors */
extern Cursor wCursor[WCUR_LAST];

static void manageAllWindows();



static int 
catchXError(Display *dpy, XErrorEvent *error)
{
    char buffer[MAXLINE];
    
    /* ignore some errors */
    if (error->resourceid != None 
	&& ((error->error_code == BadDrawable 
	     && error->request_code == X_GetGeometry)
	    || (error->error_code == BadWindow
		&& (error->request_code == X_SetInputFocus
		    || error->request_code == X_GetWindowAttributes
		    || error->request_code == X_ChangeWindowAttributes
		    || error->request_code == X_GetProperty
		    || error->request_code == X_QueryTree
		    || error->request_code == X_ConfigureWindow)))) {
#ifndef DEBUG
	return 0;
#else
	printf("got X error %x %x %x\n", error->request_code,
	       error->error_code, (unsigned)error->resourceid);
	return 0;
#endif
    }
    wFormatXError(dpy, error, buffer, MAXLINE); 
    wWarning("internal X error: %s\n", buffer);
    return -1;
}


/*
 *----------------------------------------------------------------------
 * handleSig--
 * 	general signal handler. Exits the program gently.
 *---------------------------------------------------------------------- 
 */
static void 
handleSig(int signal)
{
    wFatal("got signal %i\n", signal);

    wAbort();
    exit(1);
}


static void
buryChild(int foo)
{
    /* R.I.P. */
    waitpid(-1, NULL, WNOHANG);
}

       
/*
 *---------------------------------------------------------- 
 * StartUp--
 * 	starts the window manager and setup global data
 * 
 * display - display to run on
 * 
 * Side effects:
 * global data declared in main.c is initialized
 *----------------------------------------------------------
 */
void 
StartUp()
{
    struct sigaction sigchld_action;
    
    ModuleNo = 1; /* 0 is ourself */

    wWinContext = XUniqueContext();
    _XA_WM_STATE = XInternAtom(dpy, "WM_STATE", False);
    _XA_WM_CHANGE_STATE = XInternAtom(dpy, "WM_CHANGE_STATE", False);
    _XA_WM_PROTOCOLS = XInternAtom(dpy, "WM_PROTOCOLS", False);
    _XA_WM_TAKE_FOCUS = XInternAtom(dpy, "WM_TAKE_FOCUS", False);
    _XA_WM_DELETE_WINDOW = XInternAtom(dpy, "WM_DELETE_WINDOW", False);
    _XA_WM_SAVE_YOURSELF = XInternAtom(dpy, "WM_SAVE_YOURSELF", False);    
    _XA_WM_CLIENT_LEADER = XInternAtom(dpy, "WM_CLIENT_LEADER", False);
    _XA_WM_COLORMAP_WINDOWS = XInternAtom(dpy, "WM_COLORMAP_WINDOWS", False);
    
    _XA_GNUSTEP_WM_ATTR = XInternAtom(dpy, GNUSTEP_WM_ATTR_NAME, False);
    
    _XA_WINDOWMAKER_MENU = XInternAtom(dpy, "_WINDOWMAKER_MENU", False);
    _XA_WINDOWMAKER_WM_PROTOCOLS = XInternAtom(dpy,"_WINDOWMAKER_WM_PROTOCOLS",
					       False);
    
    /* cursors */
    wCursor[WCUR_NORMAL]=XCreateFontCursor(dpy, XC_left_ptr);    
    wCursor[WCUR_ARROW]=XCreateFontCursor(dpy, XC_top_left_arrow);
    wCursor[WCUR_MOVE]=XCreateFontCursor(dpy, XC_fleur);
    wCursor[WCUR_RESIZE]=XCreateFontCursor(dpy, XC_sizing);
    wCursor[WCUR_WAIT]=XCreateFontCursor(dpy, XC_watch);
    
    /* right now, only works for the deault screen */
    wScreen = *wScreenInit(DefaultScreen(dpy));
    XSetErrorHandler((XErrorHandler)catchXError);
    
    /* emergency exit... */
    signal(SIGINT, handleSig);
    signal(SIGHUP, handleSig);
    signal(SIGTERM, handleSig);
    signal(SIGQUIT, handleSig);  
    
    /* handle dead children */
    sigchld_action.sa_handler = buryChild;
    sigemptyset(&sigchld_action.sa_mask);
    sigchld_action.sa_flags = SA_NOCLDSTOP;
    sigaction(SIGCHLD, &sigchld_action, NULL);

    /* manage all windows that were already 
     * here before us */
    manageAllWindows(&wScreen);
}



static int
getState(Window window)
{
    Atom type;
    int form;
    unsigned long nitems, bytes_rem;
    unsigned char *data;

    if (XGetWindowProperty(dpy, window, _XA_WM_STATE, 0, 3, False,
			   _XA_WM_STATE, &type,&form,&nitems,&bytes_rem,
			   &data)==Success) {
      if (data != NULL)
	  return *(long *)data;
    }
    return 0;
}


/*
 *-----------------------------------------------------------------------
 * manageAllWindows--
 * 	Manages all windows in the screen.
 * 
 * Notes:
 * 	Called when the wm is being started.
 *----------------------------------------------------------------------- 
 */
static void
manageAllWindows(WScreen *scr)
{
    Window root, parent;
    Window *children;
    unsigned int nchildren;
    XWindowAttributes wattribs;
    unsigned int i;
    int iconic;
    WWindow *wwin;
    
    XGrabServer(dpy);
    XQueryTree(dpy, scr->root_win, &root, &parent, &children, &nchildren);

    scr->flags.startup = 1;
    /* map all windows without OverrideRedirect */
    for (i=0; i<nchildren; i++) {
	XGetWindowAttributes(dpy, children[i], &wattribs);
	
	iconic = (getState(children[i])==IconicState);
	if (!wattribs.override_redirect 
		   && (wattribs.map_state!=IsUnmapped)) {
	    XUnmapWindow(dpy, children[i]);
	    wwin = wManageWindow(scr, children[i]);
	    if (wwin) {
		if (iconic) {
		    wIconifyWindow(wwin);
		    wwin->flags.ignore_next_unmap=1;
		} else
		  wClientSetState(wwin, NormalState, None);
	    }
	}
    }
    scr->flags.startup = 0;
    wSetFocusTo(scr, NULL);
    XUngrabServer(dpy);
    XFree(children);
}



