/* window.c - client window managing class
 * 
 *  WindowMaker window manager
 * 
 *  Copyright (c) 1997 Alfredo K. Kojima
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "wconfig.h"

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#ifdef SHAPE
#include <X11/extensions/shape.h>
#endif
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "generic/wwmlib.h"
#include "WindowMaker.h"
#include "GNUstep.h"
#include "wcore.h"
#include "window.h"
#include "icon.h"
#include "properties.h"
#include "actions.h"
#include "client.h"
#include "resource.h"
#include "funcs.h"

/****** Global Variables ******/

/* modules */
extern int ModuleNo;

/* contexts */
extern XContext wWinContext;

/* cursors */
extern Cursor wCursor[WCUR_LAST];

/* protocol atoms */
extern Atom _XA_WM_DELETE_WINDOW;

extern WPreferences wPreferences;

/***** Local Variables *****/



/* local functions */
static FocusMode getFocusMode(WWindow *wwin);
static void putWindowBorders(WWindow *wwin, int width, int height, 
			     int *child_x, int *child_y);
static void setupGrabs(WWindow *wwin);
static void removeGrabs(WWindow *wwin);
static void setupResources(WWindow *wwin, WResource *resource);
static void addToStackList(WWindow *wwin);


/* event handlers */

/* titlebar */
static void titlebarMouseDown(WObjDescriptor *desc, XEvent *event);
static void titlebarDblClick(WObjDescriptor *desc, XEvent *event);

static WMethodTable wWindowTitlebarMethods = {
    (WEventHandler)titlebarMouseDown,
      NULL,
      (WEventHandler)titlebarDblClick
};

/* handle */
static void handleMouseDown(WObjDescriptor *desc, XEvent *event);

static WMethodTable wWindowHandleMethods = {
    (WEventHandler)handleMouseDown,
      NULL,
      NULL
};



/* frame window (during window grabs) */
static void frameMouseDown(WObjDescriptor *desc, XEvent *event);

static WMethodTable wWindowFrameMethods = {
    (WEventHandler)frameMouseDown,
      NULL,
      NULL
};


/* close button */
static void windowCloseClick(WObjDescriptor *desc, XEvent *event);
static void windowCloseDblClick(WObjDescriptor *desc, XEvent *event);

static WMethodTable wWindowCloseMethods;

/* iconify button */
static void windowIconifyClick(WObjDescriptor *desc, XEvent *event);

static WMethodTable wWindowIconifyMethods;


/*
 *----------------------------------------------------------------------
 * wWindowInitialize--
 * 	Initialize window "class".
 * 
 * Notes:
 * 	Method table initialization to user defined bindings is done 
 * here.
 *---------------------------------------------------------------------- 
 */
void
wWindowInitialize()
{
    memcpy(&wWindowCloseMethods, &wButtonMethods, sizeof(WMethodTable));
    wWindowCloseMethods.click = windowCloseClick;
    wWindowCloseMethods.dbl_click = windowCloseDblClick;    
    
    memcpy(&wWindowIconifyMethods, &wButtonMethods, sizeof(WMethodTable));
    wWindowIconifyMethods.click = windowIconifyClick;
}



WWindow *
wWindowFor(Window window)
{
    WObjDescriptor *desc;

    if (XFindContext(dpy, window, wWinContext, (XPointer*)&desc)==XCNOENT)
      return NULL;
    if (desc->parent_type==WCLASS_WINDOW)
      return desc->parent;
    else
      return NULL;
}




WWindow *
wWindowCreate()
{
    WWindow *wwin;
    
    wwin = wmalloc(sizeof(WWindow));

    memset(wwin, 0, sizeof(WWindow));

    /* setup global decoration defaults */
    wwin->window_flags = wResourceDefault.values.bits;

    wwin->client_descriptor.methods = &wWindowFrameMethods;
    wwin->client_descriptor.parent = wwin;
    wwin->client_descriptor.self = wwin;
    wwin->client_descriptor.parent_type = WCLASS_WINDOW;
    wwin->client_descriptor.click_timestamp = 0;
    wwin->belongs_to = NULL;
    return wwin;
}


void
wWindowDestroy(WWindow *wwin)
{
    int i;

    if (wwin->belongs_to)
      wApplicationRemoveWindow(wwin->belongs_to, wwin);

    if (wwin->title!=NULL)
      XFree(wwin->title);
        
    if (wwin->normal_hints)
      wfree(wwin->normal_hints);
    
    if (wwin->wm_hints)
      XFree(wwin->wm_hints);
    
    if (wwin->wm_instance)
      XFree(wwin->wm_instance);
    
    if (wwin->wm_gnustep_attr)
      XFree(wwin->wm_gnustep_attr);
    
    if (wwin->cmap_windows)
      XFree(wwin->cmap_windows);

    XDeleteContext(dpy, wwin->client_win, wWinContext);

    for (i=0; i<wwin->l_button_no; i++) {
	if (wwin->l_buttons[i]!=NULL) {	    
	    wButtonDestroy(wwin->l_buttons[i]);
	}	
    }    
    for (i=0; i<wwin->r_button_no; i++) {
	if (wwin->r_buttons[i]!=NULL) {
	    wButtonDestroy(wwin->r_buttons[i]);
	}
    }

    if (wwin->titlebar)
	wCoreDestroy(wwin->titlebar);

    if (wwin->handle)
   	wCoreDestroy(wwin->handle);    

    wCoreDestroy(wwin->frame);

    if (wwin->icon) 
      wIconDestroy(wwin->icon);

    wfree(wwin);
}


/*
 *----------------------------------------------------------------
 * wManageWindow--
 * 	reparents the window and allocates a descriptor for it.
 * Window manager hints and other hints are fetched to configure
 * the window decoration attributes and others.
 * 
 * Returns:
 * 	the new window descriptor
 * 
 * Side effects:
 * 	The window is reparented and appropriate notification
 * is done to the client. Input mask for the window is setup.
 * 	The window descriptor is also associated with various window 
 * contexts and inserted in the head of the window list.
 * Event handler contexts are associated for some objects 
 * (buttons, titlebar and handle)
 *---------------------------------------------------------------- 
 */
WWindow *
wManageWindow(WScreen *scr, Window window)
{
    WWindow *wwin;
    int px, py, width, height;
    XWindowAttributes wattribs;    
    XSetWindowAttributes attribs;
    WResource *resource;
    XEvent junk;
    int internal_window = 0;
    extern WWindow *colormap_win;

    /* check if the window was destroyed */
    if (XCheckTypedWindowEvent(dpy, window, DestroyNotify, &junk)) {
	return NULL;
    }
    XGrabServer(dpy);
    wwin = wWindowCreate();
#ifdef DEBUG
    printf("managing window %x\n", (unsigned)window);
#endif

#ifdef SHAPE
    {
	int junk;
	unsigned int ujunk;
	int b_shaped;
	
	XShapeSelectInput(dpy, window, ShapeNotifyMask);
	XShapeQueryExtents(dpy, window, &b_shaped, &junk, &junk, &ujunk, 
			   &ujunk, &junk, &junk, &junk, &ujunk, &ujunk);
	wwin->flags.shaped = b_shaped;
    }
#endif
    PropGetWMClass(window, &wwin->wm_class, &wwin->wm_instance);

    if (wwin->wm_class && strcmp(wwin->wm_class, "WindowMaker")==0) {
	internal_window = 1;
    }
    
    /* setup descriptor */
    wwin->client_win = window;
    wwin->screen_ptr = scr;
    wwin->focused_texture = scr->focused_texture;
    wwin->unfocused_texture = scr->unfocused_texture;

    wwin->title_font = scr->title_font;

    wwin->focused_text_gc = scr->focused_text_gc;
    wwin->unfocused_text_gc = scr->unfocused_text_gc;

    XGetWindowAttributes(dpy, window, &wattribs);
    wwin->old_border_width = wattribs.border_width;

    if (internal_window)
      attribs.event_mask = CLIENT_EVENTS|ExposureMask;
    else
      attribs.event_mask = CLIENT_EVENTS;
    attribs.do_not_propagate_mask = ButtonPressMask | ButtonReleaseMask;
    XChangeWindowAttributes(dpy, window, CWEventMask|CWDontPropagate,
			    &attribs);    
    XSetWindowBorderWidth(dpy, window, 0);

    /* fill in property/hint data */
    if (!XFetchName(dpy, window, &(wwin->title))) {
	wwin->title = NULL;
	wwin->title_width = XTextWidth(wwin->title_font, DEF_WINDOW_TITLE,
				       strlen(DEF_WINDOW_TITLE));
    } else
      wwin->title_width = XTextWidth(wwin->title_font, wwin->title,
				     strlen(wwin->title));

    /* get hints from GNUstep app */
    if (!PropGetGNUstepAttr(window, &wwin->wm_gnustep_attr)) {
	wwin->wm_gnustep_attr=NULL;
    }

    if (PropGetClientLeader(window, &wwin->client_leader)) {
	/* create an application group for the window */
	wwin->belongs_to = wApplicationAddWindowFor(wwin->client_leader, wwin);
    } else {
	wwin->client_leader = None;
    }
    
    wwin->wm_hints = XGetWMHints(dpy, window);
    
    if (wwin->wm_hints)  {
	if (wwin->wm_hints->flags & WindowGroupHint) {
	    if (wwin->wm_hints->window_group==window)
	      wwin->flags.group_leader=1;
	    wwin->group_id = wwin->wm_hints->window_group;

	    /* create an application group for the window */
	    if (!wwin->belongs_to) {
		wwin->belongs_to = wApplicationAddWindowFor(wwin->group_id, 
							    wwin);
		/* create descriptor for the leader */
		if (wwin->group_id!=wwin->client_win
		    && wwin->belongs_to->main_window==NULL) {
		    WWindow *wleader;

		    wleader = wWindowFor(wwin->group_id);
		    if (wleader!=NULL) {
			puts("ADDING LEADER TO GROUP");
			wleader->belongs_to = 
			  wApplicationAddWindowFor(wwin->group_id, wleader);
		    }
		}
	    }
	} else {
	    wwin->group_id = None;
	}

	if (wwin->wm_hints->flags & UrgencyHint)
	  wwin->flags.urgent = 1;
    } else {
	wwin->group_id = None;
    }
    

    PropGetProtocols(window, &wwin->protocols);
    
    if (!PropGetTransientForHint(window, &wwin->transient_for)) {
	wwin->transient_for = None;
    } else {
	if (wwin->transient_for==0 || wwin->transient_for==window) {
	    wwin->transient_for = scr->root_win;
	} else {
	    /* add the transient to the application window list,
	     * if not yet added */
	    if (!wwin->belongs_to && !wwin->belongs_to->main_window) {
		WWindow *wleader;

		wwin->belongs_to = 
		  wApplicationAddWindowFor(wwin->transient_for, wwin);
		/* create descriptor for the leader */
		wleader = wWindowFor(wwin->transient_for);
		if (wleader!=NULL) {
		    wleader->belongs_to = 
		      wApplicationAddWindowFor(wleader->transient_for, 
					       wleader);
		}
	    }
	    
	    /* make it a floating window */
	    if (wPreferences.floating_transients) {
		wwin->window_level = NSFloatingWindowLevel;
	    }
	}
    }

    /* guess the focus mode */
    wwin->focus_mode = getFocusMode(wwin);

    /* get geometry stuff */
    GetNormalHints(wwin, &wattribs, True);

    /* get colormap windows */
    GetColormapWindows(wwin);
    
    /* Setup the decoration/window attributes */
    
    /*
     * Decoration setting is done in this precedence (lower to higher)
     * - use global defaults (done in wWindowCreate())
     * - try to guess some settings
     * - use default in the resource database
     * - use GNUstep window attributes
     */

    if (wwin->protocols.DELETE_WINDOW)
      wwin->window_flags.broken_close = 0;
    else
      wwin->window_flags.broken_close = 1;

    /* transients can't be iconified or maximized */
    if (wwin->transient_for) {
	wwin->window_flags.miniaturizable=0;
#ifdef DEBUG
	printf("%x is transient for %x\n", (unsigned)window, 
	       (unsigned)wwin->transient_for);
#endif	
    }

    /* if the window can't be resized, remove the resize handle */
    if (wwin->normal_hints->flags & (PMinSize|PMaxSize)
	&& (wwin->normal_hints->min_width==wwin->normal_hints->max_width)
	&& (wwin->normal_hints->min_height==wwin->normal_hints->max_height)) {
	wwin->window_flags.resizable=0;
    }

    /* get default window attribute stuff from resource database */
    if ((resource=wResourceGet(wwin->title, wwin->wm_instance, 
			       wwin->wm_class))) {
	setupResources(wwin, resource);	
    }
    
    /* set workspace on which the window starts */
    wwin->workspace = scr->workspace;
    

    /* set GNUstep window attributes */
    if (wwin->wm_gnustep_attr) {
	wwin->window_flags.titlebar = ((wwin->wm_gnustep_attr->window_flags 
					& NSTitledWindowMask)?1:0);
	wwin->window_flags.closable = ((wwin->wm_gnustep_attr->window_flags 
				       & NSClosableWindowMask)?1:0);

	wwin->window_flags.miniaturizable = 
	  ((wwin->wm_gnustep_attr->window_flags 
	    & NSMiniaturizableWindowMask)?1:0);
	
	wwin->window_flags.resizable = ((wwin->wm_gnustep_attr->window_flags 
					& NSResizableWindowMask)?1:0);
	wwin->window_level = wwin->wm_gnustep_attr->window_level;
    } 

    /* setup window geometry */
    width = wwin->client.width;
    height = wwin->client.height;

    wWindowConstrainSize(wwin, &width, &height);
    
    wwin->client.width = width;
    wwin->client.height = height;
#ifdef DEBUG
    printf("Geometry %i,%i %ix%i\n",wwin->client.x, wwin->client.y,
	   width,height);
#endif
    if (!scr->flags.startup && wwin->transient_for==None 
	&& !(wwin->normal_hints->flags & (USPosition|PPosition))
	&& !internal_window) {
	PlaceWindow(wwin, &wwin->frame_x, &wwin->frame_y);
    } else {
	wwin->frame_x = wwin->client.x;
	wwin->frame_y = wwin->client.y;
    }
    
    /* 
     * Start Reparenting 
     */
    wwin->frame = wCoreCreateParent(scr, wwin->frame_x, wwin->frame_y,
				    width, height);
    /* create borders */
    putWindowBorders(wwin, width, height, &px, &py);
    wwin->client.x += px;
    wwin->client.y += py;
    XReparentWindow(dpy, wwin->client_win, wwin->frame->window, px, py);
    XResizeWindow(dpy, wwin->client_win, wwin->client.width, 
		  wwin->client.height);

    /* setup the frame descriptor */
    wwin->frame->descriptor.methods = &wWindowFrameMethods;
    wwin->frame->descriptor.parent = wwin;
    wwin->frame->descriptor.parent_type = WCLASS_WINDOW;

    /* do not save context for our windows, as they may have their own
     * contexts saved */
    if (!internal_window) {
	XSaveContext(dpy, window, wWinContext, 
		     (XPointer)&wwin->client_descriptor);
	/* don't let windows go away if we die */
	XAddToSaveSet(dpy, window);
    } else {
	/* don't let our windows be shaded */
	wwin->window_flags.shadeable = 0;
    }

    if (wwin->workspace == scr->workspace) {
	XLowerWindow(dpy, window);
	XMapSubwindows(dpy, wwin->frame->window);
	XMapWindow(dpy, wwin->frame->window);

	wwin->flags.mapped=1;
    }
    
    scr->window_count++;    
    /* insert window in window stacking list on top of all other windows,
     * taking care of stacking precedence */
    addToStackList(wwin);
    wRaiseWindow(wwin);
    
    
    /* add window at end of focus window list */
    if (!scr->focused_window) {
	wwin->next = NULL;
	wwin->prev = NULL;
	scr->focused_window = wwin;
    } else {
	WWindow *tmp;

	tmp = scr->focused_window;
	while (tmp->prev)
	  tmp = tmp->prev;
	wwin->prev = NULL;
	wwin->next = tmp;
	tmp->prev = wwin;
    }
    if (wwin->workspace == scr->workspace) {
	if (wPreferences.focus_mode==WKF_CLICK)
	  wSetFocusTo(scr, wwin);
    }
    /*
     * Prevent window withdrawal when getting the 
     * unmap notifies generated during reparenting
     */
    wwin->flags.mapped=0;
    XUngrabServer(dpy);
    XSync(dpy, 0);
   InstallWindowColormaps(colormap_win);
    return wwin;
}


/*
 *---------------------------------------------------------------------- 
 * wUnmanageWindow--
 * 	Removes the frame window from a window and destroys all data
 * related to it. The window will be reparented back to the root window
 * if restore is True.
 * 
 * Side effects:
 * 	Everything related to the window is destroyed and the window
 * is removed from the window list. Focus is set to the next on the
 * window list, unless there's a DestroyNotify event for that window.
 *---------------------------------------------------------------------- 
 */
void
wUnmanageWindow(WWindow *wwin, int restore)
{    
    int level = LevelToIndex(wwin->window_level);

    XUnmapWindow(dpy, wwin->frame->window);
    XSync(dpy, 0);

    /* reparent the window back to the root */
    if (restore)
      wClientRestore(wwin);

    /* remove from the window stack list */
    if (wwin->under)
      wwin->under->above = wwin->above;
    if (wwin->above)
      wwin->above->under = wwin->under;
    else /* this was the first window on the list */
      wwin->screen_ptr->stacking_list[level] = wwin->under;

    /* remove from window focus list */
    if (wwin->prev)
      wwin->prev->next = wwin->next;
      
    if (wwin->next)
      wwin->next->prev = wwin->prev;
    else {
	wwin->screen_ptr->focused_window = wwin->prev;

	if (wPreferences.focus_mode==WKF_CLICK)
	  wSetFocusTo(wwin->screen_ptr, wwin->prev);
    }
    
#ifdef DEBUG
    printf("destroying window %x frame %x\n", (unsigned)wwin->client_win,
	   (unsigned)wwin->frame->window);
#endif 
    wwin->screen_ptr->window_count--;
    wwin->screen_ptr->window_level_count[level]--;
    wWindowDestroy(wwin);
    XSync(dpy, 0);
}

/*
 *---------------------------------------------------------------------- 
 * wWindowPaint--
 * 	Paints the requested parts of the window.
 * If repaint is true, the background will be repainted from scratch.
 * 
 * Returns:
 * 	None
 * 
 * Side effects:
 * 	None
 *---------------------------------------------------------------------- 
 */
void
wWindowPaint(WWindow *wwin, int sides, int repaint)
{
    int x, i;
    GC gc, light_gc, dim_gc;

    if (wwin->flags.needs_full_repaint) {
	sides = BORDER_ALL;
	wwin->flags.needs_full_repaint=0;
	repaint = False;
    }
    if (wwin->window_flags.titlebar && (sides & BORDER_TOP)) {
	if (wwin->flags.focused) {
	    wTexturePaint(wwin->titlebar, wwin->focused_texture, WB_TITLEBAR,
			  repaint);
	    gc = wwin->focused_text_gc;
	} else {
	    wTexturePaint(wwin->titlebar, wwin->unfocused_texture, WB_TITLEBAR,
			  repaint);
	    gc = wwin->unfocused_text_gc;
	}
	x = (wwin->titlebar->width - wwin->title_width)/2;
	if (wwin->title)
	  XDrawString(dpy, wwin->titlebar->window, gc, x,
		      wwin->title_font->ascent + 3, wwin->title,
		      strlen(wwin->title));
	else 
	  XDrawString(dpy, wwin->titlebar->window, gc, x,
		      wwin->title_font->ascent + 3, DEF_WINDOW_TITLE,
		      strlen(DEF_WINDOW_TITLE));

	for (i=0; i<wwin->l_button_no; i++) {
	    wButtonPaint(wwin->l_buttons[i], repaint);
	}
	for (i=0; i<wwin->r_button_no; i++) {
	    wButtonPaint(wwin->r_buttons[i], repaint);
	}
    }
    if (wwin->window_flags.resizable && (sides & BORDER_BOTTOM)) {
	Window win;
	int w, h;
	int cw;

	w = wwin->handle->width;
	h = wwin->handle->height;
	cw = wwin->handle_corner_width;
	light_gc = wwin->handle_texture->relief->light_gc;
	dim_gc = wwin->handle_texture->relief->dim_gc;
	win = wwin->handle->window;

	wTexturePaint(wwin->handle, wwin->handle_texture, WB_NONE, repaint);

	XDrawLine(dpy, win, dim_gc, 0, 0, w, 0);
	XDrawLine(dpy, win, light_gc, 0, 1, w, 1);

	XDrawLine(dpy, win, dim_gc, cw, 2, cw, h);
	XDrawLine(dpy, win, light_gc, cw+1, 2, cw+1, h);

	XDrawLine(dpy, win, dim_gc, w-cw-2, 2, w-cw-2, h);
	XDrawLine(dpy, win, light_gc, w-cw-1, 2, w-cw-1, h);
    }
}


     
void
wWindowFocus(WWindow *wwin)
{
#ifdef NEWSTYLE    
    int i;
    WPixmap *pix;
    void *tmp;
#endif
    
    if (wwin->titlebar) {	
#ifdef NEWSTYLE	
	if (wwin->flags.focused) {
	    for (i=0; i<wwin->l_button_no; i++) {
		wwin->l_buttons[i]->texture = wwin->focused_texture;
		pix = wwin->l_buttons[i]->image;
		wwin->l_buttons[i]->image = wwin->l_buttons[i]->image2;
		wwin->l_buttons[i]->image2 = pix;
	    }
	    for (i=0; i<wwin->r_button_no; i++) {
		wwin->r_buttons[i]->texture = wwin->focused_texture;
		pix = wwin->r_buttons[i]->image;
		wwin->r_buttons[i]->image = wwin->r_buttons[i]->image2;
		wwin->r_buttons[i]->image2 = pix;
	    }
	}
#endif
    }    
    if (wwin->flags.focused)
      wWindowPaint(wwin, BORDER_ALL, True);
    else {
	wwin->flags.focused=1;
	wWindowPaint(wwin, BORDER_ALL, False);
    }
    if (wPreferences.focus_mode==WKF_CLICK)
      removeGrabs(wwin);
    /* keep meta-clicks grabs */
    if (!wwin->window_flags.ignore_mouse) {
	XGrabButton(dpy, Button1, Mod1Mask, wwin->client_win, True,
		    ButtonPressMask, GrabModeSync, GrabModeAsync, None, 
		    wCursor[WCUR_ARROW]);
    }
}

void
wWindowUnfocus(WWindow *wwin)
{
#ifdef NEWSTYLE    
    int i;
    WPixmap *pix;
#endif
    if (wwin->titlebar) {
#ifdef NEWSTYLE	
	if (wwin->flags.focused) {
	    for (i=0; i<wwin->l_button_no; i++) {
		wwin->l_buttons[i]->texture = wwin->unfocused_texture;
		pix = wwin->l_buttons[i]->image;
		wwin->l_buttons[i]->image = wwin->l_buttons[i]->image2;
		wwin->l_buttons[i]->image2 = pix;
	    }
	    for (i=0; i<wwin->r_button_no; i++) {
		wwin->r_buttons[i]->texture = wwin->unfocused_texture;
		pix = wwin->r_buttons[i]->image;
		wwin->r_buttons[i]->image = wwin->r_buttons[i]->image2;
		wwin->r_buttons[i]->image2 = pix;
	    }
	}
#endif	
    }
    wwin->flags.focused=0;
    wWindowPaint(wwin, BORDER_ALL, False);
    if (!wwin->window_flags.ignore_mouse) {
	XUngrabButton(dpy, Button1, Mod1Mask, wwin->client_win);
    }
    if (wPreferences.focus_mode==WKF_CLICK)
      setupGrabs(wwin);
}


/*
 *----------------------------------------------------------------------
 * wWindowConstrainSize--
 * 	Constrains size for the client window, taking the maximal size,
 * window resize increments and other size hints into account.
 * 
 * Returns:
 * 	The closest size to what was given, the client window can assume.
 * 
 *---------------------------------------------------------------------- 
 */
void
wWindowConstrainSize(WWindow *wwin, int *nwidth, int *nheight)
{
    XSizeHints *sizeh=wwin->normal_hints;
    int width = *nwidth;
    int height = *nheight;
    int winc = sizeh->width_inc;
    int hinc = sizeh->height_inc;
    int aspw = (sizeh->min_aspect.x + sizeh->max_aspect.x)/2;
    int asph = (sizeh->min_aspect.y + sizeh->max_aspect.y)/2;

    if (width < sizeh->min_width)
      width = sizeh->min_width;
    if (height < sizeh->min_height)
      height = sizeh->min_height;

    if (width > sizeh->max_width) 
      width = sizeh->max_width;
    if (height > sizeh->max_height) 
      height = sizeh->max_height;
    
    if (sizeh->flags & PAspect) {
	/* adjust min aspect ratio */
	if (width * sizeh->min_aspect.x > height * sizeh->min_aspect.y) {
	    if (aspw > asph) {
		height = (aspw/asph) * width;
		width = (asph/aspw) * height;
	    } else {
		width = (asph/aspw) * height;
		height = (aspw/asph) * width;
	    }
	}
	/* adjust max aspect ratio */
	if (width * sizeh->max_aspect.x < height * sizeh->max_aspect.y) {
	    if (aspw > asph) {
		height = (aspw/asph) * width;
		width = (asph/aspw) * height;
	    } else {
		width = (asph/aspw) * height;
		height = (aspw/asph) * width;
	    }
	}	
    }
    width = (((width - sizeh->base_width) / winc) * winc) + sizeh->base_width; 
    height = (((height - sizeh->base_height) / hinc) * hinc)
      + sizeh->base_height;
    
    *nwidth = width;
    *nheight = height;
}

#ifdef NEWSTYLE
static void
adjustButtons(WWindow *wwin)
{
    int i;
    int x, w, bsize;

    w = wwin->frame->width;
    bsize = wwin->titlebar->height;

    if ((bsize * (wwin->l_button_no + wwin->r_button_no) > w/2)) {
	if (!wwin->flags.no_buttons) {
	    wwin->flags.no_buttons=1;
	    for (i=0; i<wwin->r_button_no; i++) {	
		XUnmapWindow(dpy, wwin->r_buttons[i]->core->window);
	    }
	    return;
	}
    } else if (wwin->flags.no_buttons) {
	for (i=0; i<wwin->r_button_no; i++) {	
	    XMapWindow(dpy, wwin->r_buttons[i]->core->window);
	}
    }
    wwin->flags.no_buttons=0;
    x = w - bsize + 1;
    for (i=0; i<wwin->r_button_no; i++) {
	XMoveWindow(dpy, wwin->r_buttons[i]->core->window, x, 0);
	x -= bsize;
    }
}
#else /* !NEWSTYLE */
static void
adjustButtons(WWindow *wwin)
{
    int i;
    int x, y, w, bsize;

    w = wwin->frame->width;
    bsize = wwin->titlebar->height - 7;

    if (bsize * (wwin->l_button_no + wwin->r_button_no) > w/2) {
	if (!wwin->flags.no_buttons) {
	    wwin->flags.no_buttons=1;
	    XUnmapSubwindows(dpy, wwin->titlebar->window);
	}
	return;
    } else if (wwin->flags.no_buttons) {
	XMapSubwindows(dpy, wwin->titlebar->window);
    }
    wwin->flags.no_buttons=0;
    x = w - 3 - bsize;
    y = (wwin->titlebar->height - bsize)/2;
    for (i=0; i<wwin->r_button_no; i++) {
	XMoveWindow(dpy, wwin->r_buttons[i]->core->window, x, y);
	x -= (bsize + 3);
    }
}
#endif /* NEWSTYLE */


/*
 *----------------------------------------------------------------------
 * wWindowConfigure--
 * 	Configures the frame, decorations and client window to the 
 * given geometry. The given geometry is not checked for validity,
 * wWindowConstrainSize() must be used for that. 
 * 	Geometry parameters are for the frame, not the client window.
 * 	The client window receives a ConfigureNotify event, according
 * to what ICCCM 2.0 says.
 * 
 * Returns:
 * 	None
 * 
 * Side effects:
 * 	Window size and position are changed and client window receives
 * a ConfigureNotify event.
 *---------------------------------------------------------------------- 
 */
void 
wWindowConfigure(wwin, req_x, req_y, req_width, req_height)
WWindow *wwin;
int req_x, req_y;		       /* new position of the frame */
int req_width, req_height;	       /* new size of the frame */
{
    int synth_notify = False;
    int resize;

    resize = (req_width!=wwin->frame->width 
	      || req_height!=wwin->frame->height);
    /* 
     * if the window is being moved but not resized then
     * send a synthetic ConfigureNotify
     */
    if ((req_x!=wwin->frame_x || req_y!=wwin->frame_y) && !resize) {
	synth_notify = True;
    }
    
    /* do not let the window titlebar get unreachable */
    /*
    if (req_y<0) 
      req_y = 0;
     */

    if (resize) {
	int cx, cy, cw, ch;

	if (req_width <= 0)
	  req_width = 1;
	if (req_height <= 0)
	  req_height = 1;

	cx = req_x;
	cy = req_y;
	cw = req_width;
	ch = req_height;
	/* resize the titlebar horizontally */
	if (wwin->window_flags.titlebar) {
#ifdef NEWSTYLE
	    int ls = (wwin->titlebar->height-1)*wwin->l_button_no;
	    int rs = (wwin->titlebar->height-1)*wwin->r_button_no;
	    wCoreConfigure(wwin->titlebar, ls, 0, req_width-ls-rs, 0);
#else	    
	    wCoreConfigure(wwin->titlebar, 0, 0, req_width, 0);
#endif
	    ch -= wwin->top_width;
	    cy += wwin->top_width;
	}
	/* resize the handle horizontally and move it */
	if (wwin->window_flags.resizable) {
	    ch -= wwin->handle->height;
	    wCoreConfigure(wwin->handle, 0, wwin->top_width + ch,
			   req_width, 0);
	}
	wwin->client.x = cx;
	wwin->client.y = cy;
	wwin->client.width = cw;
	wwin->client.height = ch;
	XMoveResizeWindow(dpy, wwin->client_win, 0, wwin->top_width, cw, ch);
	wCoreConfigure(wwin->frame, req_x, req_y, req_width, req_height);
	if (wwin->window_flags.titlebar)
	  adjustButtons(wwin);	
#ifdef SHAPE
	if (wwin->flags.shaped)
	  wWindowSetShape(wwin);
#endif
	XSync(dpy, 0);
    } else {
	int cx, cy;
	
	cx = req_x;
	cy = req_y;
	if (wwin->window_flags.titlebar) {
	    cy += wwin->top_width;
	}
	wCoreConfigure(wwin->frame, req_x, req_y, 0, 0);
	wwin->client.x = cx;
	wwin->client.y = cy;
    }
    wwin->frame_x = req_x;
    wwin->frame_y = req_y;
    if (synth_notify) {
	XEvent sevent;
	
	sevent.type = ConfigureNotify;
	sevent.xconfigure.display = dpy;
	sevent.xconfigure.event = wwin->client_win;
	sevent.xconfigure.window = wwin->client_win;

	sevent.xconfigure.x = wwin->client.x;
	sevent.xconfigure.y = wwin->client.y;
	sevent.xconfigure.width = wwin->client.width;
	sevent.xconfigure.height = wwin->client.height;
 
	sevent.xconfigure.border_width = wwin->old_border_width;
	if (wwin->titlebar)
	  sevent.xconfigure.above = wwin->titlebar->window;
	else
	  sevent.xconfigure.above = None;
	sevent.xconfigure.override_redirect = False;
	XSendEvent(dpy, wwin->client_win, False, StructureNotifyMask, &sevent);
    }
    if (req_width < (HANDLE_CORNER_WIDTH*2 + HANDLE_MIN_WIDTH)) {
	wwin->handle_corner_width = (req_width - HANDLE_MIN_WIDTH)/2;
	if (wwin->handle_corner_width < HANDLE_MIN_WIDTH)
	  wwin->handle_corner_width = HANDLE_MIN_WIDTH/2;
    } else {
	wwin->handle_corner_width = HANDLE_CORNER_WIDTH;
    }
}


void
wWindowConfigureBorders(WWindow *wwin)
{
    int resized=0;
    int oy, ox;
    WScreen *scr = wwin->screen_ptr;
    int bsize;
    int new_height = wwin->frame->height;
    int width=wwin->client.width, height=wwin->client.height;
    int i;
    
    wClientGetGravityOffsets(wwin, &ox, &oy);
    
    /* titlebar and buttons */
    for (i=0; i<wwin->l_button_no; i++) {
	if (wwin->l_buttons[i]!=NULL) {	    
	    wButtonDestroy(wwin->l_buttons[i]);
	}
    }
    wwin->l_button_no=0;
    for (i=0; i<wwin->r_button_no; i++) {
	if (wwin->r_buttons[i]!=NULL) {
	    wButtonDestroy(wwin->r_buttons[i]);
	}
    }
    wwin->r_button_no=0;
    
    wwin->top_width = 0;
    if (wwin->window_flags.titlebar) {
	WButton *wbut=NULL;
	int x, w, h;
	
	h = wwin->title_font->ascent + wwin->title_font->descent + 8;
	w = width;
	if (!wwin->titlebar) {
	    resized = 1;
	    wwin->titlebar = wCoreCreate(wwin->frame, 0, 0, w, h);
	} else
	  wCoreConfigure(wwin->titlebar, 0, 0, w, h);

	/* setup object descriptor */
	wwin->titlebar->descriptor.parent = wwin;
	wwin->titlebar->descriptor.parent_type = WCLASS_WINDOW; 
	wwin->titlebar->descriptor.methods = &wWindowTitlebarMethods;

	new_height += h;
	wwin->top_width += h;

	bsize = h - 7;
	/* put default titlebar buttons on left */
	if (wwin->window_flags.miniaturizable) {
#ifdef NEWSTYLE
	    x = 0;
	    wbut = wButtonCreate(wwin->frame, x, 0, h-1, h);
#else
	    x = 3;
	    wbut = wButtonCreate(wwin->titlebar, x, (h-bsize)/2, bsize, bsize);
#endif
	    /* setup object descriptor */
	    wbut->core->descriptor.parent = wwin;
	    wbut->core->descriptor.parent_type = WCLASS_WINDOW;
	    wbut->core->descriptor.methods = &wWindowIconifyMethods;
	    wbut->image = scr->b_pixmaps[WBUT_ICONIFY];
#ifdef NEWSTYLE
	    wbut->image2 = scr->b_pixmaps[WBUT_LAST+WBUT_ICONIFY];
#endif
	    wwin->l_buttons[wwin->l_button_no++] = wbut;
	}

	/* put default titlebar buttons on right */
	if (wwin->window_flags.closable) {
#ifdef NEWSTYLE
	    x = width - h + 1;
	    wbut = wButtonCreate(wwin->frame, x, 0, h-1, h);
#else
	    x = width - 3 - bsize;
	    wbut = wButtonCreate(wwin->titlebar, x, (h-bsize)/2, bsize, bsize);
#endif
	    /* setup object descriptor */
	    wbut->core->descriptor.parent = wwin;
	    wbut->core->descriptor.parent_type = WCLASS_WINDOW;
	    wbut->core->descriptor.methods = &wWindowCloseMethods;
	    if (wwin->window_flags.broken_close) {
		wbut->image = scr->b_pixmaps[WBUT_BROKENCLOSE];
#ifdef NEWSTYLE
		wbut->image2 = scr->b_pixmaps[WBUT_LAST+WBUT_BROKENCLOSE];
#endif
	    } else {
		wbut->image = scr->b_pixmaps[WBUT_CLOSE];
#ifdef NEWSTYLE
		wbut->image2 = scr->b_pixmaps[WBUT_LAST+WBUT_CLOSE];
#endif
	    }
	    wwin->r_buttons[wwin->r_button_no++] = wbut;
	}
	XMapSubwindows(dpy, wwin->titlebar->window);
    } else {
	/* destroy titlebar */
	if (wwin->titlebar) {
	    resized = 1;
	    wCoreDestroy(wwin->titlebar);
	    wwin->titlebar=NULL;
	}
    }

    /* resize handle */
    
    wwin->bottom_width = 0;    
    if (wwin->window_flags.resizable) {
	if (!wwin->handle) {
	    resized = 1;
	    wwin->handle = wCoreCreate(wwin->frame, 0, wwin->top_width+height, 
				      width, HANDLE_HEIGHT);
	} else
	  wCoreConfigure(wwin->handle, 0, wwin->top_width+height, width, 
			 HANDLE_HEIGHT);
	/* setup object descriptor */
	wwin->handle->descriptor.parent = wwin;
	wwin->handle->descriptor.parent_type = WCLASS_WINDOW;
	wwin->handle->descriptor.methods = &wWindowHandleMethods;
	wwin->handle_texture = wwin->unfocused_texture;
	wwin->bottom_width += HANDLE_HEIGHT;
	new_height += HANDLE_HEIGHT;
	
	wwin->handle_corner_width = HANDLE_CORNER_WIDTH;
	if (width < HANDLE_CORNER_WIDTH*2 + HANDLE_MIN_WIDTH) {
	    wwin->handle_corner_width = (width - HANDLE_MIN_WIDTH)/2;
	    if (wwin->handle_corner_width < 0)
	      wwin->handle_corner_width = 0;
	}
    } else {
	/* remove handle */
	if (wwin->handle) {
	    resized = 1;
	    wCoreDestroy(wwin->handle);
	    wwin->handle=NULL;
	}
    }

    wwin->frame_x += ox * (wwin->old_border_width - FRAME_BORDER_WIDTH);
    wwin->frame_y += oy * (wwin->old_border_width - FRAME_BORDER_WIDTH);
    if (resized) {
	wWindowConfigure(wwin, wwin->frame_x, wwin->frame_y,
			 wwin->frame->width, new_height);
    }
    if (wwin->flags.focused)
      wWindowFocus(wwin);
    else
      wWindowUnfocus(wwin);
    wwin->flags.needs_full_repaint=1;
}


#ifdef SHAPE
void wWindowSetShape(WWindow *wwin)
{
    XRectangle rect[2];
    int count=0;

    XShapeCombineShape(dpy, wwin->frame->window, ShapeBounding,
		       0, wwin->top_width, wwin->client_win, 
		       ShapeBounding, ShapeSet);
    if (wwin->titlebar) {
	rect[count].x = 0;
	rect[count].y = 0;
	rect[count].width = wwin->frame->width;
	rect[count].height = wwin->top_width;
	count++;
    }
    if (wwin->handle) {
	rect[count].x = 0;
	rect[count].y = wwin->frame->height-wwin->bottom_width;
	rect[count].width = wwin->handle->width;
	rect[count].height = wwin->bottom_width;
	count++;
    }
    XShapeCombineRectangles(dpy,wwin->frame->window,ShapeBounding,
			    0,0,rect,count,ShapeUnion,Unsorted);
}
#endif

/* ====================================================================== */

static FocusMode
getFocusMode(WWindow *wwin)
{    
    FocusMode mode;
    
    if ((wwin->wm_hints) && (wwin->wm_hints->flags & InputHint)) {
	if (wwin->wm_hints->input == True) {
	    if (wwin->protocols.TAKE_FOCUS)
	      mode = WFM_LOCALLY_ACTIVE;
	    else
	      mode = WFM_PASSIVE;
	} else {
	    if (wwin->protocols.TAKE_FOCUS)
	      mode = WFM_GLOBALLY_ACTIVE;
	    else
	      mode = WFM_NO_INPUT;
	}
    } else {
	mode = WFM_PASSIVE;
    }
    return mode;
}


static void
setupGrabs(WWindow *wwin)
{
    /* setup passive grabs on the client window */
    XGrabButton(dpy, Button1, AnyModifier, wwin->client_win, True,
		ButtonPressMask, GrabModeSync, GrabModeAsync, None, None);
    XGrabButton(dpy, Button2, AnyModifier, wwin->client_win, True,
		ButtonPressMask, GrabModeSync, GrabModeAsync, None, None);
    XGrabButton(dpy, Button3, AnyModifier, wwin->client_win, True,
		ButtonPressMask, GrabModeSync, GrabModeAsync, None, None);
    XSync(dpy, 0);
}

static void
removeGrabs(WWindow *wwin)
{
    /* remove passive grabs on the client window */
    XUngrabButton(dpy, Button1, AnyModifier, wwin->client_win);
    XUngrabButton(dpy, Button2, AnyModifier, wwin->client_win);
    XUngrabButton(dpy, Button3, AnyModifier, wwin->client_win);    
    XSync(dpy, 0);
}



static void
setupResources(WWindow *wwin, WResource *resource)
{
    wwin->icon_file = resource->icon_file;
    wwin->icon_mask_file = resource->icon_mask_file;

    if (resource->valid.bits.titlebar)
      wwin->window_flags.titlebar=resource->values.bits.titlebar;
    
    if (resource->valid.bits.resizable)
      wwin->window_flags.resizable=resource->values.bits.resizable;
    
    if (resource->valid.bits.closable)
      wwin->window_flags.closable=resource->values.bits.closable;

    if (resource->valid.bits.broken_close)
      wwin->window_flags.broken_close=resource->values.bits.broken_close;
    
    if (resource->valid.bits.miniaturizable)
      wwin->window_flags.miniaturizable=resource->values.bits.miniaturizable;
    
    if (resource->valid.bits.shadeable)
      wwin->window_flags.shadeable=resource->values.bits.shadeable;
    
    if (resource->valid.bits.focusable)
      wwin->window_flags.focusable=resource->values.bits.focusable;
    
    if (resource->valid.bits.omnipotent)
      wwin->window_flags.omnipotent=resource->values.bits.omnipotent;
    
    if (resource->valid.bits.skip_window_list)
      wwin->window_flags.skip_window_list=resource->values.bits.skip_window_list;
    
    if (resource->valid.bits.on_top)
      wwin->window_flags.on_top=resource->values.bits.on_top;
    
    if (resource->valid.bits.ignore_keys)
      wwin->window_flags.ignore_keys=resource->values.bits.ignore_keys;
    
    if (resource->valid.bits.ignore_mouse)
      wwin->window_flags.ignore_mouse=resource->values.bits.ignore_mouse;
}
 

/*
 *---------------------------------------------------------------------- 
 * putWindowBorders--
 * 	Puts the titlebar and handle if specified in wwin->window_flags.
 *
 * Returns:
 * 	The position to place the child window inside the frame (child_x,
 *								 child_y)
 * 
 * Side effects:
 * 	The size of the frame is adjusted to accomodate the titlebar and
 * handle. The position of the frame is also adjusted according to the
 * gravity. wwin is associated to the created windows.
 * 
 *---------------------------------------------------------------------- 
 */
static void
putWindowBorders(wwin, width, height, child_x, child_y)
WWindow *wwin;
int width, height;		       /* size of the client window */
int *child_x, *child_y;		       /* returns the position of the client
					* window inside the frame */
{
    *child_x = 0;
    *child_y = 0;
 
    wWindowConfigureBorders(wwin);

    if (wwin->window_flags.titlebar) {
	int h;
	
	h = wwin->title_font->ascent + wwin->title_font->descent + 8;
	*child_y += h;
    }
}


/*
 *----------------------------------------------------------------------
 * addToStackList--
 * 	Inserts the window in the top of the stack window list. The 
 * stacking precedence (window levels and on_top) is kept intact. 
 * 
 * Returns:
 * 	None
 * 
 * Side effects:
 * 	The window is added to the screen window list.
 *---------------------------------------------------------------------- 
 */
static void
addToStackList(WWindow *wwin)
{
    WWindow *prev, *tmpw;
    int index = LevelToIndex(wwin->window_level);

    wwin->screen_ptr->window_level_count[index]++;
    tmpw = wwin->screen_ptr->stacking_list[index];
    if (!tmpw) {
	wwin->screen_ptr->stacking_list[index] = wwin;
	wwin->above = NULL;
	wwin->under = NULL;
	return;
    }
    prev = tmpw;

    if (!wwin->window_flags.on_top) {
	/* must put this window below other on top windows */	
	while (tmpw && tmpw->window_flags.on_top) {
	    prev = tmpw;
	    tmpw = tmpw->under;
	}
	/* add window after our place */
	if (tmpw == wwin->screen_ptr->stacking_list[index]) {
	    /* last window on list (top) */
	    tmpw->above = wwin;
	    wwin->under = tmpw;
	    wwin->above = NULL;
	    wwin->screen_ptr->stacking_list[index] = wwin;
	} else if (tmpw == NULL) {
	    /* first window on list (bottom) */
	    prev->under = wwin;
	    wwin->under = NULL;
	    wwin->above = prev;
	} else {
	    /* other cases */
	    wwin->above = tmpw->above;
	    wwin->under = tmpw;
	    tmpw->above->under = wwin;
	    tmpw->above = wwin;
	}
    } else {
	/* put on top of the stacking list */
	wwin->above = NULL;
	wwin->under = tmpw;
	tmpw->above = wwin;
	wwin->screen_ptr->stacking_list[index] = wwin;
    }
    wScreenCommitStacking(wwin->screen_ptr);
}


/* ====================================================================== */

static void 
handleMouseDown(WObjDescriptor *desc, XEvent *event)
{
    WWindow *wwin = desc->parent;
    
    if (event->xbutton.state & Mod1Mask) {
	/* move the window */
	if (XGrabPointer(dpy, wwin->titlebar->window, False, 
			 ButtonMotionMask|ButtonReleaseMask,
			 GrabModeAsync, GrabModeAsync, None, 
			 None, CurrentTime)!=GrabSuccess) {
	    wWarning("pointer grab failed for window move");
	    return;
	}
	wMoveWindow(wwin);
	XUngrabPointer(dpy, CurrentTime);		
    } else {
	if (event->xbutton.button == Button1
	    || event->xbutton.button == Button2) {
	    if (wPreferences.focus_mode==WKF_CLICK)
	      wSetFocusTo(wwin->screen_ptr, wwin);
	}
	if (event->xbutton.button == Button1) {
	    wRaiseWindow(wwin);
	    /* resize the window */
	    if (XGrabPointer(dpy, wwin->titlebar->window, False, 
			     ButtonMotionMask|ButtonReleaseMask,
			     GrabModeAsync, GrabModeAsync, None, 
			     None, CurrentTime)!=GrabSuccess) {
		wWarning("pointer grab failed for window resize");
		return;
	    }
	    wResizeWindow(wwin);
	    XUngrabPointer(dpy, CurrentTime);
	}
    }
}


static void 
titlebarDblClick(WObjDescriptor *desc, XEvent *event)
{
    WWindow *wwin = desc->parent;

    if (event->xbutton.state & ControlMask) {
	/* maximize vertically */
	if (wwin->window_flags.resizable && !wwin->flags.shaded) {
	    if (wwin->flags.maximized) {
		wUnmaximizeWindow(wwin);
	    } else
	      wMaximizeWindow(wwin, False);
	}
    } else if (wwin->window_flags.shadeable 
	       && wwin->window_flags.miniaturizable) {
	/* shade window */
	if (wwin->flags.shaded)
	  wUnshadeWindow(wwin);
	else
	  wShadeWindow(wwin);
    }
}


static void 
titlebarMouseDown(WObjDescriptor *desc, XEvent *event)
{
    WWindow *wwin = desc->parent;
    
    if (wPreferences.focus_mode==WKF_CLICK)
      wSetFocusTo(wwin->screen_ptr, wwin);
    if (event->xbutton.button == Button1
	|| event->xbutton.button == Button2) {
	if (event->xbutton.button == Button1) {
	    if (event->xbutton.state & ShiftMask) {
		wSelectWindow(wwin);
		return;
	    }
	    if (event->xbutton.state & Mod1Mask) {
		wLowerWindow(wwin);
	    } else {
		wRaiseWindow(wwin);
	    }
	}
	/* move the window */
	if (XGrabPointer(dpy, wwin->titlebar->window, False, 
			 ButtonMotionMask|ButtonReleaseMask,
			 GrabModeAsync, GrabModeAsync, None, 
			 None, CurrentTime)!=GrabSuccess) {
	    wWarning("pointer grab failed for window move");
	    return;
	}
	wMoveWindow(wwin);

	XUngrabPointer(dpy, CurrentTime);
    }
}


static void
frameMouseDown(WObjDescriptor *desc, XEvent *event)
{
    WWindow *wwin = desc->parent;

    if (wPreferences.focus_mode==WKF_CLICK)
      wSetFocusTo(wwin->screen_ptr, wwin);
    if (event->xbutton.button == Button1
	|| event->xbutton.button == Button2) {
	if (event->xbutton.button == Button1)
	  wRaiseWindow(wwin);
	if (event->xbutton.state & Mod1Mask) {
	    /* move the window */
	    if (XGrabPointer(dpy, wwin->titlebar->window, False, 
			     ButtonMotionMask|ButtonReleaseMask,
			     GrabModeAsync, GrabModeAsync, None, 
			     None, CurrentTime)!=GrabSuccess) {
		wWarning("pointer grab failed for window move");
		return;
	    }
	    wMoveWindow(wwin);

	    XUngrabPointer(dpy, CurrentTime);
	}
    }
}


static void
windowCloseClick(WObjDescriptor *desc, XEvent *event)
{
    WWindow *wwin=desc->parent;

    /* if control click, kill the client */
    if (event->xkey.state & ControlMask)
      wClientKill(wwin);
    /* send delete message */
    if (wwin->protocols.DELETE_WINDOW) {
	wClientSendProtocol(wwin, _XA_WM_DELETE_WINDOW, CurrentTime);
    }
}


static void
windowCloseDblClick(WObjDescriptor *desc, XEvent *event)
{
    WWindow *wwin=desc->parent;

    /* send delete message */
    if (wwin->protocols.DELETE_WINDOW) {
	wClientSendProtocol(wwin, _XA_WM_DELETE_WINDOW, CurrentTime);
    } else {
	wClientKill(wwin);
    }
}


static void
windowIconifyClick(WObjDescriptor *desc, XEvent *event)
{
    if (event->xbutton.state & Mod1Mask) {
	if (((WWindow*)desc->parent)->belongs_to)
	  wHideApplication(((WWindow*)desc->parent)->belongs_to);
    } else
      wIconifyWindow(desc->parent);
}

