/*
 *  WindowMaker window manager
 * 
 *  Copyright (c) 1997 Alfredo K. Kojima
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "wconfig.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
 
#include <X11/Xlib.h>
#include <X11/Xutil.h>


#include "generic/wwmlib.h"

#include "WindowMaker.h"
#include "window.h"
#include "resource.h"
#include "funcs.h"
#include "texture.h"


#ifdef offsetof
#define wOffset(type, field) ((void *) offsetof(type, field))
#else
#define wOffset(type, field) ((void *)((char *)&((type *)NULL)->field))
#endif 


/* data types */
/* assign directly to variable */
#define TBOOLEAN	2	/* char */
#define TCMAPSIZE	3	/* r/g/b (short, short, short) */
#define TDISPLAY	4	/* type of geometry display window (char) */
#define TSTRING		5	/* char* */
#define TFOCUS		6	/* focus mode */
#define TPLACEMENT	7	/* window placement mode */
/* for window attribute values */
#define TWSTRING	16
#define TWWATTRIB	17	/* window attribute bit field */


typedef struct WOption {
    char type;			       /* option type */
    char *name;			       /* name of option */
    void *ptr;			       /* pointer to data (or offset) */
} WOption;

typedef struct WSection {
    char *name;
    WOption *options;
} WSection;



/*
 * Decoration setting is done in this order:
 * - see if there's a default in the resource database
 * - if the application is GNUstep set the resources asked by the app
 * - if not, try to guess the settings
 * 
 * For icons: (ascending precedence)
 * - use global default
 * - use application specified icon bitmap
 * - use application specified icon window
 * - use database with window class
 * - use database with window instance
 * - use database with window name
 */

static HashTable *wResDatabase=NULL;

extern WResource wResourceDefault;

extern WPreferences wPreferences;

extern char *PixmapPath;


/* window attribute options */
static WOption WindowOptions[] = {
    {TWSTRING, 	"icon", 		wOffset(WResource, icon_file)},
    {TWWATTRIB,	"titlebar", 		(void*)WA_TITLEBAR},
    {TWWATTRIB,	"resizable", 		(void*)WA_RESIZABLE},
    {TWWATTRIB,	"closable", 		(void*)WA_CLOSABLE},
    {TWWATTRIB,	"miniaturizable",	(void*)WA_MINIATURIZABLE},
    {TWWATTRIB,	"shadeable",		(void*)WA_SHADEABLE},
    {TWWATTRIB,	"keepontop",		(void*)WA_ON_TOP},
    {TWWATTRIB,	"skipwindowlist",	(void*)WA_SKIP_WINDOW_LIST},
    {TWWATTRIB,	"omnipotent",		(void*)WA_OMNIPOTENT},
    {TWWATTRIB,	"focusable",		(void*)WA_FOCUSABLE},
    {TWWATTRIB,	"controlkeyboard",	(void*)WA_IGNORE_KEYS},
    {TWWATTRIB,	"controlmouse",		(void*)WA_IGNORE_MOUSE},
    {0,		NULL,			NULL},
};
#undef ofs



/* preference options */
static WOption PreferenceOptions[] = {
    {TBOOLEAN, 	"dithering", 		&wPreferences.use_dithering},
    {TCMAPSIZE,	"dithercolormap", 	&wPreferences.cmap},
    {TSTRING,	"pixmappath",		&PixmapPath},
    {TDISPLAY,	"resizedisplay",	&wPreferences.size_display},
    {TDISPLAY, 	"movedisplay",		&wPreferences.move_display},
    {TBOOLEAN,	"savecolors",		&wPreferences.save_colors},
    {TBOOLEAN, 	"colormapfollowsmouse",	&wPreferences.colormap_follows_mouse},
    {TFOCUS, 	"focusmode",		&wPreferences.focus_mode},
    {TBOOLEAN, 	"alignsubmenus",	&wPreferences.align_menus},
    {TBOOLEAN,  "floatingtransients",	&wPreferences.floating_transients},
    {TPLACEMENT,"windowplacement",	&wPreferences.window_placement},
    {TBOOLEAN,	"ignorefocusclick",	&wPreferences.ignore_focus_click},
    {0,		NULL,			NULL},
};


/* style options */
static WOption StyleOptions[] = {
    {TSTRING, 	"iconbackpixmap",	&wPreferences.icon_back_file},
    {TBOOLEAN, 	"titletextjustify",	&wPreferences.title_at_left},
      
    {TSTRING, 	"windowtitlefont",	&wPreferences.title_font},
    {TSTRING, 	"menutitlefont",	&wPreferences.menu_title_font},
    {TSTRING, 	"menutextfont",		&wPreferences.menu_entry_font},
    {TSTRING, 	"icontitlefont",	&wPreferences.icon_title_font},
    {TSTRING, 	"displayfont",		&wPreferences.info_font},
    {TSTRING, 	"dialogfont",		&wPreferences.panel_title_font},
    {TSTRING, 	"dialogmessagefont",	&wPreferences.panel_text_font},
      
    {TSTRING, 	"ftitlebackcolor",	&wPreferences.ftitle_bcolor},
    {TSTRING, 	"ftitleforecolor",	&wPreferences.ftitle_fcolor},
    {TSTRING, 	"utitlebackcolor",	&wPreferences.utitle_bcolor},
    {TSTRING, 	"utitleforecolor",	&wPreferences.utitle_fcolor},
    {TSTRING, 	"menutitlebackcolor",	&wPreferences.menu_title_bcolor},
    {TSTRING, 	"menutitleforecolor",	&wPreferences.menu_title_fcolor},
    {TSTRING, 	"menutextdisabledcolor",&wPreferences.menu_text_dcolor},
    {TSTRING, 	"menutextbackcolor",	&wPreferences.menu_text_bcolor},
    {TSTRING, 	"menutextforecolor",	&wPreferences.menu_text_fcolor},
      /*
    {TTEXTURE, 	"ftitletexture",	&wPreferences.ftitle_texture},
    {TTEXTURE, 	"utitletexture",	&wPreferences.utitle_texture},
    {TTEXTURE, 	"menutitletexture",	&wPreferences.menu_title_texture},
    {TTEXTURE, 	"menutexttexture",	&wPreferences.menu_text_texture},
       *
    {T2COLORS,	"ftitlegradient",	&wPreferences.ftitle_gradient},
    {T2COLORS,	"utitlegradient",	&wPreferences.utitle_gradient},
    {T2COLORS,	"menutitlegradient",	&wPreferences.menu_title_gradient},
    {T2COLORS,	"menutextgradient",	&wPreferences.menu_text_gradient},
       * 
       */
    {0,		NULL,			NULL},      
};


/* sections */
static WSection Sections[] = {
    {"[preferences]", 	PreferenceOptions},
    {"[style]",		StyleOptions},
    {"[attributes]",	WindowOptions},
    {NULL, 		NULL}
};



static int
separate(char *line, char *left, char *right, int separator)
{
    char *end, *tmp;
    int i;

    end=strchr(line, separator);
    if (!end) return 0;
    i = end-line;
    strncpy(left, line, i); i--;
    while (i>0 && isspace(left[i])) i--;
    left[i+1]=0;    
    tmp = end+1;
    while (*tmp && isspace(*tmp)) tmp++;
    strcpy(right, tmp);
    return 1;
}


static char *
str_upper(char *line)
{
    while (*line) {
	*line=toupper(*line);
	line++;
    }
    return line;
}



static void
setWindowAttribute(char *line, char *path)
{
    WOption *option;
    WResource *res;
    char wname[MAXLINE], value[MAXPATH];
    char *opt;
    int i, w, v, done=0;
    unsigned int data;
    
    i=0;
    w=0;
    v=0;
    while (!done) {
	switch (line[i]) {
	 case '\'':
	    wname[w++]=line[++i];
	    break;
	    
	 case ':':
	    wname[w++]=0;
	    i++;
	    while (isspace(line[i]) && line[i]!=0) i++;
	    while (line[i] && !isspace(line[i])) value[v++]=line[i++];
	    value[v]=0;
	    done = 1;
	    break;

	 case 0:
	    wWarning("%s: syntax error in line \"%s\"", path, line);
	    return;
	    
	 default:
	    wname[w++]=line[i];
	}
	i++;
    }
    
    /* separate the option from the window specification */
    opt = strrchr(wname, '.');
    *opt = 0;
    opt++;
    
    
    if (wname[0]=='*') {
	/* this is a global default */
	res = &wResourceDefault;
    } else {
	res = table_sget(wResDatabase, wname);
	if (!res) {
	    /* create new entry */
	    res = wmalloc(sizeof(WResource));
	    memset(res, 0, sizeof(WResource));
	    table_sput(wResDatabase, wname, res);
	}
    }
    
    for (option=WindowOptions; option->type!=0; option++) {
	if (strcasecmp(option->name, opt)==0)
	  break;
    }
    switch (option->type) {
     case 0:
	wWarning("%s: unknown attribute in line \"%s\"", path, line);
	return;
	
     case TWSTRING: {
	char **path=(char**)res;
	 path+=(int)option->ptr;
	if (*path)
	  free(*path);
	*path = strdup(value);
	break;
     }

     case TWWATTRIB:
	data = (unsigned int)option->ptr;
	res->valid |= data;
	
	str_upper(value);
	if (value[1]==0) {
	    if (value[0]=='Y' || value[0]=='T' || value[0]=='1')
	      i = 1;
	    else if (value[0]=='N' || value[0]=='F' || value[0]=='0')
	      i = 0;
	    else {
		wWarning("%s: syntax error in line \"%s\"", path, line);
		break;
	    }
	} else {
	    if (strcmp(value, "YES")==0 || strcmp(value, "TRUE")==0)
	      i = 1;
	    else if (strcmp(value, "NO")==0 || strcmp(value, "FALSE")==0)
	      i = 0;
	    else {
		wWarning("%s: syntax error in line \"%s\"", path, line);
		break;
	    }
	}
	if (i) {
	    res->value |= data;
	} else {
	    res->value &= ~data;
	}
	break;
    }
}




static void
parseFile(FILE *file, char *path)
{
    char buffer[MAXLINE], left[MAXLINE], right[MAXLINE];
    char *line;
    WOption *section=NULL;
    WOption *option;
    WSection *tmpsec;
    int i, sec=0;
    
    while (!feof(file)) {
	if (!fgets(buffer, MAXLINE, file)) {
	    return;
	}
	/* trim blanks */
	line = cropline(buffer);
	if (*line==0) continue;
	if (*line=='#' || (line[0]=='/' && line[1]=='/')) continue;
	if (*line=='[') {
	    /* change section */
	    for (tmpsec = Sections; tmpsec->name!=NULL; tmpsec++) {
		if (strcasecmp(tmpsec->name, line)==0) {
		    break;
		}
	    }
	    if (!tmpsec->name) {
		wWarning("%s: expected section identifier but got %s", 
			 path, line);
		return;
	    }
	    section = tmpsec->options;
	    if (section==WindowOptions)
	      sec=1;
	    else
	      sec=0;
	    continue;
	}
	/* wait for a section identifier */
	if (!section || !section->name) continue;

	if (sec) {
	    setWindowAttribute(line, path);
	    continue;
	}
	
	if (!separate(line, left, right, '=')) {
	    wWarning("%s: syntax error in line \"%s\"", path, line);
	    continue;
	}
	
	/* search the option */
	for (option = section; option->type!=0; option++) {
	    if (strcasecmp(option->name, left)==0)
	      break;
	}

	/* assign values */
	switch (option->type) {
	 case 0:
	    wWarning("%s: unknown option in line \"%s\"", path, line);
	    break;

	 case TBOOLEAN:
	    str_upper(right);
	    if (right[1]==0) {
		if (right[0]=='Y' || right[0]=='T' || right[0]=='1')
		  i = 1;
		else if (right[0]=='N' || right[0]=='F' || right[0]=='0')
		  i = 0;
		else {
		    wWarning("%s: syntax error in line \"%s\"", path, line);
		    break;
		}
	    } else {
		if (strcmp(right, "YES")==0 || strcmp(right, "TRUE")==0)
		  i = 1;
		else if (strcmp(right, "NO")==0 || strcmp(right, "FALSE")==0)
		  i = 0;
		else {
		    wWarning("%s: syntax error in line \"%s\"", path, line);
		    break;
		}
	    }
	    {
		char *val=option->ptr;
		*val=i;
	    }
	    break;
	    
	 case TSTRING: {
	     char **tmp=option->ptr;
	    if (*tmp) {
		free(*tmp);
	    }
	    *tmp = strdup(right);
	    break;
	 }
	 case TCMAPSIZE: {
	     int r, g, b;
	     struct Colmap {
		 short red, green, blue;
	     } *cmapP=option->ptr;
	     
	     if (sscanf(right, "%d/%d/%d", &r, &g, &b)!=3) {
		 wWarning("%s: syntax error in line \"%s\"", path, line);
		 break;
	     }
	     if (r<2 || g<2 || b<2 || r>256 || g>256 || b>256) {
		 wWarning("%s: invalid colormap size in line \"%s\"", path, 
			  line);
		 break;
	     }
	     cmapP->red = r;
	     cmapP->green = g;
	     cmapP->blue = b;
	     break;
	 }
	    
	 case TDISPLAY:
	    if (strcasecmp(right, "CENTER")==0)
	      i = WDIS_CENTER;
	    else if (strcasecmp(right, "CORNER")==0)
	      i = WDIS_TOPLEFT;
	    else if (strcasecmp(right, "FLOATING")==0)
	      i = WDIS_FRAME_CENTER;
	    else if (strcasecmp(right, "LINE")==0)
	      i = WDIS_NEW;
	    else {
		wWarning("%s: invalid geometry display type %s. Should be "\
			"center, corner, floating or line", path, right);
		break;
	    }
	    {
		char *val=option->ptr;
		*val = i;
	    }
	    break;
	    
	 case TFOCUS:
	    if (strcasecmp(right, "MANUAL")==0 
		|| strcasecmp(right, "CLICKTOFOCUS")==0)
	      i = WKF_CLICK;
	    else if (strcasecmp(right, "AUTO")==0
		     || strcasecmp(right, "FOCUSFOLLOWSMOUSE")==0)
	      i = WKF_POINTER;
	    else if (strcasecmp(right, "SEMIAUTO")==0 
		     || strcasecmp(right, "SLOPPYFOCUS")==0
		     || strcasecmp(right, "SLOPPY")==0)
	      i = WKF_SLOPPY;
	    else {
		wWarning("%s: invalid focus mode %s. Should be manual, auto "\
			"or semiauto", path, right);
		break;
	    }
	    {
		char *val = option->ptr;
		*val = i;
	    }
	    break;
	    
	 case TPLACEMENT:
	    if (strcasecmp(right, "AUTO")==0)
	      i = WPM_AUTO;
	    else if (strcasecmp(right, "CASCADE")==0)
	      i = WPM_CASCADE;
	    else if (strcasecmp(right, "MANUAL")==0)
	      i = WPM_MANUAL;
	    else {
		wWarning("%s: invalid window placement mode %s. Should be"\
			" auto, cascade or manual");
		break;
	    }
	    {
		char *val = option->ptr;
		*val = i;
	    }
	    break;
	    
#ifdef DEBUG
	 default:
	    printf("Invalid option type in parseFile()\n");
#endif
	}
    }
}




static void
resourceRead(char *path)
{
    FILE *file=NULL;
#ifdef USECPP
    char command[MAXLINE];
    char *args = MakeCPPArgs(path);

    if (!args) {
	wWarning("could not make arguments for preference file preprocessor");
    } else {
	sprintf(command, "%s %s %s", CPP_PATH, args, path);
	wfree(args);
	file = popen(command, "r");
	if (!file) {
	    wSysError("could not open preference file preprocessor");
	}
    }
#endif /* USECPP */
    if (!file) {
	file = fopen(path,"r");
	if (!file) {
	    wSysError("could not open configuration file: \"%s\"", path);
	} else {
	    parseFile(file, path);
	}
	fclose(file);
	file = NULL;
    }
#ifdef USECPP
    if (file) {
	parseFile(file, path);
	pclose(file);
    }
#endif
}



void
wResourceInitialize()
{
    char *path;
    
    path=find_file(DEF_CONFIG_PATHS, DEF_PREFERENCES);

    /* preference defaults */
    wPreferences.colormap_follows_mouse = 1;
    
    
    
    memset(&wResourceDefault, 0, sizeof(WResource));
    /* default global defaults */
    wResourceDefault.icon_file = NULL;
    wResourceDefault.valid = ~0;
    
    wResourceDefault.value =(WA_TITLEBAR|WA_RESIZABLE|WA_CLOSABLE
			     |WA_BROKEN_CLOSE|WA_MINIATURIZABLE|WA_SHADEABLE
			     |WA_FOCUSABLE);

    if (path) {
	wResDatabase = wmalloc(sizeof(HashTable));
	table_init(wResDatabase);

	resourceRead(path);
	if (wResourceDefault.icon_file) {
	    char *file;
	    file = wResourceDefault.icon_file;
	    wResourceDefault.icon_file=find_file(PixmapPath, file);
	    if (!wResourceDefault.icon_file) {
		wWarning("could not find default icon \"%s\"", file);
	    }
	    free(file);
	}
	free(path);
    } else
      wWarning("could not find a preference file (%s) in \"%s\"", 
	       DEF_PREFERENCES, DEF_CONFIG_PATHS);
    if (wPreferences.icon_back_file) {
	char *tmp;
	
	tmp = wPreferences.icon_back_file;
	wPreferences.icon_back_file = find_file(PixmapPath, tmp);
	if (!wPreferences.icon_back_file) {
	    wWarning("could not find icon background \"%s\"", tmp);
	}
	free(tmp);
    }
    
}



/*
 *---------------------------------------------------------------------- 
 * wResourceGet--
 * 	Get the attributes set by the user in the resource database 
 * (our own DB, not X's)
 * 
 * Returns:
 * 	The resource structure or
 * 	NULL if it's not defined for the window
 * 
 * Notes:
 * 	Order of precedence is (lower to higher):
 * 		window class
 * 		window instance name
 * 		window name (title)
 * 		window class.instance
 *---------------------------------------------------------------------- 
 */
WResource *
wResourceGet(char *name, char *instance, char *wclass)
{
    WResource *res;
    char tmp[MAXLINE];

    if (!wResDatabase) 
      return NULL;

    if (instance && wclass) {
	strcpy(tmp, wclass);
	strcat(tmp, ".");
	strcat(tmp, instance);
	res = table_sget(wResDatabase, tmp);
	if (res)
	  return res;
    }

    if (name) {
	res = table_sget(wResDatabase, name);
	if (res)
	  return res;
    }

    if (instance) {
	res = table_sget(wResDatabase, instance);
	if (res)
	  return res;
    }

    if (wclass) {
	res = table_sget(wResDatabase, wclass);
	if (res)
	  return res;
    }
    return NULL;
}
