/* setup.c - dynamic setup module interface
 * 
 *  WindowMaker window manager
 * 
 *  Copyright (c) 1997 Alfredo K. Kojima
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */



#include "wconfig.h"
#include "generic/wwmlib.h"

#include <X11/Xlib.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <fcntl.h>
#include <signal.h>

#include "screen.h"
#include "window.h"
#include "setup.h"
#include "version.h"


/**** Global Variables ****/
extern WPreferences wPreferences;


static int
loadFont(WFont *font, char *font_name)
{
    XFontStruct *new;
    
    new = XLoadQueryFont(dpy, font_name);
    if (!new) {
	return 0;
    }
    XFreeFont(dpy, font->font);
    font->font = new;
    font->height = new->ascent+new->descent;
    font->y = new->ascent;
    return 1;
}


static void
realizeMenus(WMenu *menu)
{
    int i;

    wMenuRealize(menu);
    for (i=0; i<menu->entry_no; i++) {
	if (menu->entries[i]->cascade)
	  realizeMenus(menu->entries[i]->cascade);
    }
}


void
wSetupOpen(WScreen *scr, char *prog)
{
    pid_t pid;
    int inp[2], outp[2];

    if (scr->setup_pipe_out) 
      return;

    if (pipe(inp)<0 || pipe(outp)<0) {
	wSysError("could not open pipe for setup panel");
	return;
    }
    scr->setup_pipe_in = inp[0];
    scr->setup_pipe_out = outp[1];

    pid = fork();
    if (pid<0) {
	wSysError("could not fork setup panel process");
	return;
    } else if (pid==0) {
	char *argv[2];
	
	argv[0] = prog;
	argv[1] = NULL;
	close(inp[0]);
	close(outp[1]);
	close(0);
	dup(outp[0]);
	close(1);
	dup(inp[1]);
	if (execvp(prog, argv)<0) {
	    wSysError("could not exec setup panel:%s", prog);
	    exit(1);
	}
    } else {
	close(inp[1]);
	close(outp[0]);
    }
    scr->setup_panel_pid = pid;
}


void
wSetupClose(WScreen *scr)
{
    if (scr->setup_pipe_out) {
	close(scr->setup_pipe_in);
	close(scr->setup_pipe_out);
	scr->setup_pipe_out = 0;
	kill(scr->setup_panel_pid, SIGTERM);
    }
}
	    
void
wSetupResetFont(WScreen *scr, int font, char *font_name)
{
    switch (font) {
     case WDS_WINDOW_TITLE_FONT:
	if (loadFont(&scr->title_font, font_name)) {
	    XSetFont(dpy, scr->unfocused_text_gc, scr->title_font.font->fid);
	    XSetFont(dpy, scr->focused_text_gc, scr->title_font.font->fid);
	    free(wPreferences.title_font);
	    wPreferences.title_font = strdup(font_name);
	}
	
	break;
	
     case WDS_MENU_TITLE_FONT:
	if (loadFont(&scr->menu_title_font, font_name)) {
	    XSetFont(dpy, scr->menu_title_gc, scr->menu_title_font.font->fid);
	    free(wPreferences.menu_title_font);
	    wPreferences.menu_title_font = strdup(font_name);
	}
	break;
	
     case WDS_MENU_ENTRY_FONT:
	if (loadFont(&scr->menu_entry_font, font_name)) {
	    XSetFont(dpy, scr->menu_entry_gc, scr->menu_entry_font.font->fid);
	    XSetFont(dpy, scr->select_menu_gc, scr->menu_entry_font.font->fid);
	    XSetFont(dpy, scr->disabled_menu_entry_gc, 
		     scr->menu_entry_font.font->fid);
	    free(wPreferences.menu_entry_font);
	    wPreferences.menu_entry_font = strdup(font_name);
	}
	break;
    }
    if (font == WDS_MENU_TITLE_FONT || font == WDS_MENU_ENTRY_FONT) {
	realizeMenus(scr->root_menu);
    } else if (font == WDS_WINDOW_TITLE_FONT) {
	WWindow *tmp;
	
	tmp = scr->focused_window;
	while (tmp) {
	    wWindowConfigureBorders(tmp);
	    XMoveWindow(dpy, tmp->client_win, 0, tmp->top_width);
	    tmp = tmp->prev;
	}
    }
}



void
wSetupCheck(WScreen *scr)
{
    int code, count;
    int msgsize;
    char param[MAXLINE], msg[MAXLINE];
    fd_set fds;
    struct timeval timeout;

    if (scr->setup_pipe_out==0)
      return;

    FD_ZERO(&fds);
    FD_SET(scr->setup_pipe_in, &fds);
    timeout.tv_sec = 0;
    timeout.tv_usec = 0;
    if (select(scr->setup_pipe_in+1, &fds, NULL, NULL, &timeout)<=0) {
	return;
    }
    msgsize = read(scr->setup_pipe_in, msg, MAXLINE);
    if (msgsize<0) {
	wSetupClose(scr);
	wSysError("error reading data from setup panel");
	
	return;
    }
    if (msgsize==0) {
#ifdef DEBUG
	printf("read() from setup panel returned 0");
#endif
	wSetupClose(scr);
	return;
    }
    msg[msgsize] = 0;
    count = sscanf(msg, "%i %s", &code, param);

    if (count==1) {
	switch (code) {
	 case WDS_SEND_VERSION:
	    sprintf(msg, "%i.%i.%i\n", MAJ_VERSION, MIN_VERSION, PATCHLEVEL);
	    if (write(scr->setup_pipe_out, msg, strlen(msg))<0) {
		wSetupClose(scr);
		wSysError("error writing data to setup panel");
		return;
	    }
	    break;
	    
	 case WDS_SEND_FONTS: {
	     char *tfont, *mfont, *efont;
	     
	     if (wPreferences.title_font)
	       tfont = wPreferences.title_font;
	     else
	       tfont = DEF_TITLE_FONT;
	     
	     if (wPreferences.menu_title_font)
	       mfont = wPreferences.menu_title_font;
	     else
	       mfont = DEF_MENU_TITLE_FONT;
	     
	     if (wPreferences.menu_entry_font)
	       efont = wPreferences.menu_entry_font;
	     else
	       efont = DEF_MENU_ENTRY_FONT;
	     
	     sprintf(msg, "%i %s\n%i %s\n%i %s\n",
		     WDS_WINDOW_TITLE_FONT, tfont,
		     WDS_MENU_TITLE_FONT, mfont,
		     WDS_MENU_ENTRY_FONT, efont);
	 
	    if (write(scr->setup_pipe_out, msg, strlen(msg))<0) {
		wSetupClose(scr);
		wSysError("error writing data to setup panel");
		return;
	    }
	 }
	    break;
	    
	 case WDS_CLOSEDOWN:
	    wSetupClose(scr);
	    return;

	 default:
	    wWarning("received invalid request from setup panel");
	    return;
	}
    } else if (count!=2) {
	wWarning("got invalid message from setup panel");
	return;
    }
    if (code>=WDS_WINDOW_TITLE_FONT && code<=WDS_MENU_ENTRY_FONT) {
	wSetupResetFont(scr, code, param);
    }
}
