/*
 *  WindowMaker window manager
 * 
 *  Copyright (c) 1997 Alfredo K. Kojima
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "wconfig.h"

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xproto.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>

#include "generic/wwmlib.h"
#include "WindowMaker.h"
#include "wcore.h"
#include "button.h"
#include "menu.h"
#include "actions.h"
#include "funcs.h"



extern Atom _XA_WINDOWMAKER_MENU;
extern Atom _XA_WINDOWMAKER_WM_PROTOCOLS;

static void notifyClient(WMenu *menu, WMenuEntry *entry);


/*
 * Menu specification format (subject to change):
 *
 * Codes:
 * 1 - main-menu definition
 * 2 - submenu definition
 * 3 - command entry def
 * 4 - submenu entry def
 * 
 * Ex.:
 * Header:
 * CARD16 number-of-items
 * CARD16 number-of-menu-definitions
 * 
 * Body:
 * CARD32 command-code
 * CARD32 entry-index (or menu to add as cascade)
 * CARD32 menu-index-pointer (menu to which the entry belongs)
 * CARD8  enabled
 * STRING title
 * SHORTCUT shortcut
 * 
 * STRING {
 *    	CARD8  length of string
 * 	char[length]  characters
 * }
 * 
 * SHORTCUT {
 *    	CARD8  number of keys
 *	KeySym[number] keys      
 * }
 * 
 * 12
 * 3
 * Cmd Idx  MenuPtr Enabled	Text
 * 2 	4   1		1	File
 * 3 	5   4		1	Save 	[keycodes]
 * 3 	6   4		1	Load 	[keycodes]
 * 
 * 2	7   1		1	Style
 * 3	8   7		1	Font
 * 3	9   7		1	Size
 * 
 * 1 	1   0		1	FooEdit
 * 4 	4   1		1	File
 * 4	7   1		1	Style
 * 
 * 
 * FooEdit
 * -------
 * File  --+- File
 *         |  ----
 * 	   +- Save
 * 	   -- Load
 * 
 * Style --+- Style
 * 	   |  -----
 *         +- Font
 * 	   +- Size
 */

typedef struct {
    WWindow *wwin;
    int code;
} EntryData;


WMenu*
wAppMenuGet(WScreen *scr, WWindow *wwin)
{
    Window window = wwin->client_win;
    Atom type;
    char *buffer, *bufstart;
    int fmt;
    unsigned long nitems, bytes_after;
    int i, j, k, l;
    int count, menus, cmd, idx, ptr, len;
    char enabled;
    WMenu **menu, *main=NULL;
    int *menu_ptr, menu_cnt=0;
    EntryData *cdata;
    WMenuEntry *entry;

    if (XGetWindowProperty(dpy, window, _XA_WINDOWMAKER_MENU, 0, 500000L,
			    False, AnyPropertyType, &type, &fmt, &nitems,
			    &bytes_after, (unsigned char**)&buffer)!=Success) {
	return NULL;
    }
    if (!buffer) {
	return NULL;
    }

    bufstart = buffer;
    
    count = (CARD16)*buffer;
    buffer += 2;
    menus = (CARD16)*buffer;
    buffer += 2;
#ifdef DEBUG
    printf("%i items and %i menus\n", count, menus);
#endif
    menu = wmalloc(sizeof(WMenu)*menus);
    menu_ptr = wmalloc(sizeof(int)*menus);
    for (i=0; i<count; i++) {
	cmd = (int)*buffer;	buffer+=4;
	printf("COMMAND %i\n",cmd);
	idx = (int)*buffer;	buffer+=4;
	printf("INDEX %i\n",idx);
	ptr = (int)*buffer;	buffer+=4;
	enabled = (char)*buffer;	buffer++;
	len = (unsigned char)*buffer;	buffer++;
	switch (cmd) {
	 case 1: /* main menu */
	    if (menu_cnt>=menus) {
		wWarning("AppMenu:too many menu definitions for %s", 
			 wwin->title);
		goto error;
	    }
	    main = wMenuCreate(scr, buffer, True);
	    main->flags.app_menu = 1;
	    menu_ptr[menu_cnt] = idx;
	    menu[menu_cnt] = main;
	    menu_cnt++;
	    break;

	 case 2: /* submenu */
	    if (menu_cnt>=menus) {
		wWarning("AppMenu:too many menu definitions for %s",
			 wwin->title);
		goto error;
	    }
	    menu[menu_cnt] = wMenuCreate(scr, buffer, False);
	    menu[menu_cnt]->flags.app_menu = 1;
	    menu_ptr[menu_cnt] = idx;
	    menu_cnt++;
	    break;

	 case 3: /* command entry */
	    k=-1;
	    for (j=0; j<menu_cnt; j++) {
		if (menu_ptr[j]==ptr) {
		    k=j;
		    break;
		}
	    }
	    if (k<0) {
		wWarning("AppMenu:invalid menu reference in %s", wwin->title);
		goto error;
	    }
	    cdata = wmalloc(sizeof(EntryData));
	    cdata->wwin = wwin;
	    cdata->code = idx;
	    entry = wMenuAddCallback(menu[k], buffer, notifyClient, cdata);
	    entry->flags.enabled = enabled;
	    break;

	 case 4: /* submenu entry */
	    k=-1;
	    for (j=0; j<menu_cnt; j++) {
		if (menu_ptr[j]==ptr) {
		    k=j;
		    break;
		}
	    }
	    if (k<0) {
		wWarning("AppMenu:invalid menu reference in %s", wwin->title);
		goto error;
	    }
	    l=-1;
	    for (j=0; j<menu_cnt; j++) {
		if (menu_ptr[j]==idx) {
		    l=j;
		    break;
		}
	    }
	    if (l<0) {
		wWarning("AppMenu:invalid menu reference in %s", wwin->title);
		goto error;
	    }
	    wMenuAddCascade(menu[k], buffer, menu[l]);
	    break;

	 default:
	    wWarning("AppMenu:invalid command %i in %s", cmd, wwin->title);
	    goto error;

	}
	buffer += len; /* skip title */	
	buffer++; /* skip empty keycode */
    }
    

    if (main) {
	wfree(menu);
	wfree(menu_ptr);
	XFree(bufstart);
	return main;
    }
    error:
    for (i=0; i<menu_cnt; i++) {
	wMenuDestroy(menu[i]);
    }
    wfree(menu);
    wfree(menu_ptr);
    XFree(bufstart);
    return NULL;
}


static void
notifyClient(WMenu *menu, WMenuEntry *entry)
{
    XEvent event;
    EntryData *data = entry->clientdata;
    
    event.xclient.type = ClientMessage;
    event.xclient.message_type = _XA_WINDOWMAKER_WM_PROTOCOLS;
    event.xclient.format = 32;
    event.xclient.display = dpy;
    event.xclient.window = data->wwin->client_win;
    event.xclient.data.l[0] = 1; /* SelectNotify */
    event.xclient.data.l[1] = data->code; /* Entry Code */

    XSendEvent(dpy, data->wwin->client_win, False, NoEventMask, &event);
#ifdef DEBUG    
    printf("SEND NOTIFY TO %x\n", (unsigned)data->wwin->client_win);
#endif
}


