/*
 *  WindowMaker window manager
 * 
 *  Copyright (c) 1997 Alfredo K. Kojima
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "wconfig.h"

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <stdlib.h>
#include <stdio.h>

#include "generic/wwmlib.h"
#include "WindowMaker.h"
#include "wcore.h"
#include "pixmap.h"
#include "texture.h"
#include "window.h"
#include "icon.h"
#include "actions.h"
#include "funcs.h"

/**** Global variables ****/
extern char *PixmapPath;

static void iconMouseDown(WObjDescriptor *desc, XEvent *event);
static void iconDblClick(WObjDescriptor *desc, XEvent *event);

WMethodTable wIconMethods = {
    (WEventHandler)iconMouseDown,
      NULL,
      (WEventHandler)iconDblClick
};



extern Cursor wCursor[WCUR_LAST];


inline static void
getSize(Drawable d, unsigned int *w, unsigned int *h, unsigned int *dep)
{
    Window rjunk;
    int xjunk, yjunk;
    unsigned int bjunk;
    
    XGetGeometry(dpy, d, &rjunk, &xjunk, &yjunk, w, h, &bjunk, dep);
}


WIcon*
wIconCreate(WWindow *wwin)
{
    WScreen *scr=wwin->screen_ptr;
    WIcon *icon;
    
    icon = wmalloc(sizeof(WIcon));
    icon->core = wCoreCreateTopLevel(scr, wwin->icon_x, wwin->icon_y, 
				     ICON_WIDTH, ICON_HEIGHT, 0, False);
    icon->core->descriptor.methods = &wIconMethods;
    icon->core->descriptor.parent_type = WCLASS_ICON;
    icon->core->descriptor.parent = icon;
    
    icon->core->stacking = wmalloc(sizeof(WStacking));
    icon->core->stacking->above = NULL;
    icon->core->stacking->under = NULL;
    icon->core->stacking->window_level = NSNormalWindowLevel;
    icon->core->stacking->child_of = NULL;

    icon->owner = wwin;
    icon->group_icon = 0;
    icon->icon_name = NULL;

    icon->icon_win = None;
    icon->client_window = 0;
    icon->client_pixmap = 0;
    icon->our_pixmap = 0;
    icon->pixmap = None;
    icon->mask = None;
    icon->show_title = 1;

    wIconUpdate(icon);

    return icon;
}


void
wIconDestroy(WIcon *icon)
{
    WCoreWindow *core = icon->core;
    WScreen *scr = core->screen_ptr;

    if (icon->client_window) {
	XReparentWindow(dpy, icon->icon_win, scr->root_win, 0, 0);
	XUnmapWindow(dpy, icon->icon_win);
    }
    if (icon->icon_name) 
      XFree(icon->icon_name);

#ifdef BUGGYSTACKING
    if (core->stacking->above)
      core->stacking->above->stacking->under = core->stacking->under;
    if (core->stacking->under)
      core->stacking->under->stacking->above = core->stacking->above;
#else
    RemoveFromStackList(core);
#endif
    wCoreDestroy(icon->core);
    wfree(icon);
}



/*
 *----------------------------------------------------------------------
 * wIconUpdate--
 * 	Updates or creates the icon contents, including the title and
 * pixmap. If the client does not specify anything to be shown there,
 * the default application icon is used. 
 * 
 * Side effects:
 * 	The icon window background pixmap is set and the title string
 * is reread.
 * 
 *---------------------------------------------------------------------- 
 */
void
wIconUpdate(WIcon *icon)
{
    WWindow *wwin = icon->owner;
    WScreen *scr=wwin->screen_ptr;
    unsigned int width, height, depth, th;
    int title_h=scr->icon_title_font.height;
    int had_title, title_changed=0;
    
    had_title = (icon->icon_name!=NULL);
    if (icon->icon_name)
      XFree(icon->icon_name);
    if (!XGetIconName(dpy, wwin->client_win, &icon->icon_name)) {
	icon->icon_name = NULL;
	if (had_title)
	  title_changed = 1;
    } else {
	if (!had_title)
	  title_changed = 1;
    }

    if (icon->show_title && icon->icon_name!=NULL) {
	th = title_h;
    } else {
	th = 0;
    }
    /*
     * The icon that will be used is selected in this precedence:
     * 
     * 1. application supplied icon window
     * 2. user specified icon for instance name
     * 3. user specified icon for class name
     * 4. application supplied icon pixmap
     * 5. built-in icon
     * 
     * the icon names will also be displayed if it's available
     * 
     */
    if (wwin->wm_hints && (wwin->wm_hints->flags & IconWindowHint)
	&& (!icon->client_window || title_changed)) {
	/* check if the client supplies an icon window */
	int resize=0, th;

	icon->our_pixmap = 0;
	icon->icon_win = wwin->wm_hints->icon_window;
	icon->client_window = 1;
	getSize(icon->icon_win, &width, &height, &depth);

	if (width > ICON_WIDTH) {
	    resize = 1;
	    width = ICON_WIDTH;
	}
	if (height > ICON_HEIGHT) {
	    resize = 1;
	    height = ICON_HEIGHT;
	}
	if (icon->show_title && icon->icon_name!=NULL 
	    && (height+title_h < ICON_HEIGHT)) {
	    Pixmap pix;
	    
	    pix = XCreatePixmap(dpy, scr->w_win, ICON_WIDTH, ICON_HEIGHT, 
				scr->w_depth);
	    XSetClipMask(dpy, scr->copy_gc, None);
	    XCopyArea(dpy, scr->icon_back_pixmap, pix, scr->copy_gc, 0, 0,
		      ICON_WIDTH, ICON_HEIGHT, 0, 0);
	    th = title_h;
	    XFillRectangle(dpy, pix, scr->icon_title_texture->normal_gc, 0, 0, 
			   ICON_WIDTH, th+1);
	    XDrawLine(dpy, pix, scr->icon_title_texture->light_gc, 0, 0,
		      ICON_WIDTH, 0);
	    XDrawLine(dpy, pix, scr->icon_title_texture->light_gc, 0, 0,
		      0, th+1);
	    XDrawLine(dpy, pix, scr->icon_title_texture->dim_gc, 
		      ICON_WIDTH-1, 0, ICON_WIDTH-1, th+1);
	    XSetWindowBackgroundPixmap(dpy, icon->core->window, pix);
	    XFreePixmap(dpy, pix);
	} else {
	    th = 0;
	    XSetWindowBackgroundPixmap(dpy, icon->core->window,
				       scr->icon_back_pixmap);
	}
	XSetWindowBorderWidth(dpy, icon->icon_win, 0);
	XReparentWindow(dpy, icon->icon_win, icon->core->window,
			(ICON_WIDTH-width)/2, th+(ICON_HEIGHT-height-th)/2);
	if (resize)
	  XResizeWindow(dpy, icon->icon_win, width, height);
	XMapWindow(dpy, icon->icon_win);
    } else if (wwin->icon_file && (!icon->our_pixmap || title_changed)) {
	/* look for user supplied icon file */
	char *path, *mask;
	Pixmap pixmap;

	icon->our_pixmap = 1;
	path = find_file(PixmapPath, wwin->icon_file);
	if (wwin->icon_mask_file)
	  mask = find_file(PixmapPath, wwin->icon_mask_file);
	else
	  mask = NULL;
	if (!path)
	  goto use_default_icon;

	pixmap = LoadIcon(scr, path, mask, th);
	if (pixmap) {
	    XSetWindowBackgroundPixmap(dpy, icon->core->window, pixmap);
	    XFreePixmap(dpy, pixmap);
	} else {
	    free(path);
	    if (mask)
	      free(mask);
	    goto use_default_icon;
	}
	free(path);
	if (mask)
	  free(mask);
    } else if (wwin->wm_hints && (wwin->wm_hints->flags & IconPixmapHint)) {
	int x, y;
	int w, h, d;
	Pixmap pix;

	icon->our_pixmap = 0;
	/* look for a client supplied pixmap */
	icon->client_pixmap = 1;
	icon->pixmap = wwin->wm_hints->icon_pixmap;
	if (wwin->wm_hints->flags & IconMaskHint)
	  icon->mask = wwin->wm_hints->icon_mask;
	else
	  icon->mask = None;
	/* we accept icon pixmaps of depth > 1, althought it's not 
	 * ICCCM compliant */
	getSize(icon->pixmap, (unsigned*)&w, (unsigned*)&h, (unsigned*)&d);
	
	if (w > MAX_ICON_WIDTH)
	  w = MAX_ICON_WIDTH;

	if (h > MAX_ICON_HEIGHT)
	  h = MAX_ICON_HEIGHT; 

	pix = XCreatePixmap(dpy, scr->w_win, ICON_WIDTH, ICON_HEIGHT,
			    scr->w_depth);
	XSetClipMask(dpy, scr->copy_gc, None);
	XCopyArea(dpy, scr->icon_back_pixmap, pix, scr->copy_gc, 0, 0,
		  ICON_WIDTH, ICON_HEIGHT, 0, 0);
	x = (ICON_WIDTH - w)/2;
	y = th + (ICON_HEIGHT - h - th)/2;
	XSetClipOrigin(dpy, scr->copy_gc, x, y);
	XSetClipMask(dpy, scr->copy_gc, icon->mask);
	if (d==1) {
	    XCopyPlane(dpy, icon->pixmap, pix, scr->copy_gc, 0, 0, w, h, 
		       x, y, 1);
	} else {
	    XCopyArea(dpy, icon->pixmap, pix, scr->copy_gc, 0, 0, w, h, x, y);
	}
	if (th>0) {
	    XFillRectangle(dpy, pix, scr->icon_title_texture->normal_gc, 0, 0,
			   ICON_WIDTH, th+1);
	    XDrawLine(dpy, pix, scr->icon_title_texture->light_gc, 0, 0,
		      ICON_WIDTH, 0);
	    XDrawLine(dpy, pix, scr->icon_title_texture->light_gc, 0, 0,
		      0, th+1);
	    XDrawLine(dpy, pix, scr->icon_title_texture->dim_gc, 
		      ICON_WIDTH-1, 0, ICON_WIDTH-1, th+1);
	}
	XSetWindowBackgroundPixmap(dpy, icon->core->window, pix);
	XFreePixmap(dpy, pix);
	if (y <= title_h)
	  icon->show_title = 0;
    } else if (!icon->our_pixmap || title_changed) {
	use_default_icon:

	icon->our_pixmap = 1;
	if (icon->show_title && icon->icon_name!=NULL)
	  XSetWindowBackgroundPixmap(dpy, icon->core->window, 
				     scr->default_titled_icon);
	else
	  XSetWindowBackgroundPixmap(dpy, icon->core->window,
				     scr->default_icon);
    }
    wIconPaint(icon);
}



void
wIconPaint(WIcon *icon)
{
    WScreen *scr=icon->core->screen_ptr;
    GC gc = scr->icon_title_gc;
    int x;

    XClearWindow(dpy, icon->core->window);
    
    /* draw the icon title */
    if (icon->show_title && icon->icon_name!=NULL) {
	int l = strlen(icon->icon_name);

#ifdef I18N
	x = (icon->core->width - XmbTextEscapement(scr->icon_title_font.font, 
						   icon->icon_name, l))/2;
	XmbDrawString(dpy, icon->core->window, scr->icon_title_font.font, gc, 
		      x, 1 + scr->icon_title_font.y, icon->icon_name, l);
#else
	int w = XTextWidth(scr->icon_title_font.font, icon->icon_name, l);
	if (w > icon->core->width - 4)
	  x = (icon->core->width - 4) - w;
	else
	  x = (icon->core->width - w)/2;
	XDrawString(dpy, icon->core->window, gc, x, 
		    1 + scr->icon_title_font.y, icon->icon_name, l);
#endif
    }
}


/******************************************************************/
static void 
iconDblClick(WObjDescriptor *desc, XEvent *event)
{
    WIcon *icon = desc->parent;
    WApplication *wapp;

    if (icon->group_icon) {
	wapp = wApplicationOf(icon->owner->main_window);
#ifdef DEBUG0
	if (!wapp) {
	    wWarning("could not find application descriptor for app icon!!");
	    return;
	}
#endif
	wUnhideApplication(wapp);
    } else {
	wDeiconifyWindow(icon->owner);
    }
}


static void 
iconMouseDown(WObjDescriptor *desc, XEvent *event)
{
    WIcon *icon = desc->parent;
    WWindow *wwin = icon->owner;
    XEvent ev;
    int x=wwin->icon_x, y=wwin->icon_y;
    int dx=event->xbutton.x_root, dy=event->xbutton.y_root;
    int grabbed=0;

#ifdef DEBUG
    puts("Moving icon");
#endif
    if (event->xbutton.state & Mod1Mask)
      wLowerFrame(icon->core);
    else
      wRaiseFrame(icon->core);

    if (XGrabPointer(dpy, icon->core->window, False, ButtonMotionMask
		     |ButtonReleaseMask, GrabModeAsync, GrabModeAsync, None, 
		     None, CurrentTime) !=GrabSuccess) {
	wWarning("pointer grab failed for icon move");
    }
    while(1) {
	XMaskEvent(dpy, PointerMotionMask|ButtonReleaseMask|ButtonMotionMask
		   |ExposureMask, &ev);
	switch (ev.type) {
	 case Expose:
	    wIconPaint(icon);
	    break;

	 case MotionNotify:
	    if (!grabbed) {
		if (abs(dx-ev.xmotion.x_root)>=MOVE_THRESHOLD
		    || abs(dy-ev.xmotion.y_root)>=MOVE_THRESHOLD) {
		    XChangeActivePointerGrab(dpy, ButtonMotionMask
					     |ButtonReleaseMask, 
					     wCursor[WCUR_MOVE], CurrentTime);
		    dx = ev.xmotion.x_root;
		    dy = ev.xmotion.y_root;
		    grabbed=1;
		} else {
		    break;
		}
	    }
	    x += ev.xmotion.x_root - dx;
	    y += ev.xmotion.y_root - dy;
	    dx = ev.xmotion.x_root;
	    dy = ev.xmotion.y_root;
	    XMoveWindow(dpy, icon->core->window, x, y);
	    break;

	 case ButtonRelease:
	    if (wwin->icon_x!=x || wwin->icon_y!=y)
	      wwin->flags.icon_moved = 1;

	    XMoveWindow(dpy, icon->core->window, x, y);

	    wwin->icon_x = x;
	    wwin->icon_y = y;
#ifdef DEBUG
	    puts("End icon move");
#endif
	    XUngrabPointer(dpy, CurrentTime);
	    return;
	    
	}
    }
}
    
