/* 
 *  WindowMaker window manager
 * 
 *  Copyright (c) 1997 Alfredo K. Kojima
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include "wconfig.h"

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <pwd.h>

#include "generic/wwmlib.h"
#include "WindowMaker.h"
#include "GNUstep.h"
#include "version.h"
#include "screen.h"
#include "wcore.h"
#include "window.h"

extern *DisplayName;

#ifdef USECPP
static void
putdef(char *line, char *name, char *value)
{
    if (!value) {
	wWarning("could not define value for %s for cpp", name);
	return;
    }
    strcat(line, name);
    strcat(line, value);
}



static void
putidef(char *line, char *name, int value)
{
    char tmp[64];
    sprintf(tmp, "%i", value);
    strcat(line, name);
    strcat(line, tmp);
}


static char*
username()
{
    char *tmp;
    
    tmp = getlogin();
    if (!tmp) {
	struct passwd *user;

	user = getpwuid(getuid());
	if (!user) {
	    wSysError("could not get password entry for UID %i", getuid());
	    return NULL;
	}
	if (!user->pw_name) {
	    return NULL;
	} else {
	    return user->pw_name;
	}
    }
    return tmp;
}
       
char *
MakeCPPArgs(char *path)
{
    int i;
    char buffer[MAXLINE], *buf, *line;
    char version[64];
    Visual *visual;
    
    line = wmalloc(MAXLINE);
    *line = 0;
    i=1;
    if (gethostname(buffer, MAXLINE)==0) {
	putdef(line, " -DHOST=", buffer);
    }
    buf = username();
    if (buf)
      putdef(line, " -DUSER=", buf);
    putidef(line, " -DUID=", getuid());
    buf = XDisplayName(DisplayString(dpy));
    putdef(line, " -DDISPLAY=", buf);
    putidef(line, " -DMAJ_VERSION=", MAJ_VERSION);
    putidef(line, " -DMIN_VERSION=", MIN_VERSION);
    sprintf(version, "%d.%d.%d", MAJ_VERSION, MIN_VERSION, SUBMIN_VERSION);
    putdef(line, " -DWM_VERSION=", version);
    
    visual = DefaultVisual(dpy, DefaultScreen(dpy));
    putidef(line, " -DVISUAL=", visual->class);
    
    putidef(line, " -DDEPTH=", DefaultDepth(dpy, DefaultScreen(dpy)));

    putidef(line, " -DSCR_WIDTH=", WidthOfScreen(DefaultScreenOfDisplay(dpy)));
    putidef(line, " -DSCR_HEIGHT=", 
	    HeightOfScreen(DefaultScreenOfDisplay(dpy)));

    strcpy(buffer, path);    
    buf = strrchr(buffer, '/');
    if (buf) *buf = 0; /* trunc filename */
    putdef(line, " -I", buffer);
#ifdef XPM
    strcat(line, " -DXPM");
#endif
#ifdef DITHERING
    strcat(line, " -DDITHERING");
#endif
    
#undef arg
#ifdef DEBUG
    puts("CPP ARGS");
    puts(line);
#endif
    return line;
}
#endif /* USECPP */




int
LevelToIndex(int level)
{
    switch (level) {
     case NSNormalWindowLevel:
	return 0;
     case NSFloatingWindowLevel:
	return 1;
     case NSDockWindowLevel:
	return 2;
     case NSSubmenuWindowLevel:
	return 3;
     case NSMainMenuWindowLevel:
	return 4;
     default:
	return 0;
    }
}



/*
 *----------------------------------------------------------------------
 * AddToStackList--
 * 	Inserts the frame in the top of the stacking list. The
 * stacking precedence is obeyed.
 * 
 * Returns:
 * 	None
 * 
 * Side effects:
 * 	The frame is added to it's screen's window list.
 *---------------------------------------------------------------------- 
 */
#ifdef BUGGYSTACKING
void
AddToStackList(WCoreWindow *frame)
{
    WCoreWindow *prev, *tmpw;

    frame->screen_ptr->window_count++;
    tmpw = frame->screen_ptr->stacking_list;
    if (!tmpw) {
	frame->screen_ptr->stacking_list = frame;
	frame->stacking->above = NULL;
	frame->stacking->under = NULL;
	return;
    }
    prev = tmpw;
    while (tmpw && (frame->stacking->window_level
		    <tmpw->stacking->window_level)) {
	prev = tmpw;
	tmpw = tmpw->stacking->under;
    }
    /* add window after our place */
    if (tmpw == frame->screen_ptr->stacking_list) {
	/* last window on list (top) */
	tmpw->stacking->above = frame;
	frame->stacking->under = tmpw;
	frame->stacking->above = NULL;
	frame->screen_ptr->stacking_list = frame;
    } else if (tmpw == NULL) {
	/* first window on list (bottom) */
	prev->stacking->under = frame;
	frame->stacking->under = NULL;
	frame->stacking->above = prev;
    } else {
	/* other cases */
	frame->stacking->above = tmpw->stacking->above;
	frame->stacking->under = tmpw;
	tmpw->stacking->above->stacking->under = frame;
	tmpw->stacking->above = frame;
    }
    wScreenCommitStacking(frame->screen_ptr);
}
#else /* BUGGYSTACKING */
void
AddToStackList(WCoreWindow *frame)
{
    WCoreWindow *prev, *tmpw;
    int index = LevelToIndex(frame->stacking->window_level);

    frame->screen_ptr->window_count++;
    frame->screen_ptr->window_level_count[index]++;
    tmpw = frame->screen_ptr->stacking_list[index];
    if (!tmpw) {
	frame->screen_ptr->stacking_list[index] = frame;
	frame->stacking->above = NULL;
	frame->stacking->under = NULL;
	wScreenCommitStacking(frame->screen_ptr);	
	return;
    }
    prev = tmpw;

    /* put on top of the stacking list */
    frame->stacking->above = NULL;
    frame->stacking->under = tmpw;
    tmpw->stacking->above = frame;
    frame->screen_ptr->stacking_list[index] = frame;
    wScreenCommitStacking(frame->screen_ptr);
}
#endif /* !BUGGYSTACKING */

#ifndef BUGGYSTACKING
void
RemoveFromStackList(WCoreWindow *frame)
{
    int level = LevelToIndex(frame->stacking->window_level);

    /* remove from the window stack list */
    if (frame->stacking->under)
      frame->stacking->under->stacking->above = frame->stacking->above;
    if (frame->stacking->above)
      frame->stacking->above->stacking->under = frame->stacking->under;
    else /* this was the first window on the list */
      frame->screen_ptr->stacking_list[level] = frame->stacking->under;

    frame->screen_ptr->window_count--;
    frame->screen_ptr->window_level_count[level]--;
}
#endif /* !BUGGYSTACKING */

WWindow*
NextFocusWindow(WScreen *scr)
{
    WWindow *tmp, *wwin, *closest, *min;
    Window d;
    
    wwin = scr->focused_window;
    tmp = wwin->prev;
    closest = NULL;
    min = wwin;
    d = 0xffffffff;
    while (tmp) {
	if (tmp->flags.mapped || tmp->flags.shaded) {
	    if (min->client_win > tmp->client_win)
	      min = tmp;
	    if (tmp->client_win > wwin->client_win
		&& (!closest
		    || (tmp->client_win - wwin->client_win) < d)) {
		closest = tmp;
		d = tmp->client_win - wwin->client_win;
	    }
	}
	tmp = tmp->prev;
    }
    if (!closest||closest==wwin)
      return min;
    return closest;
}


WWindow*
PrevFocusWindow(WScreen *scr)
{
    WWindow *tmp, *wwin, *closest, *max;
    Window d;
    
    wwin = scr->focused_window;
    tmp = wwin->prev;
    closest = NULL;
    max = wwin;
    d = 0xffffffff;
    while (tmp) {
	if (tmp->flags.mapped || tmp->flags.shaded) {
	    if (max->client_win < tmp->client_win)
	      max = tmp;
	    if (tmp->client_win < wwin->client_win
		&& (!closest
		    || (wwin->client_win - tmp->client_win) < d)) {
		closest = tmp;
		d = wwin->client_win - tmp->client_win;
	    }
	}
	tmp = tmp->prev;
    }
    if (!closest||closest==wwin)
      return max;
    return closest;
}
