/* 
 *  WindowMaker window manager
 * 
 *  Copyright (c) 1997 Shige Abe and 
 *		       Alfredo K. Kojima
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include "wconfig.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include "generic/wwmlib.h"
#include "WindowMaker.h"
#include "window.h"
#include "actions.h"
#include "client.h"
#include "funcs.h"
#include "stacking.h"


/********* Global Variables *******/
extern WScreen wScreen;



/*
 * TODO:
 * 
 * - disable entry of the current selected window
 * + update window title when it's changed
 *
 */


/*
 * FocusWindow
 *
 *  - Needs to check if already in the right workspace before
 *    calling wChangeWorkspace?
 *
 *  Order:
 *    Switch to correct workspace
 *    Unshade if shaded
 *    If iconified then deiconify else focus/raise.
 */
static void
focusWindow(WMenu *menu, WMenuEntry *entry)
{
    WWindow *w;
    w = (WWindow*)entry->clientdata;
    wChangeWorkspace(menu->screen_ptr,w->workspace);
    if (w->flags.shaded) {
	wUnshadeWindow(w);
    }
    if (w->flags.hidden) {
	wUnhideApplication(wApplicationOf(w->main_window));
    } else if (w->flags.miniaturized) {
	wDeiconifyWindow(w);
    } else {
	wSetFocusTo(menu->screen_ptr,w);
	wRaiseFrame(w->frame);
    }
}
 
/*
 *
 * Open switch menu 
 *
 */
void 
OpenSwitchMenu(WScreen *scr, int x, int y)
{
    WMenu *switchmenu = scr->switch_menu;
    WWindow *wwin;
    char title[MAX_MENU_TEXT_LENGTH+6];

    if (switchmenu) {
	if (switchmenu->flags.mapped) {
	    if (!switchmenu->flags.buttoned)
	      wMenuUnmap(switchmenu);
	    else
	      wRaiseFrame(switchmenu->frame);
	} else {
	    wMenuMapAt(switchmenu, x-switchmenu->frame->width/2,
		       y-switchmenu->titlebar->height/2);
	}
	return;
    }
    switchmenu = wMenuCreate(scr,_("Windows"),True);
    wwin = wScreen.focused_window;
    while (wwin) {
	WMenuEntry *entry;
	/*
	char *wspace;
	 */
	if (wwin->flags.internal_window) {
	    wwin = wwin->prev;
	    continue;
	}
	sprintf(title, "%s", wwin->title);
	entry = wMenuAddCallback(switchmenu, title, focusWindow, wwin);
	if (!wwin->window_flags.focusable)
	  wMenuSetEnabled(switchmenu, entry->order, False);
	/*
	wspace = wmalloc(16);
	if (wspace) {
	    sprintf(wspace, "[%i]", wwin->workspace+1);
	    entry->rtext = wspace;
	}
	 * */
        wwin = wwin->prev;
    }
    scr->switch_menu = switchmenu;
    if (switchmenu) {
	if (!switchmenu->flags.realized)
	  wMenuRealize(switchmenu);
	wMenuMapAt(switchmenu, x-switchmenu->frame->width/2,
		   y-switchmenu->titlebar->height/2);
    }
}

/*
 *
 * Update switch menu
 *
 *
 */
void 
UpdateSwitchMenu(WScreen *scr, WWindow *wwin, int action)
{
    WMenu *switchmenu = scr->switch_menu;
    WMenuEntry *entry;
    char title[MAX_MENU_TEXT_LENGTH+6];
    int i;
    /*
    char *wspace;
     */

    /*
     *  This menu is updated under the following conditions:
     *
     *    1.  When a window is created.
     *    2.  When a window is destroyed.
     *
     *    3.  When a window changes it's title.
     */


    if (action == ACTION_ADD) {
	if (wwin->title)
	  sprintf(title, "%s", wwin->title);
	else
	  sprintf(title, "%s", DEF_WINDOW_TITLE);
	entry = wMenuAddCallback(switchmenu, title, focusWindow, wwin);
	if (!wwin->window_flags.focusable)
	  wMenuSetEnabled(switchmenu, entry->order, False);
	/*
	wspace = wmalloc(16);
	if (wspace) {
	    sprintf(wspace, "[%i]", wwin->workspace+1);
	    entry->rtext = wspace;
	}
	 */
	wMenuRealize(switchmenu);
    } else {
	for (i=0; i<switchmenu->entry_no; i++) {
	    entry = switchmenu->entries[i];
	    /* this is the entry that was changed */
	    if (entry->clientdata == wwin) {
		switch (action) {
		 case ACTION_REMOVE:
		    wMenuRemoveItem(switchmenu, i);
		    wMenuRealize(switchmenu);
		    break;
		    
		 case ACTION_CHANGE:
		    if (entry->text)
		      free(entry->text);
		    /*
		    if (entry->rtext)
		      wfree(entry->rtext);
		     */
		    if (wwin->title)
		      sprintf(title, "%s", wwin->title);
		    else
		      sprintf(title, "%s", DEF_WINDOW_TITLE);
		    entry->text = strdup(title);
		    /*
		    wspace = wmalloc(16);
		    if (wspace) {
			sprintf(wspace, "[%i]", wwin->workspace+1);
			entry->rtext = wspace;
		    }
		     */
		    wMenuRealize(switchmenu);
		    break;
		}
		break;
	    }
	}
    }
    wMenuPaint(switchmenu);
}
