/*
 * ScrolledCanvas.c -- Routines for the ScrolledCanvas widget
 *
 * This file is part of the libXas project.
 * Copyright (c) 1997 Sbastien Phlep (phelep@iu-vannes.fr)
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
/*
 * $Log: ScrolledCanvas.c,v $
 * Revision 1.0  1997/05/10 12:17:32  seb
 * Initial revision
 *
 */

#include <X11/Xlib.h>
#include <stdio.h>
#include "ScrolledCanvas.h"
#include "GraphicalUtilities.h"

/* The following functions cannot be used outside of this module ! */
static void _XasRedrawSC_BoundingBox(ScrolledCanvas me);
static void _XasRedrawSC_VSB(ScrolledCanvas me, int state);
static void _XasRedrawSC_HSB(ScrolledCanvas me, int state);
static void _XasRedrawSC_VUAB(ScrolledCanvas me, int state);
static void _XasRedrawSC_VDAB(ScrolledCanvas me, int state);
static void _XasRedrawSC_HLAB(ScrolledCanvas me, int state);
static void _XasRedrawSC_HRAB(ScrolledCanvas me, int state);
static void _XasResizeSC_VertSlider(ScrolledCanvas me);
static void _XasResizeSC_HorizSlider(ScrolledCanvas me);
static void _XasReplaceSC_Sliders(ScrolledCanvas me);
static void _XasMoveSCVertical(ScrolledCanvas me, XEvent event);
static void _XasMoveSCHorizontal(ScrolledCanvas me, XEvent event);
static void _XasMoveChild_Vertical(ScrolledCanvas me, int free_height, int new_ycoord);
static void _XasMoveChild_Horizontal(ScrolledCanvas me, int free_width, int new_xcoord);


/****************************************
 * Public methods			*
 ****************************************/


extern ScrolledCanvas
XasAddScrolledCanvas(Window parent, Window child, int x, int y, int w, int h,
		     Bool enableVertButtons, Bool enableHorizButtons, Bool forceSlidingButtons,
		     Bool enableSoftSliders)
/*
 * Adds a ScrolledCanvas widget:
 *  'parent' is the X-Window where the widget will be displayed,
 *  'child' is the X-Window that will be managed by the widget,
 *  'x' is the coordinate from the left of the 'parent' window where the widget will be displayed,
 *  'y' is the coordinate from the top of the 'parent' window where the widget will be displayed,
 *  'w' is the width of the widget,
 *  'h' is the height of the widget,
 *  'enableVertButtons' tells if vertical arrow buttons should be displayed,
 *  'enableHorizButtons' tells if horizontal arrow buttons should be displayed,
 *  'forceSlidingButtons' tells if sliding buttons must be shown even when they are not needed
 *
 * It returns the so-created ScrolledCanvas widget structure...
 *
 */
{
  ScrolledCanvas	me;

  /* Initialize truth values */
  me.enableVB = enableVertButtons;
  me.enableHB = enableHorizButtons;
  me.forceSB  = forceSlidingButtons;
  me.enableSS = enableSoftSliders;

  /* Bounding box */
  me.bbox_w = XCreateSimpleWindow(xasDisplay,parent,x,y,w,h,0,xasBlack,xasGrey60);
  me.bbox_c = XCreateGC(xasDisplay,me.bbox_w,0,NULL);
  XSelectInput(xasDisplay,me.bbox_w,ExposureMask);

  /* Canvas */
  me.cnvs_w = XCreateSimpleWindow(xasDisplay,me.bbox_w,19,3,(w-23),(h-22),0,xasBlack,xasGrey);
  me.cnvs_c = XCreateGC(xasDisplay,me.cnvs_w,0,NULL);
  XSelectInput(xasDisplay,me.cnvs_w,ExposureMask);

  /* Vertical scrollbar structure */
  if(me.enableVB == False)
    me.vsbb_w = XCreateSimpleWindow(xasDisplay,me.bbox_w,2,2,15,(h-20),0,xasBlack,xasGrey60);
  else {
    me.vsbb_w = XCreateSimpleWindow(xasDisplay,me.bbox_w,2,33,15,((h-20)-31),0,xasBlack,xasGrey60);
    me.vubt_w = XCreateSimpleWindow(xasDisplay,me.bbox_w,2,2,15,15,0,xasBlack,xasGrey);
    me.vubt_c = XCreateGC(xasDisplay,me.vubt_w,0,NULL);
    XSelectInput(xasDisplay,me.vubt_w,ExposureMask|ButtonPressMask|ButtonReleaseMask);
    me.vdbt_w = XCreateSimpleWindow(xasDisplay,me.bbox_w,2,17,15,15,0,xasBlack,xasGrey);
    me.vdbt_c = XCreateGC(xasDisplay,me.vdbt_w,0,NULL);
    XSelectInput(xasDisplay,me.vdbt_w,ExposureMask|ButtonPressMask|ButtonReleaseMask);
  }
  if(me.enableSS == True)
    XSelectInput(xasDisplay,me.vsbb_w,Button1MotionMask|ButtonPressMask|ButtonReleaseMask);
  else
    XSelectInput(xasDisplay,me.vsbb_w,Button1MotionMask|ButtonPressMask);
  me.vsld_w = XCreateSimpleWindow(xasDisplay,me.vsbb_w,0,0,15,15,0,xasBlack,xasGrey);
  me.vsld_c = XCreateGC(xasDisplay,me.vsld_w,0,NULL);
  XSelectInput(xasDisplay,me.vsld_w,ExposureMask);

  /* Horizontal scrollbar structure */
  if(me.enableHB == False)
    me.hsbb_w = XCreateSimpleWindow(xasDisplay,me.bbox_w,18,(h-17),(w-21),15,0,xasBlack,xasGrey60);
  else {
    me.hsbb_w = XCreateSimpleWindow(xasDisplay,me.bbox_w,18,(h-17),((w-21)-31),15,0,xasBlack,xasGrey60);
    me.hlbt_w = XCreateSimpleWindow(xasDisplay,me.bbox_w,(w-33),(h-17),15,15,0,xasBlack,xasGrey);
    me.hlbt_c = XCreateGC(xasDisplay,me.hlbt_w,0,NULL);
    XSelectInput(xasDisplay,me.hlbt_w,ExposureMask|ButtonPressMask|ButtonReleaseMask);
    me.hrbt_w = XCreateSimpleWindow(xasDisplay,me.bbox_w,(w-18),(h-17),15,15,0,xasBlack,xasGrey);
    me.hrbt_c = XCreateGC(xasDisplay,me.hrbt_w,0,NULL);
    XSelectInput(xasDisplay,me.hrbt_w,ExposureMask|ButtonPressMask|ButtonReleaseMask);
  }
  if(me.enableSS == True)
    XSelectInput(xasDisplay,me.hsbb_w,Button1MotionMask|ButtonPressMask|ButtonReleaseMask);
  else
    XSelectInput(xasDisplay,me.hsbb_w,Button1MotionMask|ButtonPressMask);
  me.hsld_w = XCreateSimpleWindow(xasDisplay,me.hsbb_w,0,0,15,15,0,xasBlack,xasGrey);
  me.hsld_c = XCreateGC(xasDisplay,me.hsld_w,0,NULL);
  XSelectInput(xasDisplay,me.hsld_w,ExposureMask);

  /* Reparent the child window to make it fit in the canvas */
  me.chld_w = child;
  XSelectInput(xasDisplay,me.chld_w,ExposureMask|StructureNotifyMask);
  XReparentWindow(xasDisplay,me.chld_w,me.cnvs_w,0,0);
  _XasResizeSC_VertSlider(me);
  _XasResizeSC_HorizSlider(me);

  /* Return the widget */
  return me;
};


extern int
XasSetChildPositionInCanvas(ScrolledCanvas me, int x_coord, int y_coord)
/*
 * This routine moves the canvas' child window so that
 * 'x_coord' and 'y_coord' fit in the top-left corner of
 * the canvas. Use the 'UNCHANGED' value for 'x_coord' or
 * 'y_coord' if you don't want to change their position.
 * It returns -1 if 'x_coord' or 'y_coord' are not correct values,
 * 0 if the operation was successful, or 1 if the given coords
 * are shown in the canvas, but cannot be used as origin for the canvas.
 */
{
  Window	tmp_win1, tmp_win2;
  int		tmp_x1, tmp_y1, tmp_x2, tmp_y2;
  unsigned int	tmp_w1, tmp_h1, tmp_b1, tmp_d1, tmp_w2, tmp_h2, tmp_b2, tmp_d2;
  int		new_x, new_y, ret=0;

  XGetGeometry(xasDisplay,me.chld_w,&tmp_win1,&tmp_x1,&tmp_y1,&tmp_w1,&tmp_h1,&tmp_b1,&tmp_d1);
  XGetGeometry(xasDisplay,me.cnvs_w,&tmp_win2,&tmp_x2,&tmp_y2,&tmp_w2,&tmp_h2,&tmp_b2,&tmp_d2);
  /* Test if "UNCHANGED" is used, and replace with correct values in that case */
  if(x_coord == UNCHANGED) x_coord = -tmp_x1;
  if(y_coord == UNCHANGED) y_coord = -tmp_y1;
  /* Now, test if 'x_coord' and 'y_coord' are correct values */
  if((x_coord < 0) || (y_coord < 0) || (x_coord > tmp_w1) || (y_coord > tmp_h1))
    return -1;
  /* Now, test if the given coordinates can be used "as is" */
  if(x_coord <= (tmp_w1 - tmp_w2 + 2))
    new_x = x_coord;
  else {
    new_x = (tmp_w1 - tmp_w2 + 2);
    ret = 1;
  }
  if(y_coord <= (tmp_h1 - tmp_h2 + 2))
    new_y = y_coord;
  else {
    new_y = (tmp_h1 - tmp_h2 + 2);
    ret = 1;
  }
  /* Finally, move the window correctly and return 0 or 1 */
  XMoveWindow(xasDisplay,me.chld_w,-new_x,-new_y);
  return ret;
};


/****************************************
 * Private methods			*
 ****************************************/


static void
_XasRedrawSC_BoundingBox(ScrolledCanvas me)
/*
 * This function redraws the bounding box of the 'me' ScrolledCanvas widget
 */
{
  Window	tmp_win;
  int		tmp_x, tmp_y;
  unsigned int	tmp_w, tmp_h, tmp_b, tmp_d;

  XGetGeometry(xasDisplay,me.bbox_w,&tmp_win,&tmp_x,&tmp_y,&tmp_w,&tmp_h,&tmp_b,&tmp_d);
  XSetForeground(xasDisplay,me.bbox_c,xasGrey90);
  XDrawLine(xasDisplay,me.bbox_w,me.bbox_c,(tmp_w-5),2,18,2);
  XDrawLine(xasDisplay,me.bbox_w,me.bbox_c,18,2,18,(tmp_h-20));
  XSetForeground(xasDisplay,me.bbox_c,xasGrey44);
  XDrawLine(xasDisplay,me.bbox_w,me.bbox_c,18,(tmp_h-19),(tmp_w-4),(tmp_h-19));
  XDrawLine(xasDisplay,me.bbox_w,me.bbox_c,(tmp_w-4),(tmp_h-19),(tmp_w-4),2);

  _XasDrawObjectBorder(me.bbox_w,me.bbox_c,XAS_RELIEF_DNBOX);
};


static void
_XasRedrawSC_VSB(ScrolledCanvas me, int state)
/*
 * This function redraws the vertical sliding button of the 'me' ScrolledCanvas widget
 */
{
  Window	tmp_win;
  int		tmp_x, tmp_y, center;
  unsigned int	tmp_w, tmp_h, tmp_b, tmp_d;
  
  XGetGeometry(xasDisplay,me.vsld_w,&tmp_win,&tmp_x,&tmp_y,&tmp_w,&tmp_h,&tmp_b,&tmp_d);
  center = (tmp_h / 2);
  _XasDrawObjectBorder(me.vsld_w,me.vsld_c,state);
  XSetForeground(xasDisplay,me.vsld_c,xasGrey44);
  XDrawLine(xasDisplay,me.vsld_w,me.vsld_c,8,(center-3),6,(center-3));
  XDrawLine(xasDisplay,me.vsld_w,me.vsld_c,6,(center-3),4,(center-1));
  XDrawLine(xasDisplay,me.vsld_w,me.vsld_c,4,(center-1),4,(center+1));
  XDrawPoint(xasDisplay,me.vsld_w,me.vsld_c,5,(center+2));
  XSetForeground(xasDisplay,me.vsld_c,xasGrey90);
  XDrawLine(xasDisplay,me.vsld_w,me.vsld_c,9,(center-2),10,(center-1));
  XDrawLine(xasDisplay,me.vsld_w,me.vsld_c,10,(center-1),10,(center+1));
  XDrawLine(xasDisplay,me.vsld_w,me.vsld_c,10,(center+1),9,(center+2));
  XDrawLine(xasDisplay,me.vsld_w,me.vsld_c,9,(center+2),8,(center+3));
  XDrawLine(xasDisplay,me.vsld_w,me.vsld_c,8,(center+3),6,(center+3));
  XDrawLine(xasDisplay,me.vsld_w,me.vsld_c,9,center,7,(center+2));
  XDrawLine(xasDisplay,me.vsld_w,me.vsld_c,9,(center+1),8,(center+2));
  XDrawPoint(xasDisplay,me.vsld_w,me.vsld_c,8,center);
  XDrawPoint(xasDisplay,me.vsld_w,me.vsld_c,7,(center+1));
};


static void
_XasRedrawSC_HSB(ScrolledCanvas me, int state)
/*
 * This function redraws the horizontal sliding button of the 'me' ScrolledCanvas widget
 */
{
  Window	tmp_win;
  int		tmp_x, tmp_y, center;
  unsigned int	tmp_w, tmp_h, tmp_b, tmp_d;
  
  XGetGeometry(xasDisplay,me.hsld_w,&tmp_win,&tmp_x,&tmp_y,&tmp_w,&tmp_h,&tmp_b,&tmp_d);
  center = (tmp_w / 2);
  _XasDrawObjectBorder(me.hsld_w,me.hsld_c,state);
  XSetForeground(xasDisplay,me.hsld_c,xasGrey44);
  XDrawLine(xasDisplay,me.hsld_w,me.hsld_c,(center-3),8,(center-3),6);
  XDrawLine(xasDisplay,me.hsld_w,me.hsld_c,(center-3),6,(center-1),4);
  XDrawLine(xasDisplay,me.hsld_w,me.hsld_c,(center-1),4,(center+1),4);
  XDrawPoint(xasDisplay,me.hsld_w,me.hsld_c,(center+2),5);
  XSetForeground(xasDisplay,me.hsld_c,xasGrey90);
  XDrawLine(xasDisplay,me.hsld_w,me.hsld_c,(center-2),9,(center-1),10);
  XDrawLine(xasDisplay,me.hsld_w,me.hsld_c,(center-1),10,(center+1),10);
  XDrawLine(xasDisplay,me.hsld_w,me.hsld_c,(center+1),10,(center+2),9);
  XDrawLine(xasDisplay,me.hsld_w,me.hsld_c,(center+2),9,(center+3),8);
  XDrawLine(xasDisplay,me.hsld_w,me.hsld_c,(center+3),8,(center+3),6);
  XDrawLine(xasDisplay,me.hsld_w,me.hsld_c,center,9,(center+2),7);
  XDrawLine(xasDisplay,me.hsld_w,me.hsld_c,(center+1),9,(center+2),8);
  XDrawPoint(xasDisplay,me.hsld_w,me.hsld_c,center,8);
  XDrawPoint(xasDisplay,me.hsld_w,me.hsld_c,(center+1),7);
};


static void
_XasRedrawSC_VUAB(ScrolledCanvas me, int state)
/*
 * This function redraws the vertical up-arrow scrolling button of the 'me' ScrolledCanvas widget.
 * Depending on the value of 'state', it will look pushed or released. Cf GraphicalUtilities.h to
 * get started with the possible 'state' values.
 */
{
  _XasDrawObjectBorder(me.vubt_w,me.vubt_c,state);
  XSetForeground(xasDisplay,me.vubt_c,xasBlack);
  XDrawLine(xasDisplay,me.vubt_w,me.vubt_c,3,11,9,11);
  XDrawLine(xasDisplay,me.vubt_w,me.vubt_c,4,10,8,10);
  XDrawLine(xasDisplay,me.vubt_w,me.vubt_c,4,9,8,9);
  XDrawLine(xasDisplay,me.vubt_w,me.vubt_c,5,8,7,8);
  XDrawLine(xasDisplay,me.vubt_w,me.vubt_c,5,7,7,7);
  XDrawLine(xasDisplay,me.vubt_w,me.vubt_c,6,5,6,6);
};


static void
_XasRedrawSC_VDAB(ScrolledCanvas me, int state)
/*
 * This function redraws the vertical down-arrow scrolling button of the 'me' ScrolledCanvas widget.
 * Depending on the value of 'state', it will look pushed or released. Cf GraphicalUtilities.h to
 * get started with the possible 'state' values.
 */
{
  _XasDrawObjectBorder(me.vdbt_w,me.vdbt_c,state);
  XSetForeground(xasDisplay,me.vdbt_c,xasBlack);
  XDrawLine(xasDisplay,me.vdbt_w,me.vdbt_c,3,3,9,3);
  XDrawLine(xasDisplay,me.vdbt_w,me.vdbt_c,4,4,8,4);
  XDrawLine(xasDisplay,me.vdbt_w,me.vdbt_c,4,5,8,5);
  XDrawLine(xasDisplay,me.vdbt_w,me.vdbt_c,5,6,7,6);
  XDrawLine(xasDisplay,me.vdbt_w,me.vdbt_c,5,7,7,7);
  XDrawLine(xasDisplay,me.vdbt_w,me.vdbt_c,6,8,6,9);
};


static void
_XasRedrawSC_HLAB(ScrolledCanvas me, int state)
/*
 * This function redraws the horizontal left-arrow scrolling button of the 'me' ScrolledCanvas widget.
 * Depending on the value of 'state', it will look pushed or released. Cf GraphicalUtilities.h to
 * get started with the possible 'state' values.
 */
{
  _XasDrawObjectBorder(me.hlbt_w,me.hlbt_c,state);
  XSetForeground(xasDisplay,me.hlbt_c,xasBlack);
  XDrawLine(xasDisplay,me.hlbt_w,me.hlbt_c,4,7,5,7);
  XDrawLine(xasDisplay,me.hlbt_w,me.hlbt_c,6,6,6,8);
  XDrawLine(xasDisplay,me.hlbt_w,me.hlbt_c,7,6,7,8);
  XDrawLine(xasDisplay,me.hlbt_w,me.hlbt_c,8,5,8,9);
  XDrawLine(xasDisplay,me.hlbt_w,me.hlbt_c,9,5,9,9);
  XDrawLine(xasDisplay,me.hlbt_w,me.hlbt_c,10,4,10,10);
};


static void
_XasRedrawSC_HRAB(ScrolledCanvas me, int state)
/*
 * This function redraws the horizontal right-arrow scrolling button of the 'me' ScrolledCanvas widget.
 * Depending on the value of 'state', it will look pushed or released. Cf GraphicalUtilities.h to
 * get started with the possible 'state' values.
 */
{
  _XasDrawObjectBorder(me.hrbt_w,me.hrbt_c,state);
  XSetForeground(xasDisplay,me.hrbt_c,xasBlack);
  XDrawLine(xasDisplay,me.hrbt_w,me.hrbt_c,8,7,9,7);
  XDrawLine(xasDisplay,me.hrbt_w,me.hrbt_c,7,6,7,8);
  XDrawLine(xasDisplay,me.hrbt_w,me.hrbt_c,6,6,6,8);
  XDrawLine(xasDisplay,me.hrbt_w,me.hrbt_c,5,5,5,9);
  XDrawLine(xasDisplay,me.hrbt_w,me.hrbt_c,4,5,4,9);
  XDrawLine(xasDisplay,me.hrbt_w,me.hrbt_c,3,4,3,10);
};


static void
_XasMoveSCVertical(ScrolledCanvas me, XEvent event)
/*
 * This routine moves the vertical slider, according to
 * the method given in the 'event' XEvent.
 */
{
  Window	tmp_win1, tmp_win2;
  int		tmp_x1, tmp_y1, tmp_x2, tmp_y2;
  unsigned int	tmp_w1, tmp_h1, tmp_b1, tmp_d1, tmp_w2, tmp_h2, tmp_b2, tmp_d2;
  int		corrected_ypos, middle;

  if(event.xbutton.window == me.vubt_w) {
    /* Moving action done by pressing the up arrow button */
    XGetGeometry(xasDisplay,me.chld_w,&tmp_win1,&tmp_x1,&tmp_y1,&tmp_w1,&tmp_h1,&tmp_b1,&tmp_d1);
    XGetGeometry(xasDisplay,me.cnvs_w,&tmp_win2,&tmp_x2,&tmp_y2,&tmp_w2,&tmp_h2,&tmp_b2,&tmp_d2);
    if(tmp_h1 > tmp_h2) {
      corrected_ypos = (tmp_y1 + (tmp_h1 / 20));
      if(corrected_ypos > 0)
	corrected_ypos = 0;
      XMoveWindow(xasDisplay,me.chld_w,tmp_x1,corrected_ypos);
      XFlush(xasDisplay);
    }
  } else if(event.xbutton.window == me.vdbt_w) {
    /* Moving action done by pressing the down arrow button */
    XGetGeometry(xasDisplay,me.chld_w,&tmp_win1,&tmp_x1,&tmp_y1,&tmp_w1,&tmp_h1,&tmp_b1,&tmp_d1);
    XGetGeometry(xasDisplay,me.cnvs_w,&tmp_win2,&tmp_x2,&tmp_y2,&tmp_w2,&tmp_h2,&tmp_b2,&tmp_d2);
    if(tmp_h1 > tmp_h2) {
      corrected_ypos = (tmp_y1 - (tmp_h1 / 20));
      if(corrected_ypos < (tmp_h2 - tmp_h1))
	corrected_ypos = ((tmp_h2 - tmp_h1)-2);
      XMoveWindow(xasDisplay,me.chld_w,tmp_x1,corrected_ypos);
      XFlush(xasDisplay);
    }
  } else {
    /* Moving action done with the sliding button */ 
    if(event.xbutton.y >= 0) {
      XGetGeometry(xasDisplay,me.vsbb_w,&tmp_win1,&tmp_x1,&tmp_y1,&tmp_w1,&tmp_h1,&tmp_b1,&tmp_d1);
      XGetGeometry(xasDisplay,me.vsld_w,&tmp_win2,&tmp_x2,&tmp_y2,&tmp_w2,&tmp_h2,&tmp_b2,&tmp_d2);
      middle = (tmp_h2 / 2);
      if(event.xbutton.y < (middle+1))
	corrected_ypos = 0;
      else if(event.xbutton.y > (tmp_h1 - (middle+1)))
	corrected_ypos = tmp_h1 - tmp_h2;
      else 
	corrected_ypos = event.xbutton.y - middle;
      XMoveWindow(xasDisplay,me.vsld_w,0,corrected_ypos);
      _XasMoveChild_Vertical(me,(tmp_h1-tmp_h2),corrected_ypos);
    }
  }
};


static void
_XasMoveSCHorizontal(ScrolledCanvas me, XEvent event)
/*
 * This routine moves the horizontal slider, according to
 * the method given in the 'event' XEvent.
 */
{
  Window	tmp_win1, tmp_win2;
  int		tmp_x1, tmp_y1, tmp_x2, tmp_y2;
  unsigned int	tmp_w1, tmp_h1, tmp_b1, tmp_d1, tmp_w2, tmp_h2, tmp_b2, tmp_d2;
  int		corrected_xpos, middle;

  if(event.xbutton.window == me.hlbt_w) {
    /* Moving action done by pressing the left arrow button */
    XGetGeometry(xasDisplay,me.chld_w,&tmp_win1,&tmp_x1,&tmp_y1,&tmp_w1,&tmp_h1,&tmp_b1,&tmp_d1);
    XGetGeometry(xasDisplay,me.cnvs_w,&tmp_win2,&tmp_x2,&tmp_y2,&tmp_w2,&tmp_h2,&tmp_b2,&tmp_d2);
    if(tmp_w1 > tmp_w2) {
      corrected_xpos = (tmp_x1 + (tmp_w1 / 20));
      if(corrected_xpos > 0)
	corrected_xpos = 0;
      XMoveWindow(xasDisplay,me.chld_w,corrected_xpos,tmp_y1);
    }
  } else if(event.xbutton.window == me.hrbt_w) {
    /* Moving action done by pressing the right arrow button */
    XGetGeometry(xasDisplay,me.chld_w,&tmp_win1,&tmp_x1,&tmp_y1,&tmp_w1,&tmp_h1,&tmp_b1,&tmp_d1);
    XGetGeometry(xasDisplay,me.cnvs_w,&tmp_win2,&tmp_x2,&tmp_y2,&tmp_w2,&tmp_h2,&tmp_b2,&tmp_d2);
    if(tmp_w1 > tmp_w2) {
      corrected_xpos = (tmp_x1 - (tmp_w1 / 20));
      if(corrected_xpos < (tmp_w2 - tmp_w1))
	corrected_xpos = ((tmp_w2 - tmp_w1)-2);
      XMoveWindow(xasDisplay,me.chld_w,corrected_xpos,tmp_y1);
    }
  } else {
    /* Moving action done with the sliding button */ 
    if(event.xbutton.x >= 0) {
      XGetGeometry(xasDisplay,me.hsbb_w,&tmp_win1,&tmp_x1,&tmp_y1,&tmp_w1,&tmp_h1,&tmp_b1,&tmp_d1);
      XGetGeometry(xasDisplay,me.hsld_w,&tmp_win2,&tmp_x2,&tmp_y2,&tmp_w2,&tmp_h2,&tmp_b2,&tmp_d2);
      middle = (tmp_w2 / 2);
      if(event.xbutton.x < (middle+1))
	corrected_xpos = 0;
      else if(event.xbutton.x > (tmp_w1 - (middle+1)))
	corrected_xpos = (tmp_w1-tmp_w2);
      else
	corrected_xpos = (event.xbutton.x-middle);
      XMoveWindow(xasDisplay,me.hsld_w,corrected_xpos,0);
      _XasMoveChild_Horizontal(me,(tmp_w1-tmp_w2),corrected_xpos);
    }
  }
};


static void
_XasMoveChild_Vertical(ScrolledCanvas me, int free_height, int new_ycoord)
/*
 * This routine moves the child window vertically, depending on the 'free_height'
 * (space in which the sliding button can move) and on the 'new_ycoord' (new Y-pos
 * of the sliding button in its bounding box) values.
 */
{
  Window	tmp_win1, tmp_win2;
  int		tmp_x1, tmp_y1, tmp_x2, tmp_y2;
  unsigned int	tmp_w1, tmp_h1, tmp_b1, tmp_d1, tmp_w2, tmp_h2, tmp_b2, tmp_d2;

  int		tmp_res1, new_child_ypos;
  float		tmp_res2, tmp_res3, tmp_res4;

  XGetGeometry(xasDisplay,me.cnvs_w,&tmp_win1,&tmp_x1,&tmp_y1,&tmp_w1,&tmp_h1,&tmp_b1,&tmp_d1);
  XGetGeometry(xasDisplay,me.chld_w,&tmp_win2,&tmp_x2,&tmp_y2,&tmp_w2,&tmp_h2,&tmp_b2,&tmp_d2);
  tmp_res1 = tmp_h1 - tmp_h2;
  tmp_res4 = tmp_res1;
  tmp_res2 = tmp_res4 / free_height;
  tmp_res3 = tmp_res2 * new_ycoord;
  new_child_ypos = tmp_res3;
  if(tmp_res3 == tmp_res1)
    new_child_ypos -= 2;	/* Yes, this is needed to see the child's last line */

  XMoveWindow(xasDisplay,me.chld_w,tmp_x2,new_child_ypos);
};


static void
_XasMoveChild_Horizontal(ScrolledCanvas me, int free_width, int new_xcoord)
/*
 * This routine moves the child window horizontally, depending on the 'free_width'
 * (space in which the sliding button can move) and on the 'new_xcoord' (new X-pos
 * of the sliding button in its bounding box) values.
 */
{
  Window	tmp_win1, tmp_win2;
  int		tmp_x1, tmp_y1, tmp_x2, tmp_y2;
  unsigned int	tmp_w1, tmp_h1, tmp_b1, tmp_d1, tmp_w2, tmp_h2, tmp_b2, tmp_d2;

  int		tmp_res1, new_child_xpos;
  float		tmp_res2, tmp_res3, tmp_res4;

  XGetGeometry(xasDisplay,me.cnvs_w,&tmp_win1,&tmp_x1,&tmp_y1,&tmp_w1,&tmp_h1,&tmp_b1,&tmp_d1);
  XGetGeometry(xasDisplay,me.chld_w,&tmp_win2,&tmp_x2,&tmp_y2,&tmp_w2,&tmp_h2,&tmp_b2,&tmp_d2);
  tmp_res1 = tmp_w1 - tmp_w2;
  tmp_res4 = tmp_res1;
  tmp_res2 = tmp_res4 / free_width;
  tmp_res3 = tmp_res2 * new_xcoord;
  new_child_xpos = tmp_res3;
  if(tmp_res3 == tmp_res1)
    new_child_xpos -= 2;	/* Yes, this is needed to see the child's last column */

  XMoveWindow(xasDisplay,me.chld_w,new_child_xpos,tmp_y2);
};


static void
_XasResizeSC_VertSlider(ScrolledCanvas me)
/*
 * This routine resizes the vertical sliding button, depending on the
 * canvas' child height.
 */
{
  Window	tmp_win1, tmp_win2, tmp_win3;
  int		tmp_x1, tmp_x2, tmp_x3, tmp_y1, tmp_y2, tmp_y3;
  unsigned int	tmp_w1, tmp_h1, tmp_b1, tmp_d1, tmp_w2, tmp_h2, tmp_b2, tmp_d2, tmp_w3, tmp_h3, tmp_b3, tmp_d3;
  int		new_height;
  
  XGetGeometry(xasDisplay,me.vsbb_w,&tmp_win1,&tmp_x1,&tmp_y1,&tmp_w1,&tmp_h1,&tmp_b1,&tmp_d1);
  XGetGeometry(xasDisplay,me.cnvs_w,&tmp_win2,&tmp_x2,&tmp_y2,&tmp_w2,&tmp_h2,&tmp_b2,&tmp_d2);
  XGetGeometry(xasDisplay,me.chld_w,&tmp_win3,&tmp_x3,&tmp_y3,&tmp_w3,&tmp_h3,&tmp_b3,&tmp_d3);
  new_height = (tmp_h1 * tmp_h2) / tmp_h3;
  if(new_height < 11)
    new_height = 11;
  else if(new_height > tmp_h1)
    new_height = tmp_h1;
  XResizeWindow(xasDisplay,me.vsld_w,15,new_height);
};


static void
_XasResizeSC_HorizSlider(ScrolledCanvas me)
/*
 * This routine resizes the horizontal sliding button, depending on the
 * canvas' child height.
 */
{
  Window	tmp_win1, tmp_win2, tmp_win3;
  int		tmp_x1, tmp_x2, tmp_x3, tmp_y1, tmp_y2, tmp_y3;
  unsigned int	tmp_w1, tmp_h1, tmp_b1, tmp_d1, tmp_w2, tmp_h2, tmp_b2, tmp_d2, tmp_w3, tmp_h3, tmp_b3, tmp_d3;
  int		new_width;
  
  XGetGeometry(xasDisplay,me.hsbb_w,&tmp_win1,&tmp_x1,&tmp_y1,&tmp_w1,&tmp_h1,&tmp_b1,&tmp_d1);
  XGetGeometry(xasDisplay,me.cnvs_w,&tmp_win2,&tmp_x2,&tmp_y2,&tmp_w2,&tmp_h2,&tmp_b2,&tmp_d2);
  XGetGeometry(xasDisplay,me.chld_w,&tmp_win3,&tmp_x3,&tmp_y3,&tmp_w3,&tmp_h3,&tmp_b3,&tmp_d3);
  new_width = (tmp_w1 * tmp_w2) / tmp_w3;
  if(new_width < 11)
    new_width = 11;
  else if(new_width > tmp_w1)
    new_width = tmp_w1;
  XResizeWindow(xasDisplay,me.hsld_w,new_width,15);
};


static void
_XasReplaceSC_Sliders(ScrolledCanvas me)
/*
 * This routine tries to correctly replace the sliding buttons when
 * a ConfigureEvent is received from the child window
 */
{
  Window	tmp_win1, tmp_win2, tmp_win3, tmp_win4;
  int		tmp_x1, tmp_x2, tmp_x3, tmp_x4, tmp_y1, tmp_y2, tmp_y3, tmp_y4;
  unsigned int	tmp_w1, tmp_h1, tmp_b1, tmp_d1, tmp_w2, tmp_h2, tmp_b2, tmp_d2;
  unsigned int  tmp_w3, tmp_h3, tmp_b3, tmp_d3, tmp_w4, tmp_h4, tmp_b4, tmp_d4;
  int		new_hslider_xpos, new_vslider_ypos;
  int		tmp1, tmp2, tmp3;

  XGetGeometry(xasDisplay,me.chld_w,&tmp_win1,&tmp_x1,&tmp_y1,&tmp_w1,&tmp_h1,&tmp_b1,&tmp_d1);
  XGetGeometry(xasDisplay,me.cnvs_w,&tmp_win2,&tmp_x2,&tmp_y2,&tmp_w2,&tmp_h2,&tmp_b2,&tmp_d2);

  /* First, the horiz slider */
  XGetGeometry(xasDisplay,me.hsbb_w,&tmp_win3,&tmp_x3,&tmp_y3,&tmp_w3,&tmp_h3,&tmp_b3,&tmp_d3);
  XGetGeometry(xasDisplay,me.hsld_w,&tmp_win4,&tmp_x4,&tmp_y4,&tmp_w4,&tmp_h4,&tmp_b4,&tmp_d4);
  tmp1=(tmp_w3-tmp_w4); tmp2=(tmp_w2-tmp_w1); tmp3=(tmp_x1*tmp1);
  new_hslider_xpos = (tmp3/tmp2);
  XMoveWindow(xasDisplay,me.hsld_w,new_hslider_xpos,0);

  /* Now, the vert slider */
  XGetGeometry(xasDisplay,me.vsbb_w,&tmp_win3,&tmp_x3,&tmp_y3,&tmp_w3,&tmp_h3,&tmp_b3,&tmp_d3);
  XGetGeometry(xasDisplay,me.vsld_w,&tmp_win4,&tmp_x4,&tmp_y4,&tmp_w4,&tmp_h4,&tmp_b4,&tmp_d4);
  tmp1=(tmp_h3-tmp_h4); tmp2=(tmp_h2-tmp_h1); tmp3=(tmp_y1*tmp1);
  new_vslider_ypos = (tmp3/tmp2);
  XMoveWindow(xasDisplay,me.vsld_w,0,new_vslider_ypos);
};


extern void
_XasManageSC_Events(ScrolledCanvas me, XEvent event)
/*
 * This routine manages the X-Events for the 'me' ScrolledCanvas widget.
 * The X-Event to treat is 'event'.
 */
{
  switch(event.type) {
  case Expose:
    if(event.xexpose.count == 0) {
      if(event.xexpose.window == me.bbox_w) {
	_XasRedrawSC_BoundingBox(me);
      } else if(event.xexpose.window == me.vsld_w) {
	_XasRedrawSC_VSB(me,XAS_RELIEF_UPBOX);
      } else if(event.xexpose.window == me.hsld_w) {
	_XasRedrawSC_HSB(me,XAS_RELIEF_UPBOX);
      } else if(event.xexpose.window == me.vubt_w) {
	_XasRedrawSC_VUAB(me,XAS_RELIEF_UPBOX);
      } else if(event.xexpose.window == me.vdbt_w) {
	_XasRedrawSC_VDAB(me,XAS_RELIEF_UPBOX);
      } else if(event.xexpose.window == me.hlbt_w) {
	_XasRedrawSC_HLAB(me,XAS_RELIEF_UPBOX);
      } else if(event.xexpose.window == me.hrbt_w) {
	_XasRedrawSC_HRAB(me,XAS_RELIEF_UPBOX);
      }
    }
    break;

  case ButtonPress:
    if(event.xbutton.window == me.vubt_w) {
      _XasRedrawSC_VUAB(me,XAS_RELIEF_DNBOX);
      _XasMoveSCVertical(me,event);
    } else if(event.xbutton.window == me.vdbt_w) {
      _XasRedrawSC_VDAB(me,XAS_RELIEF_DNBOX);
      _XasMoveSCVertical(me,event);
    } else if(event.xbutton.window == me.hlbt_w) {
      _XasRedrawSC_HLAB(me,XAS_RELIEF_DNBOX);
      _XasMoveSCHorizontal(me,event);
    } else if(event.xbutton.window == me.hrbt_w) {
      _XasRedrawSC_HRAB(me,XAS_RELIEF_DNBOX);
      _XasMoveSCHorizontal(me,event);
    } else if(event.xbutton.window == me.vsbb_w) {
      if(me.enableSS == True)
	_XasRedrawSC_VSB(me,XAS_RELIEF_DNBOX);
      _XasMoveSCVertical(me,event);
    } else if(event.xbutton.window == me.hsbb_w) {
      if(me.enableSS == True)
	_XasRedrawSC_HSB(me,XAS_RELIEF_DNBOX);
      _XasMoveSCHorizontal(me,event);
    }
    break;

  case ButtonRelease:
    if(event.xbutton.window == me.vubt_w) {
      _XasRedrawSC_VUAB(me,XAS_RELIEF_UPBOX);
    } else if(event.xbutton.window == me.vdbt_w) {
      _XasRedrawSC_VDAB(me,XAS_RELIEF_UPBOX);
    } else if(event.xbutton.window == me.hlbt_w) {
      _XasRedrawSC_HLAB(me,XAS_RELIEF_UPBOX);
    } else if(event.xbutton.window == me.hrbt_w) {
      _XasRedrawSC_HRAB(me,XAS_RELIEF_UPBOX);
    } else if((event.xbutton.window == me.vsbb_w) && (me.enableSS == True)) {
      _XasRedrawSC_VSB(me,XAS_RELIEF_UPBOX);
    } else if((event.xbutton.window == me.hsbb_w) && (me.enableSS == True)) {
      _XasRedrawSC_HSB(me,XAS_RELIEF_UPBOX);
    }
    break;

  case MotionNotify:
    if(event.xbutton.window == me.vsbb_w) {
      _XasMoveSCVertical(me,event);
    } else if(event.xbutton.window == me.hsbb_w) {
      _XasMoveSCHorizontal(me,event);
    }
    break;

  case ConfigureNotify:
    if(event.xconfigure.window == me.chld_w) {
      _XasResizeSC_VertSlider(me);
      _XasResizeSC_HorizSlider(me);
      _XasReplaceSC_Sliders(me);
    }
    break;

  default:
    break;
  }
};


extern void
_XasMapSC_Windows(ScrolledCanvas me)
/*
 * This routine maps the 'me' ScrolledCanvas widget's windows.
 */
{
  XMapSubwindows(xasDisplay,me.cnvs_w);
  XMapSubwindows(xasDisplay,me.hsbb_w);
  XMapSubwindows(xasDisplay,me.vsbb_w);
  XMapSubwindows(xasDisplay,me.bbox_w);
};
